;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2013 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;;
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;; This file contains the generic definitions for hooks called
;;; throughout the BFIG search process. Extensions to BFIG should try
;;; to make use of these hooks as much as possible to extend BFIG's
;;; functionality.

(in-package :BFIG)

(defgeneric evaluate-next-node (next-node
								previous-node
								search-problem
								cutoff-cost
								maximum-visited-state-count
								full-assignment?
								disjoint?)
  (:documentation
   "This method is called once NEXT-NODE is dequeued. This method is
    responsible for:

    - Recording the new current node, if necessary.
  
    - Managing conflicts using information that can be extracted
      during this part of the search. This includes, but is not
      limited to, suboptimality conflicts in branch and bound search,
      and updating the conflict boundary in bounding conflict directed
      searches.

    - Updating search statistics.

    Returns (VALUES FOUND-CANDIDATE? CANDIDATE) where FOUND-CANDIDATE?
    is T if a candidate has been found, NIL otherwise. CANDIDATE is
    the search node containing the CANDIDATE, normally NEXT-NODE."))

(defgeneric record-node-inconsistent-with-conflict! (search-node 
													 new-conflict 
													 search-problem)
  (:documentation
   "This hook is called whenever a SEARCH-NODE is found to be
   inconsistent with a conflict."))

(defgeneric drop-node-due-to-conflicts! (node search-problem)
  (:documentation
   "This hook is called when a search node should be discarded due to
   a conflict."))