;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

(in-package :BFIG)

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;                   BFIG Testing                                ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Change:
;;; Updated todos.

;;; To Do:
;;; o Test alll conditional search methods.
;;; o Test informed search w/o conflicts.
;;; o Test all informed search methods.
;;; o Add logging functions.

;;; Fix:
;;; o Test 6 and 10 are failing, and need to be debugged.

(deftest-suite :BFIG
;;;    (log-BFIG-statistics)
;;;  (stop-and-print-BFIG-statistics)
    )

(deftest (test0 :bfig) ()

  ;; An informed search problem with one decision variable and one 
  ;; conflict.  The best candidate resolves the conflict.

  (let* ((variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2))))) 
		 (conflicts '(((= a hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test0
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= A low))
		  (next-assignment bfig-problem 
;;;						  :cutoff-cost nil
;;;						  :maximum-visited-state-count nil
;;;						  :full-assignment? T
;;;						  :disjoint? T
						   )
		  :test 
		  #'equal-state-assignments
		  :fail-info "a={(hi 3),(low 2)}, 
                      conflict: a=hi")))

;;;(deftest (test1 :bfig) ()
;;;
;;;  ;; An informed search problem with two decision variables and one 
;;;  ;; conflict to avoid.  The best candidate avoids the conflict.
;;;
;;;  (let* ((variable-descriptions
;;;		  '((a :domain ((hi :cost 3) (low :cost 2)))
;;;			(b :domain ((hi :cost 1) (low :cost 2)))))
;;;		 (conflicts '(((= a hi)(= b hi))))
;;;		 (bfig-problem (new-informed-search-problem 
;;;						variable-descriptions
;;;						conflicts
;;;						:name 'test1
;;;						:use-conflicts? t
;;;						:search-order :best-first)))
;;;	
;;;    (test '((= A hi)(= B low))
;;;		  (next-assignment bfig-problem)
;;;		  :test 
;;;		  #'equal-state-assignments
;;;		  :fail-info "a={(hi 3),(low 2)}, 
;;;                      b={(hi 1),(low 2)},
;;;                      conflict: a=hi, b=hi")
;;;
;;;	(test '((= A low)(= B low))
;;;		  (next-assignment bfig-problem)
;;;		  :test 
;;;		  #'equal-state-assignments
;;;		  :fail-info "a={(hi 3),(low 2)}, 
;;;                      b={(hi 1),(low 2)},
;;;                      conflict: a=hi, b=hi")
;;;	
;;;	(test '((= A low)(= B hi))
;;;		  (next-assignment bfig-problem)
;;;		  :test 
;;;		  #'equal-state-assignments
;;;		  :fail-info "a={(hi 3),(low 2)}, 
;;;                      b={(hi 1),(low 2)},
;;;                      conflict: a=hi, b=hi")))

(deftest (test1 :bfig) ()

  ;; An informed search problem with two decision variables and one 
  ;; conflict to avoid.  The best candidate avoids the conflict.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (conflicts '(((= a hi)(= b hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test1
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '(((= A hi)(= B low))
			((= A low)(= B low))
			((= A low)(= B hi)))
		  
		  (all-candidates bfig-problem)
		  :test 
		  #'equal-model-sequences
		  ;; equal-state-assignments
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={(hi 1),(low 2)},
                      conflict: a=hi, b=hi")))

(deftest (test2 :bfig) ()

  ;; An informed search problem with two decision variables and one 
  ;; conflict to avoid.  The best assignment violates the conflict.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (conflicts '(((= a hi)(= b low))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test2
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= A low)(= B low))			;cost 2+3=5
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      conflict: A=hi,B=low")
	
	(test '((= A hi)(= B hi))			;cost 3+1=4
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      conflict: A=hi,B=low")
	
	(test '((= A low)(= B hi))			;cost 2+1=3
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      conflict: A=hi,B=low")))

(deftest (test4 :bfig) ()

  ;; An informed search problem with two decision variables and two
  ;; conflicts to avoid.  One conflict subsumes the other.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (conflicts '(((= a hi)(= b low))
					  ((= a hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test4
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= A low)(= B low))			;cost 2+3=5
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      conflict: A=hi,B=low
		                        A=hi")
	
	(test '((= A low)(= B hi))			;cost 2+1=3
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      conflict: A=hi,B=low")))

(deftest (test5 :bfig) ()

  ;; An informed search problem with two decision variables and two
  ;; conflicts to avoid.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (conflicts '(((= a low)(= b low))
					  ((= a hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test5
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= A low)(= B hi))
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      conflict: A=low,B=low
		                        A=hi")))

(deftest (test6 :bfig) ()

  ;; An informed search problem with two decision variables and
  ;; three conflicts to avoid.  No solution.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (conflicts '(((= b low))
;;;					  ((= a hi))
					  ((= a low)(= b hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test6
						:use-conflicts? t
						:search-order :best-first)))
	
    (test 
	 `(((= a hi)(= b hi)))
	 (all-candidates bfig-problem)
	 :test 
	 #'equal-model-sequences
	 :fail-info "A={(hi 3),(low 2)} 
                 B={(hi 1),(low 3)}
                 conflict: B=low
                           A=low, B=hi"
	 )))

(deftest (test7 :bfig) ()

  ;; An informed search problem with two decision variables and
  ;; three conflicts to avoid.  No solution.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (conflicts '(
;;;					  ((= b low))
					  ((= a hi))
					  ((= a low)(= b hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test6
						:use-conflicts? t
						:search-order :best-first)))
	
    (test 
	 `(((= a low)(= b low)))
	 (all-candidates bfig-problem)
	 :test 
	 #'equal-model-sequences
	 :fail-info "A={(hi 3),(low 2)} 
                 B={(hi 1),(low 3)}
                 conflict: A=hi
                           A=low, B=hi")))

(deftest (test8 :bfig) ()

  ;; An informed search problem with two decision variables and
  ;; three conflicts to avoid.  No solution.

  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (conflicts '(((= b low))
					  ((= a low)(= b hi))
					  ((= a hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test6
						:use-conflicts? t
						:search-order :best-first)))
	
    (test 
	 '()
	 (all-candidates bfig-problem)
	 :test 
	 #'equal-model-sequences
	 :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      conflict: B=low
		                        A=hi
                                A=low, B=hi")))

;;;(deftest (test10 :bfig) ()
;;;
;;;  ;; An informed search problem with four decision variables and one 
;;;  ;; conflict to avoid.
;;;
;;;  (let* ((variable-descriptions 
;;;		  '((a :domain ((hi :cost 3) (low :cost 2)))
;;;			(b :domain ((hi :cost 1) (low :cost 2)))
;;;			(c :domain ((hi :cost 1) (low :cost 2)))
;;;			(d :domain ((hi :cost 1) (low :cost 2)))))
;;;		 (conflicts '(((= a hi)(= b hi))))
;;;		 (bfig-problem (new-informed-search-problem 
;;;						variable-descriptions
;;;						conflicts
;;;						:name 'test10
;;;						:use-conflicts? t
;;;						:search-order :best-first)))
;;;	
;;;    (test '(((= A hi)(= B low)(= C low)(= D low)) ; 1 = 9
;;;			((= A low)(= B low)(= C low)(= D low)) ;2 = 8
;;;			((= A hi)(= B low)(= C hi)(= D low)) ;  3 = 8
;;;			((= A hi)(= B low)(= C low)(= D hi)) ;  4 = 8
;;;			((= A low)(= B low)(= C low)(= D hi)) ; 5 = 7
;;;			((= A hi)(= B low)(= C hi)(= D hi)) ;   6 = 7
;;;			((= A low)(= B low)(= C hi)(= D low)) ; 7 = 7
;;;			((= A low)(= B hi)(= C low)(= D low)) ; 8 = 7
;;;			((= A low)(= B hi)(= C hi)(= D low)) ;  9 = 6
;;;			((= A low)(= B hi)(= C low)(= D hi)) ; 10 = 6 
;;;			((= A low)(= B low)(= C hi)(= D hi)) ; 11 = 6
;;;			((= A low)(= B hi)(= C hi)(= D hi)) ;  12 = 5
;;;			)
;;;		  
;;;		  (all-candidates bfig-problem)
;;;		  :test 
;;;		  #'equal-model-sequences
;;;		  ;; equal-state-assignments
;;;		  
;;;		  :fail-info "a={(hi 3),(low 2)}, 
;;;                      b=c=d={(hi 1),(low 2)},
;;;                      e=f={hi,low},
;;;                      conflict: a=hi, b=hi
;;;                      Soln: <{a=h,b=l,c=l,d=l},              ;cost 9 *
;;;                             {a=l,b=l,c=l,d=l},              ;cost 8 *
;;;                             {a=h,b=l,c=h,d=l},              ;cost 8 *
;;;                             {a=h,b=l,c=l,d=h},              ;cost 8 *
;;;                             {a=l,b=l,c=l,d=h},              ;cost 7 *
;;;                             {a=h,b=l,c=h,d=h},              ;cost 7 *
;;;
;;;                             {a=l,b=l,c=l,d=h},              ;cost 7 *
;;;                             {a=l,b=l,c=h,d=l},              ;cost 7 *
;;;                             {a=l,b=l,c=h,d=l},              ;cost 7 *
;;;                             {a=l,b=h,c=l,d=l}               ;cost 7 *
;;;                             {a=l,b=l,c=h,d=h},              ;cost 6
;;;                             {a=l,b=h,c=l,d=h}               ;cost 6
;;;                             {a=l,b=h,c=h,d=l}               ;cost 6
;;;                             {a=l,b=h,c=h,d=h}>              ;cost 5")))


;;; Change:
;;; o Added the following tests.
(deftest (test10 :bfig) ()

  ;; An informed search problem with one conditional variable,
  ;; which is always active, and one conflict.  
  ;; The best candidate resolves the conflict.

  (let* ((variable-descriptions 
		  `((a :domain ((hi :cost 3) (low :cost 2))
			   :guard ,*always-active*))) 
		 (conflicts '(((= a hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test10
						:conditional? t
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= A low))
		  (next-assignment bfig-problem 
;;;						  :cutoff-cost nil
;;;						  :maximum-visited-state-count nil
;;;						  :full-assignment? T
;;;						  :disjoint? T
						   )
		  :test 
		  #'equal-state-assignments
		  :fail-info "a={(hi 3),(low 2)}, 
                      conflict: a=hi")))

(deftest (test11 :bfig) ()

  ;; An informed search problem with two decision variables and one 
  ;; conflict to avoid.  The best candidate avoids the conflict.

  (let* ((variable-descriptions
		  `((a :domain ((hi :cost 3) (low :cost 2))
			   :guard ,*always-active*)
			(b :domain ((hi :cost 1) (low :cost 2))
			   :guard ,*always-active*)))
		 (conflicts '(((= a hi)(= b hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test11
						:conditional? t
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '(((= A hi)(= B low))
			((= A low)(= B low))
			((= A low)(= B hi)))
		  
		  (all-candidates bfig-problem)
		  :test 
		  #'equal-model-sequences
		  ;; equal-state-assignments
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={(hi 1),(low 2)},
                      conflict: a=hi, b=hi")))

;;; Fix:
;;; - Test12 calls initializes three times, 
;;;   should only initialize first two times.

;;; - Is the cost of Candidate A=low computed correctly?
;;;   What is the cost of the unassigned variables?

(deftest (test12 :bfig) ()

  ;; An informed search problem with two decision variables and one 
  ;; conflict to avoid.  The best candidate avoids the conflict.

  (let* ((variable-descriptions
		  `((a :domain ((hi :cost 3) (low :cost 2))
			   :guard ,*always-active*)
			(b :domain ((hi :cost 1) (low :cost 2))
			   :guard (= a hi))))
		 (conflicts '(((= a hi)(= b hi))))
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test12
						:conditional? t
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '(((= A hi)(= B low))
			((= A low)))
		  
		  (all-candidates bfig-problem)
		  :test 
		  #'equal-model-sequences
		  ;; equal-state-assignments
		  :fail-info "Expected: ~a.~%Received: ~a.")))

;;; Test for many variables with guards.

(deftest (test-multiple-nested-decisions :bfig) ()
  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 0) (low :cost 0)))
			(a1 :domain ((hi :cost 11) (low :cost 0))
				:guard (= a hi))
			(a2 :domain ((hi :cost 10) (low :cost 0))
				:guard (= a low))
			(b :domain ((hi :cost 0) (low :cost 0)))
			(b1 :domain ((hi :cost 11) (low :cost 0))
				:guard (= b hi))
			(b2 :domain ((hi :cost 10) (low :cost 0))
				:guard (= b low))))
		 (conflicts nil)
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test-multiple-nested-decisions
						:conditional? t
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= a hi) (= a1 hi) (= b hi) (= b1 hi))
		  
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  ;; equal-state-assignments
		  :fail-info "Expected: ~a.~%Received: ~a."))
)

(deftest (test-optimality-with-guarded-variables :bfig) ()
  (let* ((variable-descriptions
		  '((a :domain ((hi :cost 1) (low :cost 0)))
			(a1 :domain ((hi :cost 10) (low :cost 0))
				:guard (= a hi))
			(a2 :domain ((hi :cost 20) (low :cost 0))
				:guard (= a low))
			(b :domain ((hi :cost 0) (low :cost 1)))
			(b1 :domain ((hi :cost 0) (low :cost 20))
				:guard (= b hi))
			(b2 :domain ((hi :cost 10) (low :cost 0))
				:guard (= b low))))
		 (conflicts nil)
		 (bfig-problem (new-informed-search-problem 
						variable-descriptions
						conflicts
						:name 'test-optimality-with-guarded-variables
						:conditional? t
						:use-conflicts? t
						:search-order :best-first)))
	
    (test '((= a low) (= a2 hi) (= b hi) (= b1 low))
		  
		  (next-assignment bfig-problem)
		  :test 
		  #'equal-state-assignments
		  ;; equal-state-assignments
		  :fail-info "Expected: ~a.~%Received: ~a."))
)
