;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

;;; For references relating to this code see:
;;;   [1] R. Dechter. Constraint Processing. Morgan Kaufmann, San Francisco,
;;;   California, 2003.

(in-package :mtk-csp-tcsp)


;;; From [1] pg. 346
;;;
;;; <constraints>   := (<constraint>*)
;;; <constraint>    := (:scope <scope> :relation <relation>)
;;; <scope>         := (<variable> <variable>)
;;; <relation>      := (<interval>+)
;;; <interval>      := (:lower-bound <lower-bound> :upper-bound <upper-bound>)
;;; <lower-bound>   := (<number> :inequality <lb-inequality>) | (<number>)
;;; <upper-bound>   := (<number> :inequality <ub-inequality>) | (<number>)
;;; <lb-inequality> := #'> | #'>=
;;; <ub-inequality> := #'< | #'<=
;;; 
;;; Assumptions:
;;;   * Intervals:
;;;       The list of intervals is a non-intersecting disjunction of intervals.
;;;   * Time-Bounds:
;;;       If no :inequality function is specified for the <lower-bound> or
;;;       <upper-bound>, then the non-strict versions #'>= and #'<= are used
;;;       by default."

(defun make-stn-constraint-descriptions-figure-12.7 ()
  `((:scope (0 1) :relation ((:lower-bound (10) :upper-bound (20))))
    (:scope (0 4) :relation ((:lower-bound (60) :upper-bound (70))))
    (:scope (1 2) :relation ((:lower-bound (60) :upper-bound (,*inf* :inequality ,#'<))))
    (:scope (3 2) :relation ((:lower-bound (10) :upper-bound (20))))
    (:scope (3 4) :relation ((:lower-bound (20) :upper-bound (30))))))

(defun make-stn-figure-12.7 ()
  (let ((constraint-descriptions
         (make-stn-constraint-descriptions-figure-12.7))
        (stn (make-instance 'stn)))
    (add-constraints constraint-descriptions stn)
    stn))
