;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

;;; For references relating to this code see:
;;;   [1] R. Dechter. Constraint Processing. Morgan Kaufmann, San Francisco,
;;;   California, 2003.

(in-package :mtk-csp-tcsp)


;;;--------------------------------------------------------------------------------
;;; Simple Temporal Network (STN) CLASS
;;;--------------------------------------------------------------------------------

(defclass stn (tcsp)
  ())

;;;-------------------------------------------------------------------------------
;;; STN to distance graph
;;;-------------------------------------------------------------------------------

(defmethod stn-to-distance-graph ((stn stn))
  (let ((g (make-instance 'weighted-directed-graph 
	     :default-weight (make-inf-upper-bound))))
    (add-vertices (get-vertices stn) g)
    (dolist (c_i (get-constraint-list stn))
      (let ((interval (first (get-relation c_i)))
	    (s_i (get-scope c_i)))
	(add-weighted-edge (list s_i (get-upper-bound interval)) g)
	(add-weighted-edge
	 (list (reverse s_i) 
	       (make-bound 
		(* -1 (get-value (get-lower-bound interval)))
		:inequality
		(if (eql (get-inequality (get-lower-bound interval)) #'>=)
		    #'<= #'<)))
	 g)))))

;;;-------------------------------------------------------------------------------
;;; Add constraint(s) to a STN
;;;-------------------------------------------------------------------------------

(defmethod add-constraint ((c_i constraint) (n stn))
  ;; STN does not allow disjunctive intervals.
  (assert (eq (length (get-relation c_i)) 1))
  (call-next-method c_i n))
