(in-package #:drake2)

(defmethod compute-ldg-span ((ldg mtk-graph:labeled-weighted-directed-graph) (start-vertex t) (end-vertex t) &key (env nil) (inf *inf*) (-inf *-inf*))
  (cond 
	;; if the environment is a description, convert it
	((listp env)  
	 (setf env (get-clause-from-descriptions (get-state-space ldg) env)))
	;; if the env is already a state-clause (or a descendant), do nothing
	((typep env 'state-clause))
	;; otherwise, use the universal-clause
	(:otherwise 
	 (setf env (get-universal-clause (get-state-space ldg)))))

  (let ((vertex-list (get-vertices ldg))
		;; The vertex-ht and num-veritices is filled
		;;   in the following loop.
		(vertex-ht (make-hash-table))
		(num-vertices 0) 
		;; The edge-list and num-edges is filled 
		;;   in the following loop
		;; edge-list takes the form ((edge weight) (edge weight))
		(edge-list '()) 
		(num-edges 0))
	;; Create a hash table that maps events to an index number.
	(loop for vertex in vertex-list do
		 (setf (gethash vertex vertex-ht) num-vertices)
		 (incf num-vertices))
	;; Create the edge-list and num-edges
	(dolist (edge (get-edge-list (get-edges ldg)))
	  (let ((edge-lvs (gethash edge (get-weight-ht-edge ldg))))
		(dolist (pair (lvs-pairs edge-lvs)) 
		  (when (clause-subsumed? (pair-environment pair) env)
			(push (list edge (pair-value pair)) edge-list)
			(incf num-edges))
		  )))
	;; Create the A, b, and c  matricies.
	(let*  (;; Each edge in the edge list is one constraint
			(m num-edges)
			;; Each event has an occurence time
			;; plus one slack variable, one each for edge in the edge-list.
			(n (+ num-vertices num-edges))
			(A (make-array `(,m ,n) :element-type 'number :initial-element 0))
			(b (make-array m :element-type 'number :initial-element 0))
			(c (make-array n :element-type 'number :initial-element 0))
			(ldg-starti (gethash start-vertex vertex-ht))
			(ldg-endi (gethash end-vertex vertex-ht)))
	  ;; Create the A and b matrix
	  (loop for (edge weight) in edge-list
		   with mi = 0 do
		   (let ((starti (gethash (first (get-source-vertices edge)) vertex-ht))
				 (endi (gethash (first (get-target-vertices edge)) vertex-ht)))
			 ;; b - a + s1 = weight, where s1>=0
			 (setf (aref A mi endi) 1)
			 (unless (eql ldg-starti starti)
			   (setf (aref A mi starti) -1))
			 (setf (aref A mi (+ num-vertices mi)) 1)
			 (setf (aref b mi) weight)
			 (incf mi)
			 ))
	  ;; Create the c matrix
	  (setf (aref c ldg-endi) 1)
	  ;; 
	  (let ((min-sol nil)
	  		(min-c nil)
	  		(max-sol nil)
	  		(max-c nil))
	  	;; Solve the simplex for the minimum case
	  	(multiple-value-setq (min-sol min-c)
	  	  (simplex:simplex A b c))
	  	;; Create the c matrix by multiplying by -1
		;; solve for the maximum case
		(setf (aref c ldg-endi) -1)
	  	;; Solve the simplex for the maximum case
	  	(multiple-value-setq (max-sol max-c)
	  	  (simplex:simplex A b c))
	  	;; Return the min and max values
		(if min-sol
			(setf min-c (second min-c))
			(setf min-c inf))
		(if max-sol
			(setf max-c (- (second max-c)))
			(setf max-c -inf))
	  	(values min-c max-c)
		))


	)
)