(in-package :drake2)

(defvar *one-step-latency* 0)

(defmacro timing (&body forms)
  "Macro returns the time in seconds (it's usually a fraction) of system time used
to run the FORMS."
    (let ((run1 (gensym))
		  (run2 (gensym))
		  (result (gensym)))
	  `(let* ((,run1 (get-internal-run-time))
			  (,result (progn ,@forms)))
		 (let ((,run2 (get-internal-run-time)))
		   (list ,result (/ (- ,run2 ,run1) internal-time-units-per-second))) )))

(defmacro measure-one-step-latency (&body forms)
  `(setf *one-step-latency* 
		(max *one-step-latency* 
			 (second (timing ,@forms)))))

(defun run-all-benchmarking ()
  "Run a suite of benchmark problems."

  ;;Order 2
  ;; (loop for order in '(2) 
  ;; 	 do (loop for asize in '(13);(1 2 3 4 5 6 7 8 9 10)
  ;; 		   do 
  ;; 		   (run-benchmarking asize order 100 t t t t nil)
  ;; 			 (run-benchmarking asize order 100 t t t nil nil)))
  ;; ;;Order 3
  ;; (loop for order in '(3)
  ;;  do (loop for asize in '(9);(1 2 3 4 5 6 7 8)
  ;; 		   do 
  ;; 			 (run-benchmarking asize order 100 t t t t nil)
  ;; 			 (run-
  ;;			  benchmarking asize order 100 t t t nil nil)))
  ;;TPNs
  (loop for order in '(2 3)
  	 do (loop for size in (if (eql 2 order)
							  '(11 12)
							  '(6 7))
  		   do 
  		   (run-benchmarking size order 100 t))))


(defun run-benchmarking (size order problem-count new-problem?)

(with-open-file (stream (format nil "~~/lstnBenchmarks/results~a_~a.m"
								size order)
						:direction :output :if-exists :supersede)

  (loop
	 with good-problems-run = 1
	 until (eql good-problems-run (+ 1 problem-count))
	 with success? = t
	   
	 do
	   (setf *one-step-latency* 0)
	   (multiple-value-bind (random-ldg constraints) 
		   (random-structured-ldg 
			size size order 
			4 10 20 (* 4 size) 
			(* 8 size) (* 8 size)
			(* 5 size) new-problem? t :problem-number good-problems-run)

										;		 (format t "problem:~a ~%" random-ldg)
		 #+allegro (excl::gc t)
		 (org.tfeb.hax.memoize:clear-memoized-functions) ;;clear the memos
		 (format t "try to find problem: ~a ~a ~a ~%   " 
				 order size good-problems-run)
		 (debug-print-for :drake t "Problem: ~a" random-ldg)
		 (setf success? t)
		 


		 (let ((all-stns (extract-stns random-ldg constraints))
			   (compiled-stns nil)
			   (stn-compile-time 0)
			   (stn-compile-size 0))
										;			 (format t "*****raw: ~a ~a~%" (get-all-consistent-clauses
										;(get-state-space random-ldg)) all-stns )
		   ;;compile all of them
		   (setf stn-compile-time
				 (second (timing 
						   (dolist (one-stn all-stns)
							 (handler-case
						 
								 (push (construct-dispatchable-graph 
										one-stn) 
									   compiled-stns)
					   
							   (mtk-graph-algorithms::infeasible-problem ()			 
								 ))))))

		   (setf stn-compile-size (+ 
								   (length
									(format nil "~a" 
											(subseq 
											 (get-all-consistent-clauses 
											  (get-state-space random-ldg))
											 0 (length compiled-stns)
											 ))) 
								   (apply '+ (mapcar 'compute-size-of-stn 
													 compiled-stns))))

		   (when (null compiled-stns)
			 (setf success? nil)
			 (format t "bad problem! ~%"))
		   
		   (when success?				;when success, output

			 (let ((all-dispatchers (mapcar #'(lambda (x)
												(make-instance 'ldg-dispatch 
															   :ldg x))
											compiled-stns)))
			 
										;			 (format t "stp: ~a~%" (length compiled-stns))
										;			 (format t "*****stns: ~a~%"compiled-stns )

			   (format stream "results{~a,~a}.StnNumConsistent(~a)=~a;~%"
					   size order good-problems-run (length compiled-stns))
			   (format stream "results{~a,~a}.StnCompileTime(~a)=~a;~%"
					   size order good-problems-run stn-compile-time)
			   (format stream "results{~a,~a}.StnCompileSize(~a)=~a;~%"
					   size order good-problems-run stn-compile-size)
			   (format stream "results{~a,~a}.StnLatency(~a)=~a;~%"
					   size order good-problems-run 
					   (second 
						(timing 
						  (dolist (a-dispatcher all-dispatchers)
							(offline-execute! 
							 a-dispatcher
							 :max-iterations 1)))))))


		   )


		 (when (and t success?)
		   #+allegro (excl::gc t)
		   (org.tfeb.hax.memoize:clear-memoized-functions) ;;clear the memos
		   (debug-print-for :drake t "Begin full Drake")
										;		   (format t "Begin full Drake")
		   (handler-case 
			   (let* ((compiled-timing (timing (construct-dispatchable-graph 
												random-ldg)))
					  (compile-time (second compiled-timing))
					  (compiled-ldg (first compiled-timing))
					  (compiled-size (compute-size-of-ldg compiled-ldg))
					
					
					  (dispatcher (make-instance 'ldg-dispatch :ldg compiled-ldg)))

			     (debug-print-for :drake t "Compiled problem: ~a" compiled-ldg)
				 (format stream "results{~a,~a}.DrakeNumConsistent(~a)=~a;~%"
						 size order good-problems-run 
						 (length (get-all-consistent-clauses 
								  (get-state-space compiled-ldg))))
										;			 (format t "dtp: ~a~%" (length (get-all-consistent-clauses 
										;								  (get-state-space compiled-ldg))))
				 (offline-execute! dispatcher)

				 
				 ;;record the Drake results
				 (format stream "results{~a,~a}.DrakeCompileTime(~a)=~a;~%"
						 size order good-problems-run compile-time)
				 (format stream "results{~a,~a}.DrakeCompileSize(~a)=~a;~%"
						 size order good-problems-run compiled-size)
				 (format stream "results{~a,~a}.DrakeLatency(~a)=~a;~%"
						 size order good-problems-run *one-step-latency*)


				 (format t "trace: ~a~%" (dispatch-trace dispatcher))
				 ;;assert the trace is valid
				 (assert (execution-trace-valid? compiled-ldg constraints (dispatch-trace dispatcher)))
				 )

	
			 ;;handle infeasible problems
			 (mtk-graph-algorithms::infeasible-problem ()
			   (format t "bad problem v2~%")
			   (assert nil) ;;shouldn't find this here!
			 
			   (setf success? nil)))))
	   
	   when success?
	   do (setf good-problems-run (+ 1 good-problems-run))


)


)
)

(defmethod compute-size-of-ldg ((g mtk-graph:labeled-weighted-directed-graph))

  (let ((fstr (make-array '(0) :element-type 'base-char
						  :fill-pointer 0 :adjustable t))
		(all-envs nil))
	(with-output-to-string (stream fstr)

	  (format stream "~a~%" (get-vertices g))
	  (loop for variable in (state-space::state-space-variables (get-state-space g))
		 do (format stream "~a"(state-space::state-variable-domain variable)))
	  (format stream "~%~a~%" (state-space::state-space-conflicts (get-state-space g)))

	  (loop for edge in (get-edge-list g)
		 do (format stream "~a(" (get-vertices edge))
		 do (loop for pair in (lvs-pairs (get-weight edge g))
			   do (format stream "(~a,    )" (pair-value pair))
				 (pushnew (pair-environment pair) all-envs))
		 do (format stream ")~%"))
	  (format stream "~a" all-envs))

	;(format t "stn: ~a ~%" fstr)

	(length fstr)))

(defmethod compute-size-of-stn ((g mtk-graph:labeled-weighted-directed-graph))

  (let ((fstr (make-array '(0) :element-type 'base-char
						  :fill-pointer 0 :adjustable t)))
	(with-output-to-string (stream fstr)

	  (format stream "~a~%" (get-vertices g))
	 

	  (loop for edge in (get-edge-list g)
		 do (format stream "~a=" (get-vertices edge))
		 do (loop for pair in (lvs-pairs (get-weight edge g))
			   do (format stream "~a" (pair-value pair)))
		 do (format stream "~%")))

;	(format t "stn: ~a ~%" fstr)
	(length fstr)))

(defun extract-stns (ldg constraints)

  (loop for env in (get-all-consistent-clauses (get-state-space ldg))
	 collect (let* ((state-space (make-instance 'state-space-w-conflicts))
					(a-stn (make-instance 'labeled-weighted-directed-graph 
										:state-space state-space)))

			   (loop for constraint in constraints 
					
					when (clause-subsumed? env (fifth constraint))
					do (update-labeled-weighted-edge 
						(list (list (first constraint)
									(second constraint))
							  (fourth constraint)
							  (get-universal-clause state-space))
						a-stn)
					and do (update-labeled-weighted-edge 
						(list (list (second constraint)
									(first constraint))
							  (- (third constraint))
							  (get-universal-clause state-space))
						a-stn))
			   ;;return the stn from the let so it is collected
			   a-stn)))