(in-package #:drake2)

(defvar *dispatch* nil)

(defun test-dispatch-ldg-online ()
  (let*  (;; Create the labeled-weighted-directed graph
		  (var-descriptions '((a :domain ((2)))))
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))
		  ;; Create the data structure to hold onto activities
		  (activities (make-activities)))

	(insert-new-constraint graph activities :start 'a :end 'b :lb 1 :ub 2 :guard '((= a 2)) :primitive "a->b")
	;;(insert-new-constraint graph activities :start 'b :end 'c :lb 1 :ub 1.5 :guard '((= a 2)) :primitive nil)
	(insert-new-constraint graph activities :start 'b :end 'c :lb 1.2 :ub 2 :guard '((= a 2)) :primitive "b->c")
	(insert-new-constraint graph activities :start 'a :end 'c :lb 1 :ub 3 :guard '((= a 2)) :primitive "a->c")

	;; Compile the graph into dispatchable form and Create the dispatcher
	(let* ((dispatchable-graph (construct-dispatchable-graph graph))
		   (dispatch (make-instance 'online-ldg-dispatch :start 'a 
									  :ldg dispatchable-graph
									  :time-tol 0.1
									  :activities (activities-list activities))))
	  ;; Save the dispatcher into a global variable, so we can access
	  ;; the dispatcher's mailbox and force-stop it, if necessary.
	  (setf *dispatch* dispatch)
	  
	  (format t "dispatchable-graph: ~a~%" dispatchable-graph)

	  ;; To run the dispatcher in a separate thread, use this line:
	  ;;(start-dispatcher dispatch :callback #'test-callback)

	  ;; To run the dispatcher in THIS thread, use the following lines:
	  (online-execute! dispatch :activity-callback #'test-callback)
	  (format t "trace: ~a~%" (dispatch-trace dispatch))
	  (format t "result: ~a~%" (dispatch-ldg dispatch))
	  )
	))

(defun test-dispatch-ldg-online2 ()
  (let*  (;; Create the labeled-weighted-directed graph
		  (var-descriptions '((a :domain ((2)))))
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))
		  ;; Create the data structure to hold onto activities
		  (activities (make-activities)))

	(insert-new-constraint graph activities :start 'a :end 'b :lb 1 :ub 2 :guard '((= a 2)) :primitive "a->b")
	;;(insert-new-constraint graph activities :start 'b :end 'c :lb 1 :ub 1.5 :guard '((= a 2)) :primitive nil)
	(insert-new-constraint graph activities :start 'b :end 'c :lb 1.2 :ub 2 :guard '((= a 2)) :primitive "b->c")
	(insert-new-constraint graph activities :start 'a :end 'c :lb 1 :ub 3 :guard '((= a 2)) :primitive "a->c")

	;; Compile the graph into dispatchable form and Create the dispatcher
	(let* ((dispatchable-graph (construct-dispatchable-graph graph))
		   (dispatch (make-instance 'online-ldg-dispatch :start 'a 
									  :ldg dispatchable-graph
									  :time-tol 0.1
									  :activities (activities-list activities))))
	  ;; Save the dispatcher into a global variable, so we can access
	  ;; the dispatcher's mailbox and force-stop it, if necessary.
	  (setf *dispatch* dispatch)
	  
	  (format t "dispatchable-graph: ~a~%" dispatchable-graph)

	  ;; To run the dispatcher in a separate thread, use this line:
	  ;;(start-dispatcher dispatch :callback #'test-callback)

	  ;; To run the dispatcher in THIS thread, use the following lines:
	  (online-execute! dispatch :ignore-missed-activities t)	 
	  (format t "trace: ~a~%" (dispatch-trace dispatch))
	  (format t "result: ~a~%" (dispatch-ldg dispatch))
	  )
	))


(defun test-dispatch-ldg-online3 ()
  "Tests for dispatching ordering, where a->b->c->d. 
   a->b and c->d have primitives, and b->c is a temporal constraint [0,x], x>0.
   Drake may dispatch c->d before a->b. 
   In this test, the events (F and G) are dispatched before (D E)."
  (let*  (;; Create the labeled-weighted-directed graph
		  (var-descriptions '() )
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))
		  ;; Create the data structure to hold onto activities
		  (activities (make-activities)))

	(insert-new-constraint graph activities :start 'start :end 'a :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'a :end 'b :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'b :end 'c :lb 3.8504555 :ub 4.1615877 :primitive "flyA")
	(insert-new-constraint graph activities :start 'c :end 'd :lb 0.75 :ub 1.25 :primitive "lookA")
	(insert-new-constraint graph activities :start 'd :end 'e :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'e :end 'end :lb 0 :ub *inf*)

	 ;; change the lb value to 0.0001 to avoid this problem.
	(insert-new-constraint graph activities :start 'e :end 'f :lb 0 :ub *inf*)

	(insert-new-constraint graph activities :start 'start :end 'f :lb 0 :ub *inf*)
	(insert-new-constraint graph activities :start 'f :end 'g :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'g :end 'h :lb 3.92373 :ub 4.0792966 :primitive "flyB")
	(insert-new-constraint graph activities :start 'g :end 'h :lb 0 :ub 30)
	(insert-new-constraint graph activities :start 'h :end 'i :lb 0.75 :ub 1.25 :primitive "lookB")
	(insert-new-constraint graph activities :start 'i :end 'j :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'j :end 'end :lb 0 :ub *inf*)
	

	;; Compile the graph into dispatchable form and Create the dispatcher
	(let* ((dispatchable-graph (construct-dispatchable-graph graph))
		   (dispatch (make-instance 'online-ldg-dispatch :start 'start
									  :ldg dispatchable-graph
									  :time-tol 0.5
									  :activities (activities-list activities))))
	  ;; Save the dispatcher into a global variable, so we can access
	  ;; the dispatcher's mailbox and force-stop it, if necessary.
	  (setf *dispatch* dispatch)
	  
	  (format t "dispatchable-graph: ~a~%" dispatchable-graph)

	  ;; To run the dispatcher in a separate thread, use this line:
	  ;;(start-dispatcher dispatch :callback #'test-callback)

	  ;; To run the dispatcher in THIS thread, use the following lines:
	  (online-execute! dispatch :ignore-missed-activities t)	 
	  (format t "trace: ~a~%" (dispatch-trace dispatch))
	  (format t "result: ~a~%" (dispatch-ldg dispatch))
	  )
	))

(defun test-dispatch-ldg-online4 ()
  (let*  (;; Create the labeled-weighted-directed graph
		  (var-descriptions '() )
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))
		  ;; Create the data structure to hold onto activities
		  (activities (make-activities)))

	(insert-new-constraint graph activities :start 'start :end 'a :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'a :end 'b :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'b :end 'c :lb 0.0 :ub 1800.0)
	(insert-new-constraint graph activities :start 'b :end 'c :lb 117.54369 :ub 122.20402 :primitive "fly")
	(insert-new-constraint graph activities :start 'c :end 'd :lb 1.0 :ub 1.5 :primitive "look")
	(insert-new-constraint graph activities :start 'd :end 'e :lb 0 :ub 0)
	(insert-new-constraint graph activities :start 'e :end 'end :lb 0 :ub *inf*)	

	;; Compile the graph into dispatchable form and Create the dispatcher
	(let* ((dispatchable-graph (construct-dispatchable-graph graph))
		   (dispatch (make-instance 'online-ldg-dispatch :start 'start
									  :ldg dispatchable-graph
									  :time-tol 0.5
									  :activities (activities-list activities))))
	  ;; Save the dispatcher into a global variable, so we can access
	  ;; the dispatcher's mailbox and force-stop it, if necessary.
	  (setf *dispatch* dispatch)
	  
	  (format t "dispatchable-graph: ~a~%" dispatchable-graph)

	  ;; To run the dispatcher in a separate thread, use this line:
	  ;;(start-dispatcher dispatch :callback #'test-callback)

	  ;; To run the dispatcher in THIS thread, use the following lines:
	  (online-execute! dispatch :activity-callback #'test-callback :ignore-missed-activities t)	 
	  (format t "trace: ~a~%" (dispatch-trace dispatch))
	  (format t "result: ~a~%" (dispatch-ldg dispatch))
	  )
	))


;; Dispatch a TPN online that is read from a file
(defun test-dispatch-tpn-online (tpn-in-pathname &key (execution-monitor nil)
								 (activity-callback #'test-callback))
  ;; Read in and parse the TPN in file
  (with-open-file (tpn-in-stream tpn-in-pathname)
	(let* ((tpns (tpn::parse-tpn-stream tpn-in-stream
								   'tpn::get-event 'tpn::make-episode
								   'tpn::add-episode 'tpn::make-tpn))
		   (tpn (first tpns))
		   (graph nil)
		   (activities nil))

	  ;; Generate a graph of the input TPN
	  ;;(tpn:tpn2dot tpn (cl-user::mars-pathname "Exec-Monitor/graphs/tpn.pdf"))

	  ;; Convert the TPN to an LDG
	  (multiple-value-setq (graph activities) (tpn-to-ldg tpn))

	  ;; Create a dispatchable LDG from this LDG
	  (let* ((dispatchable-graph (construct-dispatchable-graph graph))
			 (dispatch (make-instance 'online-ldg-dispatch :start 'a 
									  :ldg dispatchable-graph
									  :time-tol 0.1
									  :activities (activities-list activities))))
	  	;; Save the dispatcher into a global variable, so we can access
	 	;; the dispatcher's mailbox and force-stop it, if necessary.
	  	(setf *dispatch* dispatch)

		;; Dispatch it online
		(online-execute! dispatch :activity-callback activity-callback :execution-monitor execution-monitor)

		;;(format t "trace: ~a~%" (dispatch-trace dispatch))
		;;(format t "result: ~a~%" (dispatch-ldg dispatch))

		)))
  )


(defun test-dispatch-tpn-1 ()
  ;; Dispatch the TPN online
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/test-tburton/output_qcp_1.tpn")))
	(test-dispatch-tpn-online tpn-in-pathname :activity-callback #'test-callback)))



(defun test-callback (id primitive lb ub)
  "This callback is supposed to execute the primitive between lb and ub time.
   This implementation simulates a completed activity by spawning a thread
   which notifies the dispatcher that the primitive is complete by returning
   the primitive's unique id."
  ;; Format calls NOT made from the main thread may not print.
  ;; This behavior depends on the lisp runtime, and it's slime interface.
  (format-time t "Dispatching primitive ~s: ~s [~s,~s]~%" id primitive lb ub)
  (flet ((done () 
		   (sleep lb)
		   ;; Format calls NOT made from the main thread may not print.
		   ;; This behavior depends on the lisp runtime, and it's slime interface.
		   (format-time t "Sending completed message for primitive: ~s~%" primitive)
		   (notify-complete *dispatch* id)))
	(bordeaux-threads:make-thread #'done))
  )


