(in-package #:drake2)


(defun test-tpn-ldg-conversion ()
  (let* ((tpn (make-test-tpn))
		 (graph nil)
		 (activities nil))
	;; convert the tpn into Drake's ldg + activites format
	(multiple-value-setq (graph activities) 
	  (tpn-to-ldg tpn))
	;; dispatch it!
	
	(let* ((dispatchable-graph (construct-dispatchable-graph graph))
		   (dispatch (make-instance 'online-ldg-dispatch :start 'a 
									  :ldg dispatchable-graph
									  :time-tol 0.1
									  :activities (activities-list activities))))
	  ;; we need to save off the dispatch into a global variable, so the call back can use it.
	  (setf *dispatch* dispatch)
	  (online-execute! dispatch :activity-callback #'test-callback)	 
	  (format t "trace: ~a~%" (dispatch-trace dispatch))
	  (format t "result: ~a~%" (dispatch-ldg dispatch))
	  )
	))

(defun make-test-tpn ()
  (let* ((tpn (tpn::make-tpn "TEST-TPN" "EVENTA" "EVENTC" '()))
		 (eventa (tpn::get-event "EVENTA" tpn))
		 (eventb (tpn::get-event "EVENTB" tpn))
		 (eventc (tpn::get-event "EVENTC" tpn)))
	;(declare (ignore event1 event2 event3))
	;; create a function that will make it easier to add episodes for this tpn
	(flet ((make-add-episode (start end lb ub primitive)
			 (let ((description (list :episode))) ; Note: Use LIST, do NOT use the quote (')
			   (nconc description (list (tpn::make-association :primitive primitive)))
			   (tpn::add-episode start end (tpn::make-episode start end lb ub 0 description tpn :support :CONTROL-PROGRAM )))))
	  ;; add the episodes
	  (make-add-episode eventa eventb 1   2 "a->b")
	  (make-add-episode eventb eventc 1.2 2 "b->c")
	  (make-add-episode eventa eventc 1   3 "a->c")
	  )
	;; (format t "~s~%" (with-output-to-string (tpn-out-stream)
	;; 				   (tpn::write-tpn tpn-out-stream tpn :force-labeling? t)))
	tpn))


(defun test-large-ldg ()
  (let* ((tpn-path (cl-user::mars-pathname "Drake/tests/test1/tpn.xml"))
		 (pdf-path (cl-user::mars-pathname "Drake/tests/test1/tpn.pdf"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (tpn::parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))

	;; create a pdf of the tpn
	(tpn:tpn2dot tpn pdf-path)

	(multiple-value-bind (graph activities)
		(drake2:tpn-to-ldg tpn)
	  (declare (ignore activities))
	  (let ((dispatchable-graph nil))
		;; Create the dispatchable graph, or set it to NIL if the LDG is infeasible
		(setf dispatchable-graph 
			  (handler-case (drake2:construct-dispatchable-graph graph)
				(mtk-graph-algorithms::infeasible-problem () nil)))

		dispatchable-graph))
	))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; TPN --> LDG with choice
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(deftest (test-adjust-assignments :tpn-ldg-conversion) ()
  "Test the assignment checking after decision subtrees have been visited."
  (let* ((variables '(:|x| :|y| :|z|))
		 (domains (make-hash-table :test #'equalp)))
	;; Populate domains
	(setf (gethash :|x| domains) '(1 2 3))
	(setf (gethash :|y| domains) '(1 2 3))
	(setf (gethash :|z| domains) '(1 2))
	;; Run some tests
	(test '((= :|x| 1)) 
		  (adjust-assignments '((= :|x| 1)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1))")
	(test '((= :|x| 1) (= :|y| 2)) 
		  (adjust-assignments '((= :|x| 1) (= :|y| 2)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1) (= y 2))")
	(test '()
		  (adjust-assignments '((= :|x| 1) (= :|x| 2) (= :|x| 3)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1) (= x 2) (= x 3))")
	(test '((= :|y| 2))
		  (adjust-assignments '((= :|x| 1) (= :|y| 2) (= :|x| 2) (= :|x| 3)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1) (= y 2) (= x 2) (= x 3))")
	(test '()
		  (adjust-assignments '((= :|x| 1) (= :|z| 2) (= :|x| 2) (= :|x| 3) (= :|z| 1)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1) (= z 2) (= x 2) (= x 3) (= z 1))")
	(test -1
		  (adjust-assignments '((= :|x| 1) (= :|z| 2) (= :|x| 2) (= :|z| 1)) variables domains) 
		  :test #'equalp
		  :fail-info "adjust-assignments: Failed ((= x 1) (= z 2) (= x 2) (= z 1))")))