;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-graph-algorithms)

(defun generate-graph-from-figure-25.1 ()
  (let* ((e '((1 2)
	      (1 3)
	      (1 5)
	      (2 4)
	      (2 5)
	      (3 2)
	      (4 1)
	      (4 3)
	      (5 4)))
	 (w '(3 8 -4 1 7 4 2 -5 6))
	 (g (make-instance 'mtk-graph:directed-graph))
	 (weight-ht (make-hash-table)))
    (loop
       for elist_i in e
       and w_i in w
       do (let ((e_i (mtk-graph:add-edge elist_i g)))
	    (setf (gethash e_i weight-ht) w_i)))
    (values g #'(lambda (e_i) (gethash e_i weight-ht)))))

(defun generate-graph-from-figure-25.5 ()
  (let* ((e '((2 3)
	      (2 4)
	      (3 2)
	      (4 1)
	      (4 3)))
	 (g (make-instance 'mtk-graph:directed-graph)))
    (mtk-graph:add-edges e g)
    g))

(deftest (floyd-warshall-test-figure-25.4 :mtk-graph-all-pairs-shortest-path) ()
  (multiple-value-bind (g w) (generate-graph-from-figure-25.1)
    (let* ((d (floyd-warshall g :weight-function w)))
	  (test 0 (gethash '(1 1) d))
      (test 1 (gethash '(1 2) d))
      (test -3 (gethash '(1 3) d))
      (test 2 (gethash '(1 4) d))
      (test -4 (gethash '(1 5) d))
      (test 3 (gethash '(2 1) d))
	  (test 0 (gethash '(2 2) d))
      (test -4 (gethash '(2 3) d))
      (test 1 (gethash '(2 4) d))
      (test -1 (gethash '(2 5) d))
      (test 7 (gethash '(3 1) d))
      (test 4 (gethash '(3 2) d))
	  (test 0 (gethash '(3 3) d))
      (test 5 (gethash '(3 4) d))
      (test 3 (gethash '(3 5) d))
      (test 2 (gethash '(4 1) d))
      (test -1 (gethash '(4 2) d))
      (test -5 (gethash '(4 3) d))
	  (test 0 (gethash '(4 4) d))
      (test -2 (gethash '(4 5) d))
      (test 8 (gethash '(5 1) d))
      (test 5 (gethash '(5 2) d))
      (test 1 (gethash '(5 3) d))
      (test 6 (gethash '(5 4) d))
	  (test 0 (gethash '(5 5) d))))
  (let* ((g (mtk-graph:generate-weighted-graph-from-figure-25.1))
		 (d (floyd-warshall g)))
	(test 0 (gethash '(1 1) d))
	(test 1 (gethash '(1 2) d))
	(test -3 (gethash '(1 3) d))
	(test 2 (gethash '(1 4) d))
	(test -4 (gethash '(1 5) d))
	(test 3 (gethash '(2 1) d))
	(test 0 (gethash '(2 2) d))
	(test -4 (gethash '(2 3) d))
	(test 1 (gethash '(2 4) d))
	(test -1 (gethash '(2 5) d))
	(test 7 (gethash '(3 1) d))
	(test 4 (gethash '(3 2) d))
	(test 0 (gethash '(3 3) d))
	(test 5 (gethash '(3 4) d))
	(test 3 (gethash '(3 5) d))
	(test 2 (gethash '(4 1) d))
	(test -1 (gethash '(4 2) d))
	(test -5 (gethash '(4 3) d))
	(test 0 (gethash '(4 4) d))
	(test -2 (gethash '(4 5) d))
	(test 8 (gethash '(5 1) d))
	(test 5 (gethash '(5 2) d))
	(test 1 (gethash '(5 3) d))
	(test 6 (gethash '(5 4) d))
	(test 0 (gethash '(5 5) d))
    d))

(defun transitive-closure-test-figure-25.5 ()
  (let* ((g (generate-graph-from-figure-25.5))
	 (t-fun (transitive-closure g)))
    (assert (funcall t-fun 1 1))
    (assert (null (funcall t-fun 1 2)))
    (assert (null (funcall t-fun 1 3)))
    (assert (null (funcall t-fun 1 4)))
    (assert (funcall t-fun 2 1))
    (assert (funcall t-fun 2 2))
    (assert (funcall t-fun 2 3))
    (assert (funcall t-fun 2 4))
    (assert (funcall t-fun 3 1))
    (assert (funcall t-fun 3 2))
    (assert (funcall t-fun 3 3))
    (assert (funcall t-fun 3 4))
    (assert (funcall t-fun 4 1))
    (assert (funcall t-fun 4 2))
    (assert (funcall t-fun 4 3))
    (assert (funcall t-fun 4 4))
    t-fun))
    