;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package mtk-graph-algorithms)

;; For references relating to this code see:
;;   [1] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. 
;;       Introduction to Algorithms. The MIT Press,
;;       Cambridge, Massachusetts, 2nd edition, 2001.

;; From[1] pg. 532
;;
;; BFS(G,s) 
;; 1  for each vertex u \in V[G] - {s} 
;; 2    do color[u] <- WHITE 
;; 3       d[u] <- inf 
;; 4       \pi[u] <- NIL 
;; 5  color[s] <- GRAY 
;; 6  d[s] <- 0 
;; 7  \pi[s] <- NIL 
;; 8  Q <- null 
;; 9  ENQUEUE(Q,s) 
;; 10 while Q != null 
;; 11   do u <- DEQUEUE (Q) 
;; 12     for each v \in Adj[u] 
;; 13       do if color[v] = WHITE 
;; 14             then color[v] <- GRAY 
;; 15                  d[v] <- d[u] + 1 
;; 16                  \pi[v] <- u 
;; 17                  ENQUEUE(Q,v) 
;; 18     color[u] <- BLACK

;;; BFS(G,s) 
(defmethod bfs ((g mtk-graph:graph) s)
  (let ((color (make-hash-table))
	(d (make-hash-table))
	(p (make-hash-table))
	(q nil))
;;; 1  for each vertex u \in V[G] - {s} 
    (loop for u in (set-difference (mtk-graph:get-vertices g) (list s))
;;; 2    do color[u] <- WHITE 
       do (setf (gethash u color) 'white)
;;; 3       d[u] <- inf 
	 (setf (gethash u d) most-positive-fixnum)
;;; 4       \p[u] <- NIL 
	 (setf (gethash u p) nil))
;;; 5  color[s] <- GRAY 
    (setf (gethash s color) 'gray)
;;; 6  d[s] <- 0 
    (setf (gethash s d) 0)
;;; 7  \p[s] <- NIL 
    (setf (gethash s p) nil)
;;; 8  Q <- null 
    ;; See under let.
;;; 9  ENQUEUE(Q,s) 
    (setf q (append q (list s)))
;;; 10 while Q != null 
    (loop while (not (null q))
;;; 11   do u <- DEQUEUE (Q)
       do ;;(format t "~A~%" q)
	 (let ((u (pop q)))
;;; 12     for each v \in Adj[u] 
	    (loop for v in (mtk-graph:find-children-of-vertex u g)
;;; 13       do if color[v] = WHITE 
	       do (if (eq (gethash v color) 'white)
		      (progn
;;; 14             then color[v] <- GRAY 
			(setf (gethash v color) 'gray)
;;; 15                  d[v] <- d[u] + 1 
			(setf (gethash v d) (+ (gethash u d) 1))
;;; 16                  \p[v] <- u 
			(setf (gethash v p) u)
;;; 17                  ENQUEUE(Q,v) 
			(setf q (append q (list v))))))
;;; 18     color[u] <- BLACK
	    (setf (gethash u color) 'black)))))