(in-package :mtk-graph-algorithms)

(deftest (test-labeled-floyd-warshall :mtk-graph-algorithms) ()

  (let* ((var-descriptions
		  '((a :domain ((1) (2)))))
		 ;;create the state space
		 (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
		 (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph
							   :state-space state-space)))

	(mtk-graph:add-weighted-edge `((b a) ,(mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql '((1 ())))) graph)
	(mtk-graph:add-weighted-edge `((a c) ,(mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql '((3 ((= a 1))) (5 ())))) graph)
	(mtk-graph:add-weighted-edge `((c b) ,(mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql '((1 ((= a 1))) (2 ((= a 2)))))) graph)
	
	(let ((apsp 
		   (floyd-warshall 
			graph 
			:add-function #'(lambda (x y) (mtk-graph:lvs-binary-function x y #'+))
			:min-function #'(lambda (x y) (mtk-graph:merge-lvs x y))
			:zero (mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql `((0 ())))
			:inf (mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql `((,*inf* ()))))))
	  (maphash #'(lambda ( x y) (format t "~a ~a ~%" x y)) apsp)

	  )))

(deftest (test-bellman-ford :mtk-graph-algorithms) ()

  (let* ((graph (make-instance 'mtk-graph:weighted-directed-graph)))
	(mtk-graph:add-weighted-edges 
	 '(((s t) 6)
	   ((s y) 7)
	   ((t y) 8)
	   ((t z) -4)
	   ((t x) 5)
	   ((x t) -2)
	   ((y x) -3)
	   ((y z) 9)
	   ((z s) 2)
	   ((z x) 7))
	 graph)
	
	(let ((d-pi-ht (bellman-ford graph 's)))
	  (test '(7 s) (gethash 'y d-pi-ht) :test #'equal)
	  (test '(4 y) (gethash 'x d-pi-ht) :test #'equal)
	  (test '(2 x) (gethash 't d-pi-ht) :test #'equal)
	  (test '(-2 t) (gethash 'z d-pi-ht) :test #'equal))))