;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package mtk-graph-algorithms)

;; For references relating to this code see:
;;   [1] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. 
;;       Introduction to Algorithms. The MIT Press,
;;       Cambridge, Massachusetts, 2nd edition, 2001.

;;; From[1] pg. 550
;;;
;;; TOPOLOGICAL-SORT(G)
;;; 1 call DFS(G) to compute finishing times f[v] for each vertex v 
;;; 2 as each vertex is finished, insert it onto the front of a linked list 
;;; 3 return the linked list of vertices

;;; TOPOLOGICAL-SORT(G)
(defmethod topological-sort ((G mtk-graph:graph))
;;; 1 call DFS(G) to compute finishing times f[v] for each vertex v 
  (multiple-value-bind (p c d f) (dfs g)
    (declare (ignore p c d))
;;; 2 as each vertex is finished, insert it onto the front of a linked list 
    (sort (copy-seq (mtk-graph:get-vertices g))
	  #'(lambda (x y) (> (gethash x f) (gethash y f))))
;;; 3 return the linked list of vertices
    ))