;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-graph-graph2dot)

;; (defmethod dot-generate-weighted-edge-label ((e_i mtk-graph:edge)
;; 					     &rest rest)
;;   (let* ((weight-ht-edge (first rest))
;; 	 (weight (gethash e_i weight-ht-edge)))
;;     (format nil "~A"
;; 	    (cond ((equal weight excl:*infinity-double*) "inf")
;; 		  ((equal weight excl:*negative-infinity-double*) "-inf")
;; 		  (t weight)))))

(defmethod dot-print-weighted-graph
    ((g mtk-graph:graph)
     (w hash-table)
     &key
     (stream t)
     (vertex-labeler #'dot-generate-vertex-label)
     (edge-labeler #'dot-generate-weighted-edge-label))
  (dot-print-graph g
		   :stream stream
		   :vertex-labeler vertex-labeler
		   :edge-labeler edge-labeler
		   :rest w))

(defmethod dot-print-graph ((g mtk-graph:weighted-graph)
			    &key
			    (stream t)
			    (vertex-labeler #'dot-generate-vertex-label)
			    (edge-labeler #'dot-generate-weighted-edge-label))
  (call-next-method g
		    :stream stream
		    :vertex-labeler vertex-labeler
		    :edge-labeler edge-labeler
		    :rest g))

(defmethod dot-generate-weighted-edge-label ((e_i mtk-graph:edge)
					     &rest rest)
  (let* ((g (first rest))
	 (weight (mtk-graph:get-weight e_i g)))
    (format nil "~A" (infinity-to-string weight))))

(defmethod dot-print-graph ((g mtk-graph:graph)
			    &key
			    (stream t)
			    (vertex-labeler #'dot-generate-vertex-label)
			    (edge-labeler #'dot-generate-edge-label)
			    (rest nil))
  ;;(format t "Graph dot-print-graph: ~A.~%" edge-labeler)
  (let ((v (mtk-graph:get-vertices g))
	(e (mtk-graph:get-edges g)))
    (dot-print-graph-type e stream)
    (terpri stream)
    (princ "graph [];" stream)
    (terpri stream)
    (dot-print-vertices v
			:stream stream
			:vertex-labeler vertex-labeler
			:rest rest)
    (dot-print-edges e
		     :stream stream
		     :edge-labeler edge-labeler
		     :rest rest)
    (princ "}" stream)
    (terpri stream)))

(defmethod dot-print-graph-type ((e mtk-graph:edges) &optional (stream t))
  (princ "graph G {" stream))

(defmethod dot-print-graph-type ((e mtk-graph:directed-edges)
				 &optional (stream t))
  (princ "digraph G {" stream))

(defmethod dot-print-vertices ((v list)
			       &key
			       (stream t)
			       (vertex-labeler #'dot-generate-vertex-label)
			       (rest nil))
  (dolist (v_i v)
    (let ((v_i-name (generate-vertex-name v_i)))
      (format stream
	      "\"~0A\" [\"~0A\"];~%"
	      v_i-name (funcall vertex-labeler v_i rest)))))

(defmethod dot-generate-vertex-label (v_i &rest rest)
  ;;"")
  (generate-vertex-name v_i))

(defmethod generate-vertex-name (v_i)
   (if (symbolp v_i) (symbol-name v_i) v_i))
  

(defmethod dot-print-edges ((e mtk-graph:edges)
			    &key
			    (stream t)
			    (edge-labeler #'generate-edge-label)
			    (rest nil))
  (dolist (e_i (mtk-graph:get-edge-list e))
    (let* ((v (mtk-graph:get-vertices e_i))
	   (e_1 (first v))
	   (e_2 (second v)))
      (format stream "\"~0A\"~0A\"~0A\" [label=\"~0A\",];~%"
	      e_1 
	      (generate-dot-edge-string e_i)
	      e_2
	      (funcall edge-labeler e_i rest)))))

(defmethod generate-dot-edge-string ((e mtk-graph:edge))
  "--")
(defmethod generate-dot-edge-string ((e mtk-graph:directed-edge))
  "->")

(defmethod dot-generate-edge-label ((e mtk-graph:edge) &rest rest)
  "")

(defvar *dot-path*
  #+(or win32 mswindows) "dot.exe"
  #-(or win32 mswindows) "dot"
  "Path to `dot`")

;; the path to the neato executable (used for drawing undirected
;; graphs). 
(defvar *neato-path*
  #+(or win32 mswindows) "neato.exe"
  #-(or win32 mswindows) "neato"
  "Path to `neato`")

(defmethod choose-exec ((g mtk-graph:hyper-graph))
  *neato-path*)

(defmethod choose-exec ((g mtk-graph:directed-hyper-graph))
  *dot-path*)


(defun dot-graph (graph outfile
		  &key 
		  (format "ps2")
		  (size "7.5,10")
		  (orientation "landscape")
		  (rankdir "LR")
		  (dot-path nil))
  "Renders GRAPH to OUTFILE by running the program in \*DOT-PATH* or
*NEATO-PATH* depending on the value of the DIRECTED keyword
argument.  The default is a directed graph.  The default
FORMAT is Postscript."
  (if (null dot-path) (setf dot-path (choose-exec graph)))
  (format t
		  "~0A -T~0A -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=~0A -o ~0A"
		  dot-path format size orientation rankdir outfile)
  #+sbcl
  (let ((dot-string (with-output-to-string (stream)
		      (dot-print-graph graph :stream stream))))
    (sb-ext:run-program 
     dot-path
     (list (format nil
		   "-T~0A -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=~0A"
		   format size orientation rankdir)
	   "-o" outfile)
     :input (make-string-input-stream dot-string)
     :output *standard-output*))
  #+allegro
  (excl.osi:with-command-io
      ((format nil
	       "~0A -T~0A -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=~0A -o ~0A"
	       dot-path format size orientation rankdir outfile))
    (:input (dot-stream)
	    (dot-print-graph graph :stream dot-stream)))
  #+lispworks
  (with-open-stream
      (dot-stream 
       (sys:open-pipe 
	(format nil 
		"~0A -T~0A -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=~0A -o ~0A"
		dot-path format size orientation rankdir outfile)
	:direction :input))
    (dot-print-graph graph :stream dot-stream)
    (force-output dot-stream))
  #+clisp
  (with-open-stream 
      (out 
       (ext:make-pipe-output-stream
	(format nil 
		"~0A -T~0A -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=~0A -o ~0A"
		dot-path format size orientation rankdir outfile)))
    (dot-print-graph graph :stream out))
  #-(or sbcl lispworks allegro clisp)
  (error "Don't know how to execute a program on this platform"))