;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung
;;;;   David C. Wang

(in-package :mtk-graph)


;;-------------------------------------------------------------------------------
;; Edge CLASS:
;;   * Directed Edge
;;-------------------------------------------------------------------------------

(defclass directed-edge (directed-hyper-edge edge) ())

;; CONSTRUCTOR
(defmethod make-directed-edge (&key source target)
  (assert source)
  (assert target)
  (assert (not (eql source target)))
  (let ((vertices (list source target)))
    (make-instance 'directed-edge
      :vertices vertices
      :source-vertices (list source)
      :target-vertices (list target))))

;; GET SOURCE VERTEX
(defmethod get-source-vertex ((e_i directed-edge))
  (first (get-source-vertices e_i)))

;; GET TARGET VERTEX
(defmethod get-target-vertex ((e_i directed-edge))
  (first (get-target-vertices e_i)))


;;-------------------------------------------------------------------------------
;; Edges CLASS:
;;   * Directed Edges
;;-------------------------------------------------------------------------------

(defclass directed-edges (directed-hyper-edges edges) ())

;;----------------------------------------
;; Add Edges

(defmethod add-edge ((e_i directed-edge) (e directed-edges))
  (call-next-method e_i e))

(defmethod add-edge ((edge-description list) (e directed-edges))
  "Adds an edge to a set of directed edges, where the edge-description
is a sequence of source and target vertices the edge connects."
  (let* ((source-vertices (list (first edge-description)))
	 (target-vertices (list (second edge-description)))
	 (vertices (append source-vertices target-vertices))
	 (e_i (make-instance 'directed-edge
			     :vertices vertices
			     :source-vertices source-vertices
			     :target-vertices target-vertices)))
    (add-edge e_i e)))

;;----------------------------------------
;; Delete Edges

(defmethod delete-edge ((e_i directed-edge) (e directed-edges))
  (call-next-method e_i e))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex-pair ((vertices list) (e directed-edges))
  (gethash vertices (get-edges-ht-vertices e)))


;;-------------------------------------------------------------------------------
;; Graph CLASS:
;;   * Directed Graph
;;-------------------------------------------------------------------------------

(defclass directed-graph (directed-hyper-graph graph)
  ((e :type directed-edges
      :initform (make-instance 'directed-edges))))

;; TRANSPOSE
(defmethod transpose ((g directed-graph))
  (let ((g^t (make-instance (class-of g))))
    (add-vertices (get-vertices g) g^t)
    (dolist (e_i (get-edge-list (get-edges g)))
      (add-edge (list (get-target-vertex e_i) (get-source-vertex e_i)) g^t))
    g^t))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex-pair ((vertices list) (g directed-graph))
  (find-edges-with-vertices vertices (get-edges g)))
