;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-graph)


;;-------------------------------------------------------------------------------
;; Edge CLASS:
;;   * Edge
;;-------------------------------------------------------------------------------

(defclass edge (hyper-edge) ())

;; CONSTRUCTOR
(defmethod make-edge (&key vertices)
  (assert (= (length (remove-duplicates vertices :test #'eql)) 2))
  (make-instance 'edge :vertices vertices))


;;-------------------------------------------------------------------------------
;; Edges CLASS:
;;   * Edges
;;-------------------------------------------------------------------------------

(defclass edges (hyper-edges)
  ((edges-ht-vertices :type hash-table
                      :accessor get-edges-ht-vertices
                      :initform (make-hash-table :test #'equal)
                      :documentation "")))

;;----------------------------------------
;; Add Edges

(defmethod add-edge ((e_i edge) (e edges))
  (call-next-method e_i e)
  (push e_i (gethash (get-vertices e_i) (get-edges-ht-vertices e)))
  e_i)

(defmethod add-edge ((edge-description list) (e edges))
  "The edge-description is a list of two vertices that the edge connects."
  (let ((e_i (make-instance 'edge :vertices edge-description)))
    (add-edge e_i e)))

;;----------------------------------------
;; Delete Edges

(defmethod delete-edge ((e_i edge) (e edges))
  (call-next-method e_i e)
  (setf (gethash (get-vertices e_i) (get-edges-ht-vertices e))
	(delete e_i (gethash (get-vertices e_i) (get-edges-ht-vertices e)))))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex-pair ((vertices list) (e edges))
  (or (gethash vertices (get-edges-ht-vertices e))
      (gethash (reverse vertices) (get-edges-ht-vertices e))))


;;-------------------------------------------------------------------------------
;; Graph CLASS:
;;   * Graph (undirected)
;;-------------------------------------------------------------------------------

(defclass graph (hyper-graph)
  ((e :type edges
      :initform (make-instance 'edges))))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex-pair ((vertices list) (g graph))
  (find-edges-with-vertex-pair vertices (get-edges g)))
