(in-package :mtk-graph)


(defclass labeled-directed-edge (directed-edge)
  ((environment 
	  :type state-clause
	  :accessor get-edge-env
	  :initarg :env
	  :initform (error "Must specify an environment using :env key.")
	  :documentation "The environment under which the edge holds.")))

(defmethod print-object ((e labeled-directed-edge) stream)
  (with-slots (vertices environment) e
	  (format stream "~a ~a" vertices environment)))

(defclass labeled-directed-edges (directed-hyper-edges edges) ())

(defclass labeled-directed-graph (directed-hyper-graph graph)
  ((state-space
	:type state-space
	:accessor get-state-space
	:initform (error "Must provide a state-space.")
	:documentation "The state-space of the graph."
	:initarg :state-space)

   (e :type labeled-directed-edges
	  :initform (make-instance 'directed-hyper-edges))))

(defmethod add-edge ((e_i labeled-directed-edge) (e labeled-directed-edges))
  (call-next-method e_i e))

(defmethod add-labeled-edge ((edge-description list) 
							 (g labeled-directed-graph)
							 (env state-clause))
 
  (let* ((source-vertices (list (first edge-description)))
		 (target-vertices (list (second edge-description)))
		 (vertices (append source-vertices target-vertices))
		 		 
		 (e_i (make-instance 'labeled-directed-edge
							 :vertices vertices
							 :source-vertices source-vertices
							 :target-vertices target-vertices
							 :env env)))
	(add-vertices vertices g)
    (add-edge e_i (get-edges g))))



;; (defmethod find-strongly-connected-components
;; 	((g labeled-directed-graph))
;;   "Identify only those SCC that the compiling algorithm cares about!"

;;   (let ((rigid-component-ht (make-hash-table)))

;; 	(dolist (start-vertex (get-vertices g))
;; 	  (debug-print-for :mtk-graph t "Beginning SCC search ~a" start-vertex)
;; 	  (search-for-start-vertex g start-vertex start-vertex
;; 							 (get-universal-clause (get-state-space g))
;; 							 nil
;; 							 nil
;; 							 rigid-component-ht))
;; 	rigid-component-ht))


;; (defmethod search-for-start-vertex ((g labeled-directed-graph)
;; 									start-vertex
;; 									curr-vertex
;; 									curr-env
;; 									curr-path
;; 									most-recent-edge
;; 									rigid-component-ht)

;;     (debug-print-for :mtk-graph t "At ~a with ~a, path: ~a " curr-vertex curr-env curr-path)
;;   (let ((outgoing-paths (find-edges-with-source-vertex curr-vertex g)))
;; 	(dolist (outgoing-edge outgoing-paths)
;; 	  (debug-print-for :mtk-graph t "Considering path ~a" outgoing-edge)
;; 	  (let* ((next-env (clause-intersection
;; 						curr-env
;; 						(get-edge-env outgoing-edge)))
;; 			 (outgoing-target (get-target-vertex outgoing-edge)))
;; 		(unless (null next-env)

;; 		;;when taking the first step, you can follow every path in front of you
;; 		  (when (or (eql start-vertex curr-vertex)
;; 					;;either you've found a loop back to the start or it 
;; 					;;only loops back on itself once, allows us to find
;; 					;;long chains
;; 					(and (or (eql outgoing-target start-vertex)
;; 							 (>= 1 (apply #'+ (mapcar #'(lambda (x)
;; 														  (if (eql x outgoing-target)
;; 															  1
;; 															  0))
;; 													  curr-path))) 
;; 							 ;;(not (member outgoing-target curr-path))
;; 							 )

					   
;; 						 ;;and the environment is more specific
;; 										;					(clause-subsumed? next-env
;; 										;									  curr-env)

;; 						 ;;and it's not subsumed by another incoming path
;; 						 ;;that isn't equivalent to the route we took
;; 						 ;; (notany #'(lambda (alt-edge)
;; 						 ;; 			 (and (clause-subsumed? (get-edge-env alt-edge)
;; 						 ;; 									(get-edge-env most-recent-edge))
;; 						 ;; 				  (clause-subsumed? (get-edge-env outgoing-edge)
;; 						 ;; 									(get-edge-env alt-edge))
;; 						 ;; 				  (not (eql (get-edge-env most-recent-edge)
;; 						 ;; 							(get-edge-env alt-edge)))))
;; 						 ;; 		 (find-edges-with-target-vertex curr-vertex g))
;; ))
		  
;; 										;(and 

;; 										;			   (or (null curr-env)
;; 										;				   (clause-subsumed? next-env curr-env))

;; 										;			   (notany #'(lambda (alt-edge) 
;; 										;						   (and (clause-subsumed? next-env (get-edge-env alt-edge))
;; 										;								(not (eql next-env (get-edge-env alt-edge)))))
;; 										;					   (find-edges-with-target-vertex curr-vertex g)))

;; 			(if (eql start-vertex outgoing-target)
;; 				(progn
;; 				  (debug-print-for :mtk-graph t "Loops to start")
;; 				  (add-rigid-component! rigid-component-ht (remove-duplicates
;; 															(append (list curr-vertex)
;; 																	curr-path)) 
;; 										next-env))
;; 										;unless (member outgoing-target curr-path)
;; 				(search-for-start-vertex
;; 				 g
;; 				 start-vertex
;; 				 outgoing-target
;; 				 next-env
;; 				 (append (list curr-vertex) curr-path)
;; 				 outgoing-edge
;; 				 rigid-component-ht))))))
;; 	(debug-print-for :mtk-graph t "backtrack!")
;; )

  

;; )


(defun add-rigid-component! (rigid-component-ht
							 new-component
							 env)

  ;;check that a subsuming, larger component doesn't exist before you try to add
  (loop for component-env being the hash-keys of rigid-component-ht
	 if (and (clause-subsumed? env component-env)
			 (member new-component (gethash component-env rigid-component-ht)
					 :test #'subsetp))
	   do (return-from add-rigid-component!))


 ;;if there's an existing one that's a subset, remove it
  (setf (gethash env rigid-component-ht)
		(remove-if #'(lambda (x) (subsetp x new-component)) 
				   (gethash env rigid-component-ht)
				))

  ;;remove old components if they're subsets
  (setf (gethash env rigid-component-ht)
		(remove-if #'(lambda (old-component) (subsetp old-component new-component)) 
				   (gethash env rigid-component-ht)))
  ;;Add the new component iff it is different than the other rigid components known
  (pushnew  new-component (gethash env rigid-component-ht)
			:test #'(lambda (x y) 
					  (or (subsetp x y) 
						  (null (set-exclusive-or x y))))))