(in-package :mtk-graph)

(defclass labeled-weighted-directed-graph (weighted-directed-graph)
  ((state-space :type state-space
				:initarg :state-space
				:accessor get-state-space
				:documentation "Holds the state-space associated with the graph.")
   (zero-components
	:initform nil
	:accessor get-zero-components
	:documentation "A list of zero-groups, (group env), where every event in the group is zero-bound together under env.")))

(defmethod print-object ((ldg labeled-weighted-directed-graph) stream)
  (call-next-method ldg stream)
  (format stream "Zero groups: ~a~%" (get-zero-components ldg)))

(defmethod update-labeled-weighted-edges ((labeled-weighted-edge-descriptions list)
										  (g labeled-weighted-directed-graph)
										  &key (strict-order #'<)
										  (value-equalp #'eql))
"Add edges by a list of descriptions."

(dolist (edge-description labeled-weighted-edge-descriptions)
  (update-labeled-weighted-edge edge-description g 
								:strict-order strict-order
								:value-equalp value-equalp)))


(defmethod update-labeled-weighted-edge ((labeled-weighted-edge-description list)
										 (g labeled-weighted-directed-graph)
										 &key (strict-order #'<)
										 (value-equalp #'eql))
"Add an edge. Description is: ((start end) weight env)"

(debug-print-for :mtk-graph t "Update weight ~a" labeled-weighted-edge-description)
  (assert (= (length labeled-weighted-edge-description) 3))
   (let* ((edge-description (first labeled-weighted-edge-description))
		  (new-weight (second labeled-weighted-edge-description))
		  (env (third labeled-weighted-edge-description))
		  ;;try to find this edge
		  (e_i (first (find-edges-with-vertex-pair edge-description 
												   (get-edges g))) ))

	 ;;forbid self-loops for compilation
	 (unless (eql (first edge-description)
				  (second edge-description))

	 ;;if it's not there, make it
	   (when (null e_i)
		 (setf e_i (add-edge edge-description g))

	   
		 ;;and initialize the weight
		 (setf (gethash e_i (get-weight-ht-edge g))
			   (make-instance 'strict-labeled-value-set 
							  :value-equalp value-equalp
							  :strict-order strict-order
							  :state-space (get-state-space g))))
	 ;;then add the new pair to the lvs
	   (add-pair-to-lvs! (gethash e_i (get-weight-ht-edge g))
						 (make-instance 'labeled-pair
										:environment env
										:value new-weight))
	   (debug-print-for :mtk-graph t "Updated: ~a" (gethash e_i (get-weight-ht-edge g))))))