(in-package :mtk-graph)

(deftest (test-make-lvs-less-strict :mtk-graph) ()
  (let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
		 (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
		 (lvs (make-instance 'strict-labeled-value-set :strict-order #'< 
							 :state-space state-space
							 :value-equalp #'eql))

		 (lvs-from-description (make-labeled-value-set-from-description
								state-space 'strict-labeled-value-set
								#'<  #'eql '((2 ()) (3 ((= a 1))) (1 ((= a 2)))
											 (0 ((= a 2))) (3 ((= a 2))) 
											 (4 ((= a 2) (= b 1)))  
											 (-1 ((= a 2) (= b 2)))  ))))


	;;Add some clauses and make sure the right ones stick
	;;first one, keep
	(add-pair-to-lvs-by-description! lvs '(2 ()))
	(add-pair-to-lvs-by-description! lvs '(2 ()))
	(add-pair-to-lvs-by-description! lvs '(3 ((= a 1))))
	;;(format t "~a~%" lvs)
	(test 1 (lvs-length lvs))

	;;1 <= 2, keep
	(add-pair-to-lvs-by-description! lvs '(1 ((= a 2))))
	;;(format t "~a~%" lvs)
	(test 2 (lvs-length lvs))

	;;keep and replace the old one
	(add-pair-to-lvs-by-description! lvs '(0 ((= a 2))))
	;;(format t "~a ~a ~a ~%" lvs (lvs-length lvs) (query-lvs lvs (get-clause-from-descriptions state-space '((= a 2)))))
	(test 2 (lvs-length lvs))
	(test 0 (first (query-lvs lvs (get-clause-from-descriptions state-space '((= a 2))))))

	;;do not keep
	(add-pair-to-lvs-by-description! lvs '(3 ((= a 2))))
	;;(format t "~a~%" lvs)
	(test 2 (lvs-length lvs))

	;;do not keep
	(add-pair-to-lvs-by-description! lvs '(4 ((= a 2) (= b 1))))
	;;(format t "~a~%" lvs)
	(test 2 (lvs-length lvs))

	;;should keep
	(add-pair-to-lvs-by-description! lvs '(-1 ((= a 2) (= b 2))))
	;;  (format t "~a~%" lvs)
	(test 3 (lvs-length lvs))

	(test -1 (first (query-lvs-by-description lvs '((= a 2) (= b 2)))))

	(test -1 (first (query-lvs-by-description lvs-from-description '((= a 2) (= b 2)))))
	(test 3 (lvs-length lvs-from-description))
	))

(deftest (test-make-lvs-greater :mtk-graph) ()
  (let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
		 (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
		 (lvs (make-instance 'strict-labeled-value-set :strict-order #'> 
							 :state-space state-space
							 :value-equalp #'eql)))

	;;Add some clauses and make sure the right ones stick
	;;first one, keep
	(add-pair-to-lvs-by-description! lvs '(2 ()))
	(add-pair-to-lvs-by-description! lvs '(1 ((= a 1))))
	(test 1 (lvs-length lvs))

	;;1 <= 2, keep
	(add-pair-to-lvs-by-description! lvs '(3 ((= a 2))))
	(test 2 (lvs-length lvs))

	;;do not keep
	(add-pair-to-lvs-by-description! lvs '(0 ((= a 2))))
	(test 2 (lvs-length lvs))
  
	;;do not keep
	(add-pair-to-lvs-by-description! lvs '(-1 ((= a 2) (= b 1))))
	(test 2 (lvs-length lvs))
  
	;;should keep
	(add-pair-to-lvs-by-description! lvs '(4 ((= a 2) (= b 2))))
	(test 3 (lvs-length lvs))))

(deftest (test-lvs-binary-ops :mtk-graph) ()
(let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
	   (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
	   (lvs1 (make-instance 'strict-labeled-value-set :strict-order #'< 
						   :state-space state-space
						   :value-equalp #'eql))
	   (lvs2 (make-instance 'strict-labeled-value-set :strict-order #'< 
						   :state-space state-space
						   :value-equalp #'eql))
	   )

;;Add some clauses
  (add-pair-to-lvs-by-description! lvs1 '(7 ()))
  (add-pair-to-lvs-by-description! lvs1 '(5 ((= a 1))))
  (add-pair-to-lvs-by-description! lvs1 '(4 ((= b 1))))

  (add-pair-to-lvs-by-description! lvs2 '(5 ()))
  (add-pair-to-lvs-by-description! lvs2 '(2 ((= b 2))))
  (add-pair-to-lvs-by-description! lvs2 '(1 ((= c 2))))

  (let ((lvs-sum (lvs-binary-function lvs1 lvs2 #'+))
		(lvs-min (lvs-binary-function lvs1 lvs2 #'min)))


	(test 5 (first (query-lvs lvs-sum (get-clause-from-descriptions state-space '((= b 1) (= c 2))))))
	(test 8 (first (query-lvs lvs-sum (get-clause-from-descriptions state-space '((= b 2) (= c 2))))))
	(test 12 (first (query-lvs lvs-sum (get-clause-from-descriptions state-space '()))))

	(test 5 (first (query-lvs lvs-min (get-clause-from-descriptions state-space '()))))
	(test 2 (first (query-lvs lvs-min (get-clause-from-descriptions state-space '((= b 2) (= c 1))))))
	(test 1 (first (query-lvs lvs-min (get-clause-from-descriptions state-space '((= b 2) (= c 2))))))

)))




(deftest (test-make-lvs-less-weak :mtk-graph) ()
(let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
	   (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
	   (lvs (make-instance 'weak-labeled-value-set 
						   :strict-order #'(lambda (x y) (< (first x) (first y)))
						   :state-space state-space
						   :value-equalp #'equalp)))

;;Add some clauses and make sure the right ones stick
  ;;first one, keep
  (add-pair-to-lvs-by-description! lvs '((2 e) ()))
  (add-pair-to-lvs-by-description! lvs '((2 e) ()))
  (add-pair-to-lvs-by-description! lvs '((2 f) ()))
  (add-pair-to-lvs-by-description! lvs '((3 i) ()))
;;  (format t "~a~%" lvs)

  (add-pair-to-lvs-by-description! lvs '((2 k) ((= a 1))))
  ;;(format t "~a ~a " lvs (query-lvs lvs (get-clause-from-descriptions state-space '()) ))
  (test 3 (lvs-length lvs))
  
  (test 3 (length (query-lvs lvs (get-clause-from-descriptions state-space '((= a 1))))))
  (test 2 (first (first (query-lvs lvs (get-clause-from-descriptions state-space '((= a 1)))))))

  ;;1 <= 2, keep
  (add-pair-to-lvs-by-description! lvs '((0 g) ((= a 2))))
  ;;(format t "~a~%" lvs)
  (test 4 (lvs-length lvs))

  ;;replace all the universal ones
  (add-pair-to-lvs-by-description! lvs '((1 h) ()))
  ;;(format t "~a~%" lvs)
  (test 2 (lvs-length lvs))

  (add-pair-to-lvs-by-description! lvs '((0 g) ()))
  ;;(format t "~a~%" lvs)
  (test 1 (lvs-length lvs)))
)