;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(defpackage mtk-graph
  (:nicknames :mtk-graph)
  (:use :cl :mtk-utils :mtk-utils-infinity :state-space :abstract-data-types)
  (:export
   hyper-graph
   graph
   directed-hyper-graph
   directed-graph
   undirected-graph
   hyper-edges
   edges
   directed-hyper-edges
   directed-edges
   hyper-edge
   edge
   directed-hyper-edge
   directed-edge

   make-hyper-edge
   make-directed-hyper-edge
   make-edge
   make-directed-edge
   
   get-vertices
   get-edges
   get-edges-with-vertex
   get-edge-list
   get-source-vertices
   get-source-vertex
   get-target-vertices
   get-target-vertex
   get-vertices
   
   add-vertex
   add-vertices
   delete-vertex
   delete-vertices
   add-edge
   add-edges
   delete-edge
   delete-edges
   
   copy-graph
   transpose
   
   find-edges-with-vertex
   find-edges-with-vertices
   find-edges-with-source-vertex
   find-edges-with-target-vertex
   find-edges-with-source-and-target-vertices
   find-edges-with-vertex-pair
   find-vertices-adjacent-to-vertex
   find-neighbors-of-vertex
   find-children-of-vertex
   find-parents-of-vertex

   graph-test
   directed-graph-test

   weighted-hyper-graph
   weighted-directed-hyper-graph
   weighted-graph
   weighted-directed-graph
   edge-weights
   
   get-weight
   set-weight
   add-weighted-edge
   add-weighted-edges
   delete-weighted-edge
   delete-weighted-edges

   generate-weighted-graph-from-figure-25.1

   strict-labeled-value-set
   weak-labeled-value-set
   labeled-pair
   query-lvs
   query-lvs-by-description
   make-labeled-value-set-from-description
   lvs-binary-function
   merge-lvs
   merge-lvs-in-place!
   pair-value
   pair-environment
   add-pair-to-lvs!
   lvs-pairs
   lvs-binary-function
   lvs-binary-function-candidates
   lvs-environment
   lvs-value
   remove-invalid-pairs!
   copy-lvs
   update-lvs-with-binary-operation!
   prune-lvs-of-invalid-environments!

   get-weight-ht-edge
   zero-components 
   labeled-weighted-directed-graph
   get-state-space
   update-labeled-weighted-edges
   update-labeled-weighted-edge

   labeled-directed-graph
   add-labeled-edge
   find-strongly-connected-components
   remove-subsumed-pairs!
   get-zero-components
   apply-stable-unary-operator! 
   clear-dependence-on-env!
   ))


(mtk-utils::deftest-suite :mtk-graph)
(mtk-utils::declare-debug-print-system! :mtk-graph)

