(in-package :ITC)

;;; Test of Constraint System Explanation for ITC

;;; (added by Peng):
;;; Test41 to 49 are the basic tests for modified constraint checking.
;;; The REMOVED? flag is set to NIL for all event successors. 
;;; This is so that every edge will be considered during propagation.
;;;; Their edges are relaxed to -1000000 to 1000000 (don't know why *infinity* doesn't work here).



(deftest (test40-1 :itc) ()

  ;; Test the relaxation def functions.
  
  (let ((stn (make-instance 'ITC-STN :name  'test40-1	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test40-1
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test40-1
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test40-1
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (cd (insert-new-constraint stn :source 'test40-1
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  
	  (let ((relaxationA (make-instance 'stc-relaxation :constraint ca :relaxing-ub? t :relaxed-ub 5))
			(relaxationB (make-instance 'stc-relaxation :constraint cb :relaxing-lb? t :relaxed-lb 0.5))
			(relaxationC (make-instance 'stc-relaxation :constraint cc :relaxing-lb? t :relaxed-lb 0))
			(relaxationD (make-instance 'stc-relaxation :constraint cd :relaxing-ub? t :relaxed-ub 100)))

		(test nil (consistent-stn? stn :initialize? t))

		(apply-relaxation relaxationA)
		(test nil (consistent-stn? stn :initialize? t))

		(undo-relaxation relaxationA)
		(test nil (consistent-stn? stn :initialize? t))

		(apply-relaxation relaxationB)
		(test t (consistent-stn? stn :initialize? t))

		(undo-relaxation relaxationB)
		(test nil (consistent-stn? stn :initialize? t))

		(apply-relaxation relaxationC)
		(test t (consistent-stn? stn :initialize? t))

		(undo-relaxation relaxationC)
		(test nil (consistent-stn? stn :initialize? t))

		(apply-relaxation relaxationD)
		(test t (consistent-stn? stn :initialize? t))

		(undo-relaxation relaxationD)
		(test nil (consistent-stn? stn :initialize? t)))

	  )))


(deftest (test40-2 :itc) ()

  ;; Test the ITC d-edge conflict extracting function.
  
  (let ((stn (make-instance 'ITC-STN :name 'test40-2	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test40-2
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test40-2
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test40-2
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (cd (insert-new-constraint stn :source 'test40-2
									 :start 'e1 :end 'e3 :lb 4 :ub 15))
		  (ce (insert-new-constraint stn :source 'test40-2
									 :start 'e3 :end 'e4 :lb 4 :ub 5))
		  (cf (insert-new-constraint stn :source 'test40-2
									 :start 'e2 :end 'e4 :lb 4 :ub 5)))

	  (multiple-value-bind (consistent? conflicts) (consistent-stn? stn :initialize? t :return-negative-cycle? t)

		(test
		 `(nil ("E3->E2[-4](lb)" "E2->E4[5]" "E4->E3[-4](lb)"))
		 (list nil (second conflicts))		 
		 :multiple-values nil
		 :test #'equal-consistent?-conflict-result
		 :fail-info 
		 "Test 40-2: minimal conflicts is (E3->E2[-4](lb) E2->E4[5] E4->E3[-4](lb))")))))


(deftest (test41 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test41	:start-name 'START)))

	(insert-new-constraint stn :source 'test41
						   :start 'start :end 'e1 :lb 0 :ub 0)
	(insert-new-constraint stn :source 'test41
						   :start 'e1 :end 'e2 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test41
						   :start 'e2 :end 'e3 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test41
						   :start 'e1 :end 'e3 :lb 4 :ub 5)
	
	(stn-initialize! stn)

	(test
	 `(nil 
	   ((ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
		(ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
		(ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)))

	 (consistent-stn? stn)
	 
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info 
	 "Test 41: ((E1,E3) for [4,5], (E2,E3) for [4,5], ~
           (E1,E2) for [4,5]) -> Inconsistent (conflict contains all STCs except (Start,E1)).")))


(deftest (test42 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test42	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test42
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test42
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test42
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test42
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  
	  (let ((constraints (list ca cb cc c2)))
		  
		  (test
		   `((
			  (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
              (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
			  ))

		   (all-conflicts-of-constraints stn constraints	:method :min-unsat)
   
		   :multiple-values nil
		   :test #'equal-conflicts
		   :fail-info "Test 42: Min-Unsat Conflicts = {(E1,E2)[4,5], (E2,E3)[4,5], (E1,E3)[4,5])}.")))))

(deftest (test43 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test43	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test43
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test43
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test43
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test43
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
		  
	  (let ((constraints (list ca cb cc c2)))	  
	  
		  (test
		   `(  
             (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			 (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			 (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))

		   (conflict-of-constraints stn constraints
									:method :rePlayXplain)
   
		   :multiple-values nil
		   :test #'equal-conflict
		   :fail-info "Test 43: rePlayXplain Conflicts = {(E1,E2)[4,5], (E2,E3)[4,5], (E1,E3)[4,5])}.")))))



(deftest (test44 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test44	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test44
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test44
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test44
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test44
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  (let ((constraints (list ca cb cc c2)))

		(test
		 `(
		   (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
		   (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
		   (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))

		 (Preferred-conflict-of-constraints stn
											constraints
											:method :PreferredConflict)
		 
		 :multiple-values nil
		 :test #'equal-conflict
		 :fail-info "Test 44: PreferredConflict Conflicts = {(E1,E2)[4,5], (E2,E3)[4,5], (E1,E3)[4,5])}.")))))
	

(deftest (test45 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test45	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test45
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test45
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test45
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test45
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  (let ((constraints (list ca cb cc c2)))
		
		(test
		 `(
		   (ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*)
		   (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
		   (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*))

		 (preferred-relaxation-of-constraints stn constraints
											  :method :PreferredRelax)
		 
		 :multiple-values nil
		 :test #'equal-relaxation
		 :fail-info "Test 45: PreferredRelax relaxation = (START,E1)[0,0], (E1,E2)[4,5], (E2,E3)[4,5])")))))
	  

(deftest (test46 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test46	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test46
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test46
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test46
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test46
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  (let ((constraints (list ca cb cc c2)))

		(test
		 `((ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
		   (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
		   (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))

		 (preferred-conflict-of-constraints stn constraints
											:method :LexQuicXplain)
		 
		 :multiple-values nil
		 :test #'equal-conflict
		 :fail-info "Test 46: PreferredConflict conflict = ~
                       {(E1,E2)[4,5], (E2,E3)[4,5],  (E1,E3)[4,5])}.")))))		  


(deftest (test47 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test47	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test47
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test47
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test47
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test47
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  (let ((constraints (list ca cb cc c2)))

		(test
		 `((
			(
			 ((ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*))
			 
			 ((ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))
			 )
			
			(
			 ((ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*)
			  (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)) 
			 
			 ((ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*))
			 )
			
			(
			 ((ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*)
			  (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)) 
			 
			 ((ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*))
			 )
			)

		   (  
			(
			 (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			 (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			 (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))
			)
		   
		   )
		 
		 (CS::all-explanations-and-conflicts-of-constraints stn constraints :method :daa-min-unsat)
		 
		 :multiple-values T
		 :test #'equal-explanations-and-conflicts
		 :fail-info "Test 47: daa-min-unsat explanations and conflicts")))))


(deftest (test48 :itc) ()
  
  (let ((stn (make-instance 'ITC-STN :name  'test48	:start-name 'START)))

	(let ((ca (insert-new-constraint stn :source 'test48
									 :start 'start :end 'e1 :lb 0 :ub 0))
		  (cb (insert-new-constraint stn :source 'test48
									 :start 'e1 :end 'e2 :lb 4 :ub 5))
		  (cc (insert-new-constraint stn :source 'test48
									 :start 'e2 :end 'e3 :lb 4 :ub 5))
		  (c2 (insert-new-constraint stn :source 'test48
									 :start 'e1 :end 'e3 :lb 4 :ub 5)))
	  (let ((constraints (list ca cb cc c2)))
		
		(test
		 `(
		   (
			(
			 ((ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*))
			 
			 ((ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*))
			 )
			  
			(
			 ((ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*)
			  (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)) 
			 
			 ((ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*))
			 )
			
			(
			 ((ITC-STC :start start :end e1 :lb 0 :ub 0 :guard ,*true*)
			  (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			  (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)) 
			 
			 ((ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*))
			 )
			)

		   (
			(
			 (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)
			 (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
			 
			 (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
			 )
			)
		   )
		 
		 (representative-explanations-of-constraints 
		  stn constraints
		  :method :representativeXPlain)
		 
		 :multiple-values T
		 :test #'equal-explanations-and-conflicts
		 :fail-info "Test 48: representative explanations and conflicts")))))





