;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; FIX:
;;; o arguments for new-constraint and insert-new-constraint is inconsistent
;;;   between ITC and ISAT.  Unify.

;;; To Do:
;;; o stn-initialize! should be generalized to initialize! and unified with
;;;;  ISAT initialization.

;;; ***********************************************************************
;;;
;;;                Testing Incremental Temporal Consistency
;;;
;;; ***********************************************************************

;;; Last update: April 27th, 2008.

;;; ***********************************************************************
;;;
;;;                Test Utilities
;;;
;;; ***********************************************************************

(defun test-consistent-stn! (stn)
  (multiple-value-bind (consistent? conflict)
	  (consistent-stn? stn)
	(if consistent?
		(format t "~% Test returns consistent.")
	  (format t "~% Test returns inconsistent, with conflict = ~a"
			  conflict))))


;;; ***********************************************************************
;;;
;;;                ITC Tests
;;;
;;; ***********************************************************************

(deftest-suite :ITC
;;;    (log-ITC-statistics)
;;;  (stop-and-print-ITC-statistics)
  )

(deftest (test1 :itc) ()
  (let ((stn (make-instance 'ITC-STN :name  'test1
							:start-name 'START)))


	(insert-new-constraint stn :source 'test1 :start 'start :end 'e1 :lb 0 :ub 0)
	(insert-new-constraint stn :source 'test1 :start 'e1 :end 'e2 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test1 :start 'e2 :end 'e3 :lb 4 :ub 5)

 	(let ((c2 (insert-new-constraint stn :source 'test1 :start 'e1 :end 'e3 :lb 4 :ub 5))
		  (c3 (new-constraint stn :source 'test1 :start 'e1 :end 'e3 :lb 0 :ub 20)))

	  (stn-initialize! stn)

	  (test
	   `(nil 
		 ((ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
		  (ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
		  (ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)))


	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((START,E1) for [0,0],(E1,E3) for [4,5], (E2,E3) for [4,5], ~
                   (E1,E2) for [4,5]) -> Inconsistent, all STCs but first.")

	  (insert-constraint stn c3)
	  (remove-constraint stn c2)

	  (test
	   `(t nil)
	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((E1,E3) for [0,20], (E2,E3) for [4,5], (E1,E2) for [4,5]) ~
               -> Consistent.")
	  
	  )))

;;;        e1 - e3
;;; start / \ /   \  end
;;;       \ / \   /
;;;        e2 - e4

(deftest (test2 :itc) ()

  (let ((stn (make-instance 'ITC-STN :name  'test2
							:start-name 'START)))

	(insert-new-constraint stn :source 'test2 :start 'start :end 'e1 :lb 3 :ub 4)
	(insert-new-constraint stn :source 'test2 :start 'e1 :end 'e3 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test2 :start 'e3 :end 'end :lb 5 :ub 6)
	(insert-new-constraint stn :source 'test2 :start 'start :end 'e2 :lb 0 :ub 3)
	(insert-new-constraint stn :source 'test2 :start 'e2 :end 'e4 :lb 3 :ub 4)

	(let ((c2 (insert-new-constraint stn :source 'test2 :start 'e4 :end 'end :lb 3 :ub 4))
		  (c3 (new-constraint stn :source 'test2 :start 'e2 :end 'e3 :lb 3 :ub 4))
		  )
	  (stn-initialize! stn)

	  (test
	   `(nil 
		 ((ITC-STC :start start :end e2 :lb 0 :ub 3 :guard ,*true*)
		  (ITC-STC :start e2 :end e4 :lb 3 :ub 4 :guard ,*true*)
		  (ITC-STC :start e4 :end end :lb 3 :ub 4 :guard ,*true*)
		  (ITC-STC :start e3 :end end :lb 5 :ub 6 :guard ,*true*)
		  (ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
		  (ITC-STC :start start :end e1 :lb 3 :ub 4 :guard ,*true*)))
	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((start,E2) for [0,3], ~
                    (E2,E4) for [3,4], ~
                    (E4,END) for [3,4]) ~
                    (E3,END) for [5,6]) ~
                    (E1,E3) for [4,5], ~
                    (START,E1) for [3,4], ~
               -> Inconsistent, all STCs.")
;;;	  (test-consistent-stn! stn)
;;;	  (print-stn! stn)

	  (remove-constraint stn c2)
	  ;; (test
	  ;;  `(T nil)
	  ;;  (consistent-stn? stn)
   
	  ;;  :multiple-values t
	  ;;  :test #'equal-consistent?-conflict-result
	  ;;  :fail-info "((start,E2) for [0,3], ~
      ;;               (E2,E4) for [3,4], ~
      ;;               (E3,END) for [5,6]) ~
      ;;               (E1,E3) for [4,5], ~
      ;;               (START,E1) for [3,4], ~
      ;;          -> Consistent, nil.")
	  
;;;	  (test-consistent-stn! stn)
;;;	  (print-stn! stn)

	  (insert-constraint stn c3)
	  (test
	   `(T nil)
	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((start,E2) for [0,3], ~
                    (E2,E4) for [3,4], ~
                    (E4,END) for [3,4]) ~
                    (E3,END) for [5,6]) ~
                    (E1,E3) for [4,5], ~
                    (START,E1) for [3,4], ~
                    (E2,E3) for [3,4], ~
               -> Consistent, nil.")
;;;	  (test-consistent-stn! stn)
;;;	  (print-stn! stn)
	  )))

(deftest (test3 :itc) ()
  
  "Same as Test 1, but with guards."
  
  (let ((stn (make-instance 'ITC-STN :name  'test3
							:start-name 'START)))

	(define-state-variable stn 'test3 'A '(A1 A2))

	(insert-new-constraint stn :source 'test3 :start 'start :end 'e1 :lb 0 :ub 0)
	(insert-new-constraint stn :source 'test3 :start 'e1 :end 'e2 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test3 :start 'e2 :end 'e3 :lb 4 :ub 5)

	(insert-new-constraint stn :source 'test3 :start 'e1 :end 'e3 :lb 4 :ub 5
						   :guard '((= A A1)))
	(insert-new-constraint stn :source 'test3 :start 'e1 :end 'e3 :lb 0 :ub 20
						   :guard '((= A A2)))

	(stn-initialize! stn)
;;;	(format t "~2%Assign A = a1")
	(assign! stn 'A 'A1)
	(test
	 `(nil 
	   ((ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ((= A A1)))
		(ITC-STC :start e2 :end e3 :lb 4 :ub 5 :guard ,*true*)
		(ITC-STC :start e1 :end e2 :lb 4 :ub 5 :guard ,*true*)))
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E1) for [0,0], ~
                    (E1,E2) for [4,5], ~
                    (E2,E3) for [4,5]) ~
                    A=A1 -> (E1,E3) for [4,5]) ~
                    A=A2 -> (E1,E3) for [0,20], ~
                    A=A1 ~
               -> Inconsistent, all STCs.")
;;;	(test-consistent-stn! stn)
;;;	(print-stn! stn)

	(unassign! stn 'A 'A1)
	(assign! stn 'A 'A2)

;;;	(format t "~2%Unassign A = a1, assign A = a2")

	(test
	 `(T nil)
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E1) for [0,0], ~
                    (E1,E2) for [4,5], ~
                    (E2,E3) for [4,5]) ~
                    A=A1 -> (E1,E3) for [4,5]) ~
                    A=A2 -> (E1,E3) for [0,20], ~
                    A=A2 ~
               -> Consistent, nil.")
;;;	(test-consistent-stn! stn)
;;;	(print-stn! stn)
	))

(deftest (test4 :itc) ()

  (let ((stn (make-instance 'ITC-STN :name  'test4
							:start-name 'START)))

	(define-state-variable stn 'test4 'A '(A1 A2))

	(insert-new-constraint stn :source 'test4 :start 'start :end 'e1 :lb 3 :ub 4)
	(insert-new-constraint stn :source 'test4 :start 'e1 :end 'e3 :lb 4 :ub 5)
	(insert-new-constraint stn :source 'test4 :start 'e3 :end 'end :lb 5 :ub 6)
	(insert-new-constraint stn :source 'test4 :start 'start :end 'e2 :lb 0 :ub 3)
	(insert-new-constraint stn :source 'test4 :start 'e2 :end 'e4 :lb 3 :ub 4)

	(insert-new-constraint stn :source 'test4 :start 'e4 :end 'end :lb 3 :ub 4 :guard '((= A A1)))
	(insert-new-constraint stn :source 'test4 :start 'e2 :end 'e3 :lb 3 :ub 4 :guard '((= A A2)))
	
	(stn-initialize! stn)

;;;	(format t "~2%Assign A = a1")
	(assign! stn 'A 'A1)

	(test
	 `(nil 
	   (
		(ITC-STC :start start :end e2 :lb 0 :ub 3 :guard ,*true*)
		(ITC-STC :start e2 :end e4 :lb 3 :ub 4 :guard ,*true*)
		(ITC-STC :start e4 :end end :lb 3 :ub 4 :guard ((= A A1)))
		(ITC-STC :start e3 :end end :lb 5 :ub 6 :guard ,*true*)
		(ITC-STC :start e1 :end e3 :lb 4 :ub 5 :guard ,*true*)
		(ITC-STC :start start :end e1 :lb 3 :ub 4 :guard ,*true*)))
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E1) for [3,4], ~
                    (E1,E3) for [4,5], ~
                    (E3,End) for [5,6], ~
                    (start,E2) for [0,3]) ~
                    (E2,e4) for [3,4]) ~
                    A=A1 -> (E4,End) for [3,4]) ~
                    A=A2 -> (E2,E3) for [3,4], ~
                    A=A1
               -> Inconsistent, All STCs except last.")
;;;	(test-consistent-stn! stn)
;;;	(print-stn! stn)

;;;	(format t "~2%Unassign A = a1")
	(unassign! stn 'A 'A1)
	(test
	 `(T nil)
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E1) for [3,4], ~
                    (E1,E3) for [4,5], ~
                    (E3,End) for [5,6], ~
                    (start,E2) for [0,3]) ~
                    (E2,e4) for [3,4]) ~
                    A=A1 -> (E4,End) for [3,4]) ~
                    A=A2 -> (E2,E3) for [3,4]
               -> Consistent, nil.")
;;;	(test-consistent-stn! stn)
;;;	(print-stn! stn)

;;;	(format t "~2% assign A = a2")
	(assign! stn 'A 'A2)
	(test
	 `(T nil)
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E1) for [3,4], ~
                    (E1,E3) for [4,5], ~
                    (E3,End) for [5,6], ~
                    (start,E2) for [0,3]) ~
                    (E2,e4) for [3,4]) ~
                    A=A1 -> (E4,End) for [3,4]) ~
                    A=A2 -> (E2,E3) for [3,4]
                    A=A2
               -> Consistent, nil.")
;;;	(test-consistent-stn! stn)
;;;	(print-stn! stn)
	))

(deftest (test5 :itc) ()

  "Similar to test20 in opitc, but no state variables."
  
  (let ((stn (make-instance 'ITC-STN :name  'test5
							:start-name 'START)))

	(insert-new-constraint stn :source 'test1 :start 'start :end 'e2 :lb 6 :ub 6)

	(let ((c2 (insert-new-constraint stn :source 'test5 :start 'start :end 'e2 :lb 4 :ub 5))
		  (c3 (new-constraint stn :source 'test5 :start 'start :end 'e2 :lb 0 :ub 20)))

	  (stn-initialize! stn)
	  (test
	   `(nil 
		 (
		  (ITC-STC :start start :end e2 :lb 4 :ub 5 :guard ,*true*)
		  (ITC-STC :start start :end e2 :lb 6 :ub 6 :guard ,*true*)))
	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((start,E2) for [4,5], ~
                      (start,E2) for [6,6], ~
                    A=A1
               -> Inconsistent, All STCs.")
;;;	  (print-stn! stn)
;;;	  (test-consistent-stn! stn)

	  (insert-constraint stn c3)
	  (remove-constraint stn c2)
	  
	  (test
	   `(T nil)
	   (consistent-stn? stn)
   
	   :multiple-values t
	   :test #'equal-consistent?-conflict-result
	   :fail-info "((start,E2) for [4,5], ~
                      (start,E2) for [0,20], ~
                    A=A1
               -> Consistent, nil.")

;;;   (print-stn! stn)
;;;	  (test-consistent-stn! stn)
	  )))

(deftest (test6 :itc) ()

  (let ((stn (make-instance 'ITC-STN :name  'test6
							:start-name 'START)))

	"Similar to test20 in OPITC, using state variables."

	(define-state-variable stn 'test6 'A '(hi low))

	(insert-new-constraint stn :source 'test6 :start 'start :end 'e2
					:lb 6 :ub 6)

	(insert-new-constraint stn  :source 'test6 :start 'start :end 'e2
						   :lb 4 :ub 5 :guard '((= A hi)))
	(insert-new-constraint stn :source 'test6 :start 'start :end 'e2
						   :lb 0 :ub 20 :guard '((= A low)))

;;;  (stn-initialize! stn)
	(assign! stn 'A 'hi)
	(test
	 `(nil 
	   (
		(ITC-STC :start start :end e2 :lb 4 :ub 5 :guard ((= A hi)))
		(ITC-STC :start start :end e2 :lb 6 :ub 6 :guard ,*true*)))
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E2) for [4,5], ~
                   A=hi -> (start,E2) for [6,6], ~
                   A=low -> (start,E2) for [0,20], ~
                   A=hi
               -> Inconsistent, First two STCs.")
;;;  (print-stn! stn)
;;;	(test-consistent-stn! stn)

	(unassign! stn 'A 'hi)
	(assign! stn 'A 'low)
	(test
	 `(T nil )
	 (consistent-stn? stn)
   
	 :multiple-values t
	 :test #'equal-consistent?-conflict-result
	 :fail-info "((start,E2) for [4,5], ~
                   A=hi -> (start,E2) for [6,6], ~
                   A=low -> (start,E2) for [0,20], ~
                   A=low
               -> Consistent, nil.")
;;; (print-stn! stn)
;;;	(test-consistent-stn! stn)
	))

