(in-package #:kirk)

(defmethod compute-span ((tpn tpn::temporal-plan-network) &key (inf *inf*) (-inf *-inf*) (rational-inf most-positive-double-float) (-rational-inf most-negative-double-float))
  (let ((event-ht (make-hash-table))
		(num-events 0) ;; the number is incremented in the loop
		(num-episodes (length (tpn::tpn-episodes tpn))))
	;; Create a hash table that maps events to an index number.
	(loop for event in (tpn::tpn-events tpn) do
		 (setf (gethash event event-ht) num-events)
		 (incf num-events))
	;; Create the A, b, and c  matricies.
	(let*  (;; Each episode has a lower and upperbound constraint
			(m (* 2 num-episodes))
			;; Each event has an occurence time
			;; plus two slack variable, one each for the lower and upper bounds.
			(n (+ num-events (* 2 num-episodes)))
			(A (make-array `(,m ,n) :element-type 'number :initial-element 0))
			(b (make-array m :element-type 'number :initial-element 0))
			(c (make-array n :element-type 'number :initial-element 0))
			(tpn-starti (gethash (tpn::tpn-start tpn) event-ht))
			(tpn-endi (gethash (tpn::tpn-end tpn) event-ht)))
	  ;; Create the A and b matrix
	  (loop for episode in (tpn::tpn-episodes tpn)
		   with mi = 0 do
		   (let ((starti (gethash (tpn::episode-start episode) event-ht))
				 (endi (gethash (tpn::episode-end episode) event-ht))
				 (ub (to-rinf (tpn::episode-upperbound episode) :rational-inf rational-inf :-rational-inf -rational-inf))
				 (lb (to-rinf (tpn::episode-lowerbound episode) :rational-inf rational-inf :-rational-inf -rational-inf)))
			 ;; b - a + s1 = ub, where s1>=0
			 (setf (aref A mi endi) 1)
			 (unless (eql tpn-starti starti)
			   (setf (aref A mi starti) -1))
			 (setf (aref A mi (+ num-events mi)) 1)
			 (setf (aref b mi) ub)
			 (incf mi)
			 ;; b - a - s2 = lb, where s2>=0
			 (setf (aref A mi endi) 1)
			 (unless (eql tpn-starti starti)
			   (setf (aref A mi starti) -1))
			 (setf (aref A mi (+ num-events mi)) -1)
			 (setf (aref b mi) lb)
			 (incf mi)))
	  ;; Create the c matrix
	  (setf (aref c tpn-endi) 1)
	  ;; 
	  (let ((min-sol nil)
	  		(min-c nil)
	  		(max-sol nil)
	  		(max-c nil))
	  	;; Solve the simplex for the minimum case
	  	(multiple-value-setq (min-sol min-c)
	  	  (simplex:simplex A b c))
	  	;; Create the c matrix by multiplying by -1
		;; solve for the maximum case
		(setf (aref c tpn-endi) -1)
	  	;; Solve the simplex for the maximum case
	  	(multiple-value-setq (max-sol max-c)
	  	  (simplex:simplex A b c))
	  	;; Return the min and max values
		(if min-sol
			(setf min-c (second min-c))
			(setf min-c inf))
		(if max-sol
			(setf max-c (- (second max-c)))
			(setf max-c -inf))
	  	(values (to-pinf min-c :rational-inf rational-inf :-rational-inf -rational-inf) 
				(to-pinf max-c :rational-inf rational-inf :-rational-inf -rational-inf))
		))))