;;;; Copyright (c) 2009-2010 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Peng Yu

(in-package #:kirk)

;;-------------------------------------------------------------------------------
;; EXPLANATIONS
;; The following is work-in-progress, code for allowing Kirk to generate
;; explanations.
;;-------------------------------------------------------------------------------

(defun write-explanations (stream explanations &key 
						   ;;(explanation-in-JSON nil)
						   (*episodes-queue-table* nil)
						   (*queue* nil)
						   (*event-name-table* nil)
						   (*string-head*  (format nil "{\"explanations\": ["))
						   (*string-end*  (format nil "]}")) 
						   (*choice-index* 1)
						   (*exp-string* nil)
						   (*single-exp* nil))
  "This function writes explanations in to a .xml file with stream"


  (if (not *episodes-queue-table*)

	  (loop for explanation in (reverse explanations)
		   
		 do (setf *exp-string* (explanation-to-JSON-string explanation *event-name-table*))
		 (let ((current-choice (first explanation))) 

		   (setf *single-exp* (format nil "{\"Choice~a\": [{\"Choice\": \"~a\"}, ~a]}, " 
									  *choice-index* current-choice *exp-string*))
		   

		   (setf *string-head* (concatenate 'string *string-head* *single-exp*))

		   (setf *choice-index* (+ *choice-index* 1))
		   ))

	  (loop for explanation in (reverse explanations)

		 do (setf *exp-string* (explanation-to-JSON-string explanation *event-name-table*))
		 (let ((current-choice (get-choice (first explanation) *queue*))) 

		   (setf *single-exp* (format nil "{\"Choice~a\": [{\"Choice\": \"~a\"}, ~a]}, " 
									  *choice-index* (gethash current-choice *episodes-queue-table*) *exp-string*))
		   

		   (setf *string-head* (concatenate 'string *string-head* *single-exp*))

		   (setf *choice-index* (+ *choice-index* 1))
		   ))
	  
	  )

   (setf *string-head* (subseq *string-head* 0 (- (length *string-head*) 3)))
   (setf *string-head* (concatenate 'string *string-head* *string-end*))
   
   (format stream "~a" *string-head*)
   'T)

(defun get-choice (input-choice *queue* &key
				   (output-choice nil)
				   (match T))

  "Find the corresponding choice in *queue* to input-choice"

  (loop for formal-choice in *queue*

	   do (setf match T)
	  ;; do (format t "Decision: ~A~%" formal-choice)
	   

	   (loop for individual-decision in formal-choice
			
			;;do (format t "   Individual Decision: ~A~%" individual-decision)
			
			do (when (not (choice-match? input-choice individual-decision))
				   
				 (setf match nil)))

			(when match
			  (setf output-choice formal-choice)
			;;  (format t "Result ~A~%" output-choice)
			  ))

  output-choice)


(defun choice-match? (input-choice individual-decision
					  &key (match nil)
					  (regs nil)
					  (choice-match? nil)
					  (decision nil)
					  (decision-result nil))

  "Check if there is a matching decision in input-choice"

 ;;  (multiple-value-setq (match regs)
;;	(cl-ppcre:scan-to-strings "\\(\\s*(=*)\\s*\\)" individual-decision))

  (multiple-value-setq (match regs)
	(cl-ppcre:scan-to-strings "\\W\\W\\s(\\d{1,4})\\s(\\d{1,4})\\W" (format nil "~A" individual-decision)))

  (when match
	(setf decision (aref regs 0))
	(setf decision-result (aref regs 1)))


  (loop for choice in input-choice

	 do  (multiple-value-setq (match regs)
		 (cl-ppcre:scan-to-strings "(\\d{1,4})\\W(\\d{1,4})" (format nil "~A" choice)))
	;; do (format t "      Choice ~A ~A~%" (aref regs 0) (aref regs 1))

	   (when (and (equal decision (aref regs 0)) (equal decision-result (aref regs 1)))

		;; (format t "Match ~A ~A~%" individual-decision input-choice)
		 (setf choice-match? T)))
  
  choice-match?)

(defun explanation-to-JSON-string (explanation *event-name-table* &key 
								   (*exp-string* (format nil "{\"exp1\": ["))
								   (*exp-string-end* (format nil "]}"))
								   (*constraint-index* 1)
								   (*single-relaxation* nil))

  ;;(format t "Number of entries in hash table: ~A~%" (hash-table-count *event-name-table*))



  (loop for relaxation in (second explanation)
	   
	 do (if (not *event-name-table*)
			(setf *single-relaxation* (format nil "{\"constraint~A\": [{\"fromnode\": \"~A\"}, {\"tonode\": \"~A\"}, {\"mintime\": ~A}, {\"maxtime\": ~A}, {\"validmintime\": ~A}, {\"validmaxtime\": ~A}]}, " 
											  *constraint-index*
											  (format nil "~a" (first (first relaxation)))
											  (format nil "~a" (second (first relaxation)))
											  (first (second relaxation))
											  (second (second relaxation))
											  (third (second relaxation))
											  (fourth (second relaxation))))

			(setf *single-relaxation* (format nil "{\"constraint~A\": [{\"fromnode\": \"~A\"}, {\"tonode\": \"~A\"}, {\"mintime\": ~A}, {\"maxtime\": ~A}, {\"validmintime\": ~A}, {\"validmaxtime\": ~A}]}, " 
											  *constraint-index*
											  (gethash (format nil "~a" (first (first relaxation))) *event-name-table*)
											  (gethash (format nil "~a" (second (first relaxation))) *event-name-table*)
											  (first (second relaxation))
											  (second (second relaxation))
											  (third (second relaxation))
											  (fourth (second relaxation))))
			)

	 ;; (format t "Looking for ~A and get ~A~%" (first (first relaxation)) (gethash (first (first relaxation)) *event-name-table*))

	   
	   (setf *exp-string* (concatenate 'string *exp-string* *single-relaxation*))

	   (setf *constraint-index* (+ *constraint-index* 1))
	   )
  
  (setf *exp-string* (subseq *exp-string* 0 (- (length *exp-string*) 3)))
  (setf *exp-string* (concatenate 'string *exp-string* *exp-string-end*))
  
  *exp-string*)


(defun get-decision-queue (tpn &key 
                               (*node-index* 0)
                               (*node-choice* 0)
                               (*current-choice*  (format nil " "))
                               (*current-choice-path*  nil)
                               (*temp-queue* nil)
                               (*queue* nil)
                               (*episode-queue* nil)
                               (*temp-episode-queue* nil)
                               (temp-current-episode-path nil)
                               (choice-episode-table nil)
						   (episode-name-table (make-hash-table :test 'equal))
						   (event-name-table (make-hash-table :test 'equal)))
  
  "This function enumerate all possible combinations of decisions in the tpn.
It returns all combinations in a queue."

     
  (loop for choice-node in (tpn::tpn-decisions tpn)
        
        ;; Generate the first choice node in queue (choice and episode)
      do (setf *node-index* (+ 1 *node-index*))
        
       (when (= *node-index* 1) ;; if this is the first decision node in queue
           (loop for choice-episodes in (tpn::event-successor-episodes choice-node)
               do (setf *node-choice* (+ *node-choice* 1)) ;; Count the current choice index for this decision node
                (setf *current-choice-path*  (list `(= ,(tpn::event-name choice-node) ,*node-choice*)))
                (push *current-choice-path* *queue*) ;; Construct the choice queue
                (push (list choice-episodes) *episode-queue*)) ;; Construct the episode queue
           (setf *node-choice* 0)) ;; Reset choice index in decision nodes 
        

        ;; If this is not the first decision node in queue
      do (when (/= *node-index* 1)
           
           (setf *temp-queue* *queue*) ;;Copy current choice queue
           (setf *temp-episode-queue* *episode-queue*) ;; Copy current episode queue
           
           (setf *queue* nil) ;; Clean current choice queue for new choice paths
           (setf *episode-queue* nil) ;; Clean current choice queue for new episode paths
           
           
           ;; Start with choice path. Loop throuth all choice path in current queue and add new decisions
           (loop for current-choice-path in *temp-queue*
               do (setf temp-current-episode-path (gethash current-choice-path choice-episode-table)) ;; Map current choice path to episode path
                 ;; Check if current choice node belongs to the current choice path. If not, it will not be added
                  (when (in-queue? (get-previous-choice tpn choice-node) temp-current-episode-path) 
                    ;; If yes, add the new decisions to the choice path
                     (loop for choices in (tpn::event-successor-episodes choice-node) 
                           do (setf *node-choice* (+ *node-choice* 1))
                            (setf *current-choice*  `(= ,(tpn::event-name choice-node) ,*node-choice*))
                            (setf *current-choice-path* current-choice-path)
                            (push *current-choice* *current-choice-path*)
                            (push *current-choice-path* *queue*)))
                 ;; Keep a copy of the original choice path in the queue in case other decisions need to use it.
               do (push current-choice-path *queue*)
               do (setf *node-choice* 0))
           
           ;; Next, do the same for the episode queue. Loop throught all episode path in the queue.
           (loop for current-episode-path in *temp-episode-queue*
               do (when (in-queue? (get-previous-choice tpn choice-node) current-episode-path)
                    ;; If current episode path contains current choice node, continue.
                     (loop for choice-episode in (tpn::event-successor-episodes choice-node)
                           do (setf temp-current-episode-path  current-episode-path)
                            (push choice-episode temp-current-episode-path)
                            (push temp-current-episode-path *episode-queue*)))
                 ;; Keep a copy of the original episode path in the queue in case other decisions need to use it.
                  (push current-episode-path *episode-queue*))
           )

        ;; Generate the hash table between choice path and episode path
        (setq choice-episode-table (hash-choice-episode *queue* *episode-queue*)))
  
  (when (not (= *node-index* 1))
  ;; Update the choice queue. Remove redundant choice paths. 
  (setf *temp-queue* *queue*) ;;Copy current choice queue
  (setf *queue* nil) ;; Clean current choice queue for new choice paths
  
  ;; Start with choice path. Loop throuth all choice path in current queue and add new decisions
  (loop for current-choice-path in *temp-queue*
      do (setf temp-current-episode-path (gethash current-choice-path choice-episode-table)) ;; Map current choice path to episode path
        do (when (toward-end? tpn (first temp-current-episode-path))
           (push current-choice-path *queue*))))

  ;;Map the episodes to their represented names
  (loop for current-episode in (TPN::tpn-episodes tpn)
	 do (when (not (equal (TPN::episode-primitive current-episode) NIL))
		  ;;(format t "Current episode ~A ~A~%" current-episode (TPN::episode-primitive current-episode))
		  
		  (let ((current-primitive (TPN::episode-primitive current-episode))
				(index1 (position #\" (TPN::episode-primitive current-episode) :start 0))
				(index2 0)
				(index3 0)
				(index4 0)
				(start-event-name nil)
				(end-event-name nil))

		  (when index1
			(setf index2 (position #\" current-primitive :start (+ index1 1)))
			(setf index3 (position #\" current-primitive :start (+ index2 1)))
			(setf index4 (position #\" current-primitive :start (+ index3 1)))

			(setf start-event-name (subseq current-primitive (+ index1 1) index2))
			(setf end-event-name (subseq current-primitive (+ index3 1) index4)))

		  (when (not index1)

			(setf start-event-name (format nil "~a" current-primitive))
			(setf end-event-name (format nil "~a" current-primitive)))

		  
		  (setf (gethash (list (TPN::event-name (TPN::episode-start current-episode))
							   (TPN::event-name (TPN::episode-end current-episode)))
						 episode-name-table) 
				(list start-event-name end-event-name))

		  (setf (gethash (TPN::event-name (TPN::episode-start current-episode)) event-name-table) start-event-name)
		  (setf (gethash (TPN::event-name (TPN::episode-end current-episode)) event-name-table) end-event-name)
		  ;;(format t "Adding ~A ~A~%" (TPN::event-name (TPN::episode-start current-episode)) start-event-name)
		  
		  )))

  (loop for current-selection in *queue*
	   
	 
	 ;;  (format t "Current Selection ~A  ~A~%" current-selection current-episode)
	 do (let ((current-selection-name (format nil ""))
		   (current-episode (gethash current-selection choice-episode-table))) 
	   (loop for target-episode in current-episode
			
		  do (loop while (equal (TPN::episode-primitive target-episode) NIL)
				do (setf target-episode (first (TPN::event-successor-episodes (TPN::episode-end target-episode))))
				  (setf current-selection-name (format nil "~a/~a-~a"
													   (first (gethash (list (TPN::event-name (TPN::episode-start target-episode))
																			 (TPN::event-name (TPN::episode-end target-episode)))
																	   episode-name-table))
													   (second (gethash (list (TPN::event-name (TPN::episode-start target-episode))
																			  (TPN::event-name (TPN::episode-end target-episode)))
																		episode-name-table))
                                                       current-selection-name 
													   
													   )))
			
			(setf (gethash current-selection choice-episode-table) 
                  (subseq current-selection-name 0 (- (length current-selection-name) 1))))))


  (list *queue* choice-episode-table event-name-table))



(defun hash-choice-episode (choice-queue episode-queue &key
							(choice-episode-table nil)
							(size nil)
							(episode-path nil)
							(choice-path nil))
  
  "This function creates a hash table between the choice path and episode path"
  
  (setf size (- (list-length choice-queue) 1))
  (setq choice-episode-table (make-hash-table)) 
  (loop for x from 0 to size
	 do (setf choice-path (nth x choice-queue))
       (setf episode-path (nth x episode-queue))
       (setf (gethash choice-path choice-episode-table) episode-path))
  
  choice-episode-table)

(defun in-queue? (episode episode-queue &key
                          (in-queue nil))
  
  "This function checks whether an episode queue constains a specific episode"
  
   (loop for choice-episode in episode-queue
             do (when (equal choice-episode episode)
                     (setf in-queue T))
               
        )
  
  in-queue)


(defun get-event-parent (event &key
                               (parent-event nil))
  
  "Return the parent node of the input node"
  
  (setf parent-event (tpn::episode-start (first (tpn::event-predecessor-episodes event))))
  
  parent-event)


(defun get-event-child (event &key
                               (child-event nil))
  
  "Return the child node of the input node"
  
  (loop for successor in (tpn::event-successor-episodes event)
      do (push (tpn::episode-end successor) child-event))

  
  child-event)


(defun get-previous-choice (tpn choice-event &key
                                (parent-choice-event nil)
                                (parent-choice-episode nil)
                                (child-event nil))
  
  "Return the choice node before the current one. If not exist, return nil"
  
  (setf child-event choice-event)
  (loop for node in (tpn::tpn-events tpn)
      do (setf parent-choice-episode (first (tpn::event-predecessor-episodes child-event)))
       (setf parent-choice-event (tpn::episode-start parent-choice-episode))
       (setf child-event parent-choice-event)
        ;;do (format t "is decision? ~a     is start-event? ~a ~%" (tpn::decision-event? parent-choice-event) (equal parent-choice-event (tpn::tpn-start tpn)))
      until (or (tpn::decision-event? parent-choice-event) (equal parent-choice-event (tpn::tpn-start tpn))))
  
  (when (equal parent-choice-event (tpn::tpn-start tpn))
     (setf parent-choice-episode nil)
    )
  
  parent-choice-episode)


(defun toward-end? (tpn episode &key
                                (event-queue nil)
                                (toward-end? nil))
  
  "Tell if the current decision node links to the end node."
  (push (tpn::episode-end episode) event-queue)

  
  (loop while (> (list-length event-queue) 0)
	 with event
	 do (setf event (pop event-queue))

	 do (loop for child-event in (get-event-child event)
               
		   do (when (not (tpn::decision-event? child-event))

				(push child-event event-queue))
               
		   do (when (equal child-event (tpn::tpn-end tpn))
				(setf toward-end? T))
               
		   until (equal toward-end? T))

	 until (equal toward-end? T))
  

  toward-end?)

  
  



(defun add-to-queue (new-choice *queue*)	
  "Adds a single choice sequence information to the choices queue."
	
  (push new-choice *queue*)
  *queue*)
