;;; TEST SCRIPT

(in-package #:kirk)

(defparameter *test-tpn-xml* 
"<TPNS>
  <TPN>
    <FORMAT>Spec2</FORMAT>
    <START>START</START>
    <END>NODE2</END>
    <ARC>
      <START>START</START>
      <END>NODE1</END>
      <COST>0.0</COST>
      <LOWERBOUND>5.0</LOWERBOUND>
      <UPPERBOUND>6.0</UPPERBOUND>
    </ARC>
    <ARC>
      <START>NODE1</START>
      <END>NODE2</END>
      <PRIMITIVE>UAV.flyTo(\"92f1260c-8ce4-4f71-85c2-f71ab1cc1caa\", \"2b990672-21df-4af3-b707-b2b247f4c593\")</PRIMITIVE>
      <COST>0.0</COST>
      <LOWERBOUND>0.0</LOWERBOUND>
      <UPPERBOUND>+INF</UPPERBOUND>
    </ARC>
  </TPN>
</TPNS>"
)

(defparameter *test-env-xml* 
"<ENVIRONMENT>
 <VEHICLE>
    <NAME>UAV</NAME>
    <INITIAL_POSITION>
      <X2D>0.0</X2D>
      <Y2D>0.0</Y2D>
      <Z2D>10000.0</Z2D>
    </INITIAL_POSITION>
    <INITIAL_VELOCITY>
      <DIRECTION>
        <X2D>-0.004249852685447234</X2D>
        <Y2D>-0.007159751818306398</Y2D>
        <Z2D>0.9999653377522904</Z2D>
      </DIRECTION>
      <SPEED>0.1</SPEED>
    </INITIAL_VELOCITY>
    <ACTION>
      <NAME>flyTo</NAME>
      <MIN_ACCELERATION>1.0</MIN_ACCELERATION>
      <MAX_ACCELERATION>99.0</MAX_ACCELERATION>
      <MIN_VELOCITY>2.0</MIN_VELOCITY>
      <MAX_VELOCITY>88.0</MAX_VELOCITY>
    </ACTION>
 </VEHICLE>
  <FEATURE>
    <NAME>2b990672-21df-4af3-b707-b2b247f4c593</NAME>
    <TARGET>UAV</TARGET>
    <CENTER>
      <X2D>2.0</X2D>
      <Y2D>3.0</Y2D>
      <Z2D>10000.0</Z2D>
    </CENTER>
    <RADIUS>100.0</RADIUS>
  </FEATURE>
  <FEATURE>
    <NAME>92f1260c-8ce4-4f71-85c2-f71ab1cc1caa</NAME>
    <TARGET>UAV</TARGET>
    <CENTER>
      <X2D>0.0</X2D>
      <Y2D>0.0</Y2D>
      <Z2D>10000.0</Z2D>
    </CENTER>
    <RADIUS>100.0</RADIUS>
  </FEATURE>
</ENVIRONMENT>"
)

(defun test-kirk-vz-graph ()
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/temp/tempTPN.rmpl2.tpn"))
		 (env-in-pathname (cl-user::mars-pathname "Kirk/tests/temp/tempEnv.xml"))
		 (tpn-out-pathname (cl-user::mars-pathname "Kirk/tests/temp/tempTPNout.rmpl.tpn" ))
		 (env-out-pathname (cl-user::mars-pathname "Kirk/tests/temp/tempEnvout.xml" )))
	(kirk-vz-graph-file tpn-in-pathname env-in-pathname tpn-out-pathname env-out-pathname)
	))

(defun test-kirk-vz-graph-no-choice ()
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/test2/tempTPNnc.rmpl2.tpn"))
		 (env-in-pathname (cl-user::mars-pathname "Kirk/tests/test2/tempEnvnc.xml"))
		 (tpn-out-pathname (cl-user::mars-pathname "Kirk/tests/test2/tempTPNoutnc.rmpl.tpn" ))
		 (env-out-pathname (cl-user::mars-pathname "Kirk/tests/test2/tempEnvoutnc.xml" )))
	(kirk-vz-graph-file tpn-in-pathname env-in-pathname tpn-out-pathname env-out-pathname)
	))

(defun test-kirk-vz-graph-with-exp ()
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempTPN.rmpl2.tpn"))
		 (env-in-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempEnv.xml"))
		 (tpn-out-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempTPNout.rmpl.tpn" ))
		 (env-out-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempEnvout.xml" ))
		 (exp-out-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempExpout.xml")))
	(kirk-vz-graph-file tpn-in-pathname env-in-pathname tpn-out-pathname env-out-pathname :exp-out-pathname exp-out-pathname)
	))

(defun test-kirk-boeing-manufacturing-with-exp ()
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/manufacturing/plan_inputwC.tpn"))
		 (tpn-out-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/manufacturing/plan_good.tpn" ))
		 (exp-out-pathname (cl-user::mars-pathname "Kirk/tests/boeing-2012/manufacturing/plan_exp.xml")))
	(kirk-file tpn-in-pathname tpn-out-pathname :exp-out-pathname exp-out-pathname)
	))

(defun test-kirk-boeing-manufacturing ()
  (let* ((tpn-in-pathname (cl-user::mars-pathname "Kirk/tests/test-boeing/Block-Stack-Simple/ManufacturingScenario.rmpl2.tpn"))
		 (tpn-out-pathname (cl-user::mars-pathname "Kirk/tests/test-boeing/Block-Stack-Simple/KirkOutput.rmpl.tpn" )))
	(kirk-file tpn-in-pathname tpn-out-pathname)))

(defun test-kirk-vz-graph-string () 
  (let* ((tpn *test-tpn-xml*)
		 (env *test-env-xml*)
		 (tpn-in-stream (make-string-input-stream tpn))
		 (env-in-stream (make-string-input-stream env))
		 (tpn-out-stream (make-string-output-stream))
		 (env-out-stream (make-string-output-stream)))
	(kirk-vz-graph-stream tpn-in-stream env-in-stream tpn-out-stream env-out-stream))
	)


(defun test-kirk-test4 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test4/tpn.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	;(break)
	;;(rationalize-tpn tpn)
	(kirk tpn)
	))

(defun test-kirk-test5 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test5/tpn.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	;;(rationalize-tpn tpn)
	(kirk tpn)
	))

(defun test-kirk-test6 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test6/tpn.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	;;(rationalize-tpn tpn)
	(kirk tpn)
	))

(defun test-kirk-test7 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test7/tpn.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns))
		 (env-path (cl-user::mars-pathname "Kirk/tests/test7/env.xml"))
		 (env (read-environment-file env-path)) )
	;;(rationalize-tpn tpn)
	(kirk-vz-graph tpn env)
	))

(defun test-kirk-test8 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test8/tpn.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns))
		 (env-path (cl-user::mars-pathname "Kirk/tests/test8/env.xml"))
		 (env (read-environment-file env-path)) )
	;;(rationalize-tpn tpn)
	(kirk-vz-graph tpn env)
	
	(loop for episode in (tpn::tpn-episodes tpn) do
		 (format t "TELL: ~a~%" (tpn::episode-tell episode))

		 )
	))

(defun test-kirk-test9 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempTPN.rmpl2.tpn"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns))
		 (env-path (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempEnv.xml"))
		 (env (read-environment-file env-path)) )
	;;(rationalize-tpn tpn)
	(multiple-value-bind (feasible tpn-out env-out best-choice explanations tpn-int)
		(kirk-vz-graph tpn env)
	  
	  ;; (loop for episode in (tpn::tpn-episodes tpn) do
	  ;; 	 (format t "TELL: ~a~%" (tpn::episode-tell episode))
	  ;; 	 )
	  (declare (ignore feasible env-out best-choice explanations tpn-int))

	  (tpn2dot tpn-out (cl-user::mars-pathname "Kirk/tests/boeing-2012/tempTPNout.pdf")))
	))

(defun test-kirk-test10 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test-spec3/test1.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (tpn::parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn));builds the TPN Lisp objects
		 (tpn (first tpns));just the first TPN, no matter how many are present in the file
		 )
	
	;;(break)
	;;(rationalize-tpn tpn)
	(kirk tpn)

	(loop for ep in (tpn::tp-episodes tpn) do
		 (format t "~%Description: ~a" (tpn::episode-description ep))
		 (format t "~%Label: ~a" (tpn::episode-activation-label ep))
		 (format t "~%Assignment: ~a~%" (tpn::episode-activation-assignment ep)))

	))

(defun test-kirk-test11 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test-spec3/test2.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (tpn::parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	
	;;(break)
	;;(rationalize-tpn tpn)
	(kirk tpn)

	(loop for ep in (tpn::tp-episodes tpn) do
		 (format t "~%Description: ~a" (tpn::episode-description ep))
		 (format t "~%Label: ~a" (tpn::episode-activation-label ep))
		 (format t "~%Assignment: ~a~%" (tpn::episode-activation-assignment ep)))

	))

(defun test-kirk-test12 ()
  (let* ((tpn-path (cl-user::mars-pathname "Kirk/tests/test-spec3/test3.xml"))
		 (tpn-xml (mtk-utils:read-file-to-string tpn-path :multibyte t))
		 (tpns (tpn::parse-tpn-string tpn-xml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	
	;;(break)
	;;(rationalize-tpn tpn)
	(kirk tpn)

	(loop for ep in (tpn::tp-episodes tpn) do
		 (format t "~%Description: ~a" (tpn::episode-description ep))
		 (format t "~%Label: ~a" (tpn::episode-activation-label ep))
		 (format t "~%Assignment: ~a~%" (tpn::episode-activation-assignment ep)))

	))



(defun rationalize-tpn (tpn)
  (loop for episode in (tpn::tpn-episodes tpn) do
	   (with-slots ((lowerbound tpn::lowerbound) (upperbound tpn::upperbound)) episode
		 (setf lowerbound (brationalize lowerbound))
		 (setf upperbound (brationalize upperbound))
		 )))