
;;; resource-manager-interface.lisp
(in-package #:resource-manager)

(defclass resource-manager-interface ()
  ((notification-instance)
   (dispatchable)))

(defgeneric getFeasibility ((interface resource-manager-interface) (id string) (tpn string) (env string) &key (priority 5) (start-time `(:relative 0))))
(defgeneric submit ((interface resource-manager-interface) (id string) (tpn string) (env string) &key (priority 5) (start-time `(:relative 0))))
(defgeneric cancel ((interface resource-manager-interface) (id string) (tpn string)))
(defgeneric start ((interface resource-manager-interface) (env string) &key (time-tolerance 0.5) (ignore-missed-activities t)))
(defgeneric stop ((interface resource-manager-interface)))
(defgeneric addObstacle ((interface resource-manager-interface) (obstacle string) &key (replace t)))
(defgeneric removeObstacle ((interface resource-manager-interface) (obstacle string)))
(defgeneric replan ((interface resource-manager-interface)))

(defmethod set-notification-instance ((interface resource-manager-interface) notification-instance)
  (setf (slot-value interface 'notification-instance) notification-instance))

(defmethod set-dispatchable ((interface resource-manager-interface) x)
  (setf (slot-value interface 'dispatchable) x))

#+abcl
(defmethod notify-task-status ((interface resource-manager-interface) (taskID string) (submitterID string) (status string) (description string) (resourceUri string) (resourceName string))
  (handler-case 
	  (let* ((class (java:jclass "edu.mit.ll.resourceBroker.resourceManager.IResourceManagerNotification"))
			 (stringclass (java:jclass "java.lang.String"))
			 (method (java:jmethod class "notifyTaskStatus" stringclass stringclass stringclass stringclass stringclass stringclass)))
		(java:jcall method (slot-value interface 'notification-instance) taskID submitterID status description resourceUri resourceName))
	;; handles the case when the jclass is not found
	(error () (mtk-utils:format-time t "error in notify-task-status~%")))
  t)

#+abcl
(defmethod notify-task-status-xml ((interface resource-manager-interface) (xml-string string))
  (handler-case
	  (let* ((class (java:jclass "edu.mit.ll.resourceBroker.resourceManager.IResourceManagerNotification"))
			 (stringclass (java:jclass "java.lang.String"))
			 (method (java:jmethod class "notifyTaskStatus" stringclass)))
		(java:jcall method (slot-value interface 'notification-instance) xml-string))
	;; handles the case when the jclass is not found
	(error () (mtk-utils:format-time t "error in notify-task-status-xml: ~a ~%" xml-string)))
  t)

;; TODO: This method has undefined variables "taskID, submittedIR, status, description, resourceUri, resourceName."
#+abcl
(defmethod dispatch-task ((interface resource-manager-interface) (the-task string))
  (handler-case
	  (let* ((class (java:jclass "edu.mit.ll.resourceBroker.resourceManager.Dispatchable"))
			 (stringclass (java:jclass "java.lang.String"))
			 (method (java:jmethod class "dispatchRequest" stringclass)))
		(java:jcall method (slot-value interface 'notification-instance) the-task))
	;; handles the case when the jclass is not found
	(error () (mtk-utils:format-time t "error in dispatch-task: ~a ~%" the-task)))
  t)

   