;;; -*- Mode:Common-Lisp; Package:LP-MODEL; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :lp-model)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; LP-VARIABLE Constructor                                                  ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (defmethod initialize-instance :after ((var lp-variable) &rest init-args)

;;   ;; Coerce the lower and upper bounds to DOUBLE-FLOAT
;;   (with-slots (lower-bound upper-bound) var
;; 	(setf lower-bound (coerce lower-bound 'double-float))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Functions on variables.                                                  ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro make-variable-and-add-to-lp (lp &key
									   (name nil name-p)
									   (lb nil lb-p) 
									   (ub nil ub-p)
									   (lin-obj-val nil lin-obj-val-p))
  "Constructs a new variable and adds it to the LP formulation. Returns the 
   index of the new variable and a reference to the new variable."

  `(let ((var (make-instance 'lp-variable 
							 ,@(if name-p (list :name name))
							 ,@(if lb-p (list :lower-bound lb))
							 ,@(if ub-p (list :upper-bound ub)))))
	 (values (add-variable-to-lp var ,lp ,@(if lin-obj-val-p (list :lin-obj-val lin-obj-val)))
			 var)))

(defmethod add-variable-to-lp ((var lp-variable) (lp lp) &key (lin-obj-val nil lin-obj-val-p) (auto-set-col-index? T))
  "Add the variable to the specified LP. Returns the index of the added
   variable."
  (with-slots (variables variables-ht) lp
	(let ((var-idx (if auto-set-col-index? 
					   (length variables)
					   (col-index-in-formulation var)))) 
	  (with-slots (name col-index-in-formulation) var
		;; Record this variable's index in the formulation.
		(setf col-index-in-formulation var-idx)
		;; Add this variable to the LP's variables list.
		(push var variables)
		;; Add this variable to the hash table using its name
		(setf (gethash name variables-ht) var))

	  ;; If a linear objective value is given, update the LP objective to reflect it.
	  (when lin-obj-val-p
		(set-lp-linear-objective-idx! lp var-idx lin-obj-val))
	  
	  ;; Return the column index of this variable.
	  var-idx)))

(defmethod (setf linear-part) (val item (var lp-variable))
  "This method determines the index of VAR and calls the appropriate specialized method."
  (call-next-method val item (col-index-in-formulation var)))

(defmethod linear-part (item (var lp-variable))
  "This method determines the index of VAR and calls the appropriate specialized method."
  (call-next-method item (col-index-in-formulation var)))