;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:mtk-utils-2d-geometry)

;;------------------------------------------------------------
;; Tolerance
;;------------------------------------------------------------

(defparameter *geometry-tolerance* 0)

(defun equal-tol (e1 e2 &key (tolerance *geometry-tolerance*))
  "Two numbers are equal if their absolute difference 
is less than or equal to *geometry-tolerance*"
  (assert (numberp e1))
  (assert (numberp e2))
  (if (equalp tolerance 0) 
	  (equalp e1 e2)
	  (<= (abs (- e1 e2)) tolerance))
)

;;------------------------------------------------------------
;; Equality
;;------------------------------------------------------------

(defmethod equal-geo (e1 e2 &key (tolerance *geometry-tolerance*))
  (declare (ignore tolerance))
  (equal e1 e2))

(defmethod equalp-geo (e1 e2 &key (tolerance *geometry-tolerance*))
  (declare (ignore tolerance))
  (equalp e1 e2))

(defmacro pushnew-geo (item place)
  `(pushnew ,item ,place :test #'equalp-geo))