;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David Wang, Seung H. Chung

(in-package mtk-utils-symbol-equality)


(defun symbol-eql (i1 i2) 
  "Two symbols are equal if they have the same SYMBOL-NAME, regardless of
their package"
  (equal (symbol-name i1) (symbol-name i2)))

(defun symbol-equal (i1 i2) 
  "identical to the EQUAL function, but compares
symbols using EQL-SYM"
  (cond
    ;; CASE 1: both are null
    ((and (null i1) (null i2))
     T)
    ;; CASE 2: both are lists
    ((and (listp i1) (listp i2))
     (and (symbol-equal (car i1) (car i2))
	  (symbol-equal (cdr i1) (cdr i2))))
    ;; CASE 3 both are symbols
    ((and (symbolp i1) (symbolp i2))
     (symbol-eql i1 i2))
    ;; CASE 4: 
    (:otherwise
     (equal i1 i2))
    ))


(defun symbol-equalp (i1 i2) 
"identical to the EQUALP function, but compares
symbols using EQL-SYM"	 
  (cond 
    ;; CASE 1: both are null
    ((and (null i1) (null i2))
     T)
    ;; CASE 2: both are lists
    ((and (listp i1) (listp i2))
     (and (symbol-equalp (car i1) (car i2))
	  (symbol-equalp (cdr i1) (cdr i2))))
    ;; CASE 3: both are symbols
    ((and (symbolp i1) (symbolp i2))
     (symbol-eql i1 i2))
    ;; CASE 4: 
    (:otherwise
     (equalp i1 i2))
    ))


