;;;; Copyright (c) 2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David Wang

(in-package #:mtk-utils-threading)

(defclass flag ()
  ((lock 
	:initform (bordeaux-threads:make-recursive-lock) 
	:initarg :lock
	:accessor flag-lock
	:documentation "Lock used to make flag access atomic.")
   (condition 
	:initform (bordeaux-threads:make-condition-variable)
	:initarg :condition
	:accessor flag-condition
	:documentation "The conditional variable used to indicate the flag was set.")
   (value
	:type t
	:initform nil
	:initarg :value
	:accessor flag-value
	:documentation "The boolean value of the flag, nil by default.")
   ))

(defun make-flag (&rest args)
  (apply #'make-instance 'flag args))

(defmethod flag? ((flag flag) &key (wait nil) (wait-value t) (test #'eq))
  "Returns the value of the flag. 
If WAIT is nil, the function will return immediately.
If WAIT is t, the function will block until the TEST function returns
t when comparing two arguments, the WAIT-VALUE and the value of the flag."
  (bordeaux-threads:with-recursive-lock-held ((flag-lock flag))
	(when wait
	  (loop
		 (when (funcall test wait-value (flag-value flag))
		   (return))
		 (bordeaux-threads:condition-wait 
		  (flag-condition flag)
		  (flag-lock flag))))
	(flag-value flag)))

(defmethod flag! ((flag flag) value)
  "Sets the value of the flag. The value is normally true or nil, 
but can be set to any value."
  (bordeaux-threads:with-recursive-lock-held ((flag-lock flag))
	(setf (flag-value flag) value)
	(bordeaux-threads:condition-notify (flag-condition flag))))