;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 2000 - 2009 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :opsat)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Creating Opsat Clause Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Contains functions for creating the different types of Clause
;;; descriptions that are employed by OpSat.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do:
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; o Make the following functions be methods on the search-problem
;;;   instances and generalize method names to apply to
;;;   generic search problems.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface Routines
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-opsat-clause-description (opsat-literals)

  "Given a list of OPSAT-LITERAL descriptions, returns
   an opsat clause description, composed of literals."

  (list :CLAUSE opsat-literals))

(defun make-opsat-state-variable-description (variable-name domain)

  "Given a VARIABLE-NAME and a list of DOMAIN elements, 
   RETURNS an opsat state variable description."

  (list variable-name :DOMAIN domain))

(defun make-opsat-decision-variable-description (variable-name valuation)

  "Given a VARIABLE-NAME and its VALUATION, that is, a list of
   value domain elements,
   RETURNS an opsat decision variable description."

  (list variable-name :DOMAIN valuation))

(defun make-opsat-valued-domain-element (element cost)

  "Given a domain ELEMENT and its associated COST,
   RETURNS an opsat valueddecision domain element description."

  (list element :COST cost))
