;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			new-opsat-problem 
;;;			add-constraint-system!
;;;
;;;			start-sat-constraints-set
;;;			push-sat-constraints
;;;			add-sat-constraints-to-top
;;;			pop-sat-constraints-from-top
;;;			clear-sat-constraints
;;;			
;;;			start-unsat-constraints-set
;;;			push-unsat-constraints
;;;			add-unsat-constraints-to-top
;;;			pop-unsat-constraints-from-top
;;;			clear-unsat-constraints
;;;			)
;;;          :OPSAT)
;;;  )

;;; This file contains code to setup a new opsat problem, and
;;; to associate a set of constraints with the problem, defined
;;; as a constraint system.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Specifying the Interface to the Constraint System

;;; (NEW-OPSAT-PROBLEM
;;;    DECISION-VARIABLE-DESCRIPTIONS
;;;     STATE-VARIABLE-DESCRIPTIONS
;;;     BASE-CONSTRAINT-DESCRIPTIONS
;;;     &OPTIONAL &KEY
;;;			   (CONDITIONAL? NIL)
;;;			   (CONSTRAINT-SYSTEM-TYPE 'THEORY)
;;;			   (SAT-CONSTRAINT-DESCRIPTIONS-LIST 
;;;				'(())
;;;				)
;;;			   (UNSAT-CONSTRAINT-DESCRIPTIONS-LIST 
;;;				:NO-UNSAT-CONSTRAINTS)
;;;			   (SOURCE NIL)
;;;			   (NAME (GENTEMP "OPSAT-PROBLEM-"))
;;;			   
;;;            (CONFLICT-DIRECTED? T)
;;;			   (SEARCH-ORDER :BEST-FIRST)
;;;			   
;;;			   (BINARY-OBJECTIVE-FUNCTION #'+)  
;;;			   (COST-OFFSET 0)
;;;			   (COST-ORDER-FUNCTION #'>=))
;;;  
;;;   Creates and returns a combinatorial optimization problem with 
;;;   propositional state logic constraints.  Defaults to a problem that 
;;;   maximizes the sum of the decision variable assignment costs.  The 
;;;   objective function can be modified by calling add-objective-function!.
;;;
;;;   State variables are finite domain variables that are mentioned 
;;;   in state constraints.
;;;   Decision variables is a subset of the state variables 
;;;   that are being optimized over.
;;;
;;;   STATE-VARIABLE-DESCRIPTIONS is a list:
;;;     (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
;;;   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
;;;    supplied to make-instance of state variable.  It currently consists of
;;;    one required keyword pair :DOMAIN DOMAIN where DOMAIN is a list of values.
;;;   
;;;   DECISION-VARIABLE-DESCRIPTIONS is a list:
;;;    (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)*
;;;   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
;;;    supplied to make-instance of decision variable.  It currently consists of
;;;    one required keyword :DOMAIN and one optional keyword :GUARD:
;;;
;;;      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
;;;       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *true*])
;;;
;;;   where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
;;;    a domain element, or list of the form:
;;;
;;;      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
;;;  
;;;   where DOMAIN-ELEMENT is the value of the assignment, and  
;;;   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
;;;   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
;;;   This currently consists of :COST.
;;;
;;;   :GUARD is optional, and denotes an assignment that
;;;   activates variable.
;;;
;;;   CONDITIONAL?  
;;;     A BOOLEAN.  If true, then decision variables are guarded by assignments, 
;;;     A guard is specified within a variable description by 
;;;       :GUARD '(= <VAR> <VAL>).  
;;;     Use :guard *always-active* (= *true* to specify a variable 
;;;     that is permanently active.
;;;
;;;   The multi-attribute objective function of the optimization problem 
;;;   is defined over the decision variables.  The attribute cost of
;;;   each decision variable value is specified in DECISION-VARIABLE-DESCRIPTIONS.
;;;
;;;   The composition of attribute costs specified jointly by 
;;;   BINARY-OBJECTIVE-FUNCTION, COST-OFFSET and COST-ORDER-FUNCTION. 
;;;   BINARY-OBJECTIVE-FUNCTION specifies how attribute costs are pair-wise
;;;    combined, and defaults to +.  This function must satisfy preferential
;;;    independence.
;;;   COST-OFFSET specifies a constant value to be added to all costs.
;;;   COST-ORDER-FUNCTION places a total order on values, and is used to
;;;    define the best value.  The default, >=, corresponds to a solution
;;;    that maximizes cost (utility).
;;;
;;;   BASE-CONSTRAINT-DESCRIPTIONS is a list of state logic 
;;;    constraint descriptions.  The BASE STATE CONSTRAINTS, should
;;;    be consistent for every decision-variable assignment; the
;;;    modeler violates this rule at his or her own risk.
;;;
;;;   A STATE LOGIC CONSTRAINT DESCRIPTION is a clause, denoted by
;;;   a list of STATE LITERALS.
;;;
;;;   A STATE LITERAL is a literal whose proposition is an assignment 
;;;    (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
;;;    STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
;;;    VARIABLE-NAME's DOMAIN in that description.
;;;
;;;   SAT-CONSTRAINT-DESCRIPTIONS-LIST is either the keyword
;;;   :NO-SAT-CONSTRAINTS, or a list of ONE or more 
;;;   SAT-CONSTRAINT-DESCRIPTIONS, each of which denotes a list of
;;;   ZERO or more constraints.  Each constraint is described by a 
;;;   CONSTRAINT DESCRIPTION, as specified above.  
;;;   If a list is provided, then for each list element SAT-CONSTRAINTS,
;;;   the solutions generated must satisfy the union of BASE-CONSTRAINTS 
;;;   and SAT-CONSTRAINTS.
;;;   Note that this condition will be enforced even when the 
;;;   SAT-CONSTRAINTS list is empty.
;;;
;;;   If keyword :NO-SAT-CONSTRAINTS is provided, then no sat constraint
;;;   is imposed, and the base constraints are not tested for consistency
;;;   alone.  This is equivalent to SAT-CONSTRAINT-DESCRIPTIONS = ().
;;;
;;;   SAT-CONSTRAINT-DESCRIPTIONS-LIST defaults to '(()), which
;;;   impose the condition that the solution satisfy the base constraints.
;;;  
;;;   UNSAT-CONSTRAINT-DESCRIPTIONS is either a list of 0 or 
;;;   more CONSTRAINT-DESCRIPTIONS or the keyword :NO-UNSAT-CONSTRAINTS.
;;;   If a list is provided, then the solution satisfies the condition
;;;   that the union of BASE-CONSTRAINTS and UNSAT-CONSTRAINTS
;;;   is unsatisfiable.  If the keyword :NO-UNSAT-CONSTRAINT, then this is
;;;   equivalent to providing the empty list.
;;;
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;  
;;;   NAME is an sexpression naming the OPSAT problem.
;;;
;;;   CONSTRAINT-SYSTEM-TYPE is an object type with slot NAME and
;;;     initarg :NAME. CONSTRAINT-SYSTEM-TYPE defaults to THEORY.
;;;     CONSTRAINT-SYSTEM-TYPE supports the following methods:
;;;
;;;      It supports the following methods: 
;;;
;;;      NEW-CONSTRAINT is a method that creates a state constraint
;;;       for CONSTRAINT-SYSTEM.  Its arguments are 
;;;       (CONSTRAINT-SYSTEM :SOURCE SOURCE . CONSTRAINT-DESCRIPTION.
;;;
;;;      INSERT-CONSTRAINT is a method that inserts a constraint
;;;       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
;;;       CONSTRAINT-OBJECT as arguments.
;;;
;;;      GET-ASSIGNMENT is a method that takes arguments
;;;      CONSTRAINT-SYSTEM, VARIABLE AND VALUE, and looks up the object 
;;;       representing assignment within CONSTRAINT-SYSTEM.
;;;
;;;      ASSERT! is a method that takes arguments:
;;;        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
;;;           &optional (TRUTH-VALUE *true*) (SOURCE NIL))
;;;       It asserts that ASSIGNMENT 
;;;       has TRUTH-VALUE (*true* or *false*) in CONSTRAINT-SYSTEM, 
;;;       justified by SOURCE.
;;;       It returns nothing useful.
;;;
;;;      UNASSERT! is a method that takes arguments:
;;;        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
;;;          &optional (TRUTH-VALUE *true*)(SOURCE NIL) )
;;;       It removes the assertion that ASSIGNMENT has TRUTH-VALUE 
;;;       (*true* or *false*) in CONSTRAINT-SYSTEM due to justification SOURCE.
;;;       It returns nothing useful.
;;;
;;;      CONSISTENT-WITH? is a method that takes as arguments:
;;;        (CONSTRAINT-SYSTEM &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL) (SOURCE NIL)
;;;          (PERFORM-SEARCH? T) (RETURN-CONFLICT? T))
;;;       It tests the consistency of the current constraints and assignments 
;;;       in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.
;;;
;;;       It returns (VALUES CONSISTENT? CONFLICT),
;;;       where CONSISTENT? is true unless theory is proven inconsistent.  
;;;
;;;       Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;;       Each assignment in ADDED-ASSIGNMENTS is a list:
;;;         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
;;;       and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.
;;;
;;;       It first tests consistency by propagation.
;;;       If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
;;;       continues with splitting.
;;;
;;;       If RETURN-CONFLICT? is TRUE then it returns a subset of the current inserted 
;;;       assignments that led to an inconsistency.
;;;     
;;;      INCONSISTENT-WITH? is a method that takes as arguments:
;;;        (CONSTRAINT-SYSTEM &optional &KEY (added-constraints nil) (source nil)
;;;          (perform-search? T) (return-conflict? T)
;;;          (base-theory-consistent-with-all-assignments? T))
;;;       It tests the INCONSISTENCY of the current constraints and 
;;;       assignments in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS.
;;;
;;;       It RETURNS (values INCONSISTENT? CONFLICT).  INCONSISTENT? is 
;;;       false unless CONSTRAINT-SYSTEM is proven inconsistent.  It first tests 
;;;       consistency by propagation.  Then if CONSTRAINT-SYSTEM is consistent
;;;       by propagation and PERFORM-SEARCH? is TRUE, 
;;;       it continues with splitting.
;;;
;;;       Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;;       Each assignment in ADDED-ASSIGNMENTS is a list:
;;;         (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))
;;;       and specifies that proposition (= variable value) has TRUTH-VALUE.
;;; 
;;;       If RETURN-CONFLICT? is TRUE and CONSTRAINT-SYSTEM is proven necessarily 
;;;       true then it returns a subset of the current 
;;;       decision variable assignments that is sufficient to guarantee that 
;;;       constraint-system is satisfied.
;;;
;;;       If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
;;;       it assumes that the inserted assignments are consistent with the 
;;;       base constraints in CONSTRAINT-SYSTEM (minus added-constraints).
;;;       In this case a conflict is a subset of the decision variable 
;;;       assignments that, together with CONSTRAINT-SYSTEM (minus assignments) entail 
;;;       ADDED-CONSTRAINTS.
;;;
;;;       A conflict is created by searching over a subset of the current 
;;;       constraints that support the truth of all ADDED-CONSTRAINTS, or all 
;;;       non-unit clauses, and collecting the set of all propositions in
;;;       these clauses such that the proposition denotes an assignment to 
;;;       a decision variable, and the proposition's truth assignment was
;;;       made during search by splitting.

;;;   CONFLICT-DIRECTED? is a flag, which defaults to T.  Conflict-directed search
;;;     is performed if and only if this flag is true.

;;;   SEARCH-ORDER specifies the order in which the search tree is
;;;     expolored.  Defaults to :BEST-FIRST.

;;; Incrementally Adding and Removing Sat Constraints

;;; (ADD-CONSTRAINT-SYSTEM! SEARCH-PROBLEM
;;;			    CONSTRAINT-SYSTEM)
;;;
;;;   Adds CONSTRAINT-SYSTEM to SEARCH-PROBLEM, 
;;;   and records all functions used by SEARCH-PROBLEM to
;;;   modify constraints and assignments, and to query
;;;   constraint consistency.
;;;
;;;   SEARCH-PROBLEM is a search-problem object.
;;;
;;;   CONSTRAINT-SYSTEM is a constraint system that is being searched over,
;;;   It supports the following methods: 
;;;
;;;      NEW-CONSTRAINT is a method that creates a state constraint
;;;       for CONSTRAINT-SYSTEM.  Its arguments are 
;;;       (CONSTRAINT-SYSTEM :SOURCE SOURCE . CONSTRAINT-DESCRIPTION.
;;;
;;;      INSERT-CONSTRAINT is a method that inserts a constraint
;;;       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
;;;       CONSTRAINT-OBJECT as arguments.
;;;
;;;      GET-ASSIGNMENT is a method that takes arguments
;;;      THEORY, VARIABLE AND VALUE, and looks up the object 
;;;       representing assignment within CONSTRAINT-SYSTEM.
;;;
;;;      ASSERT! is a method that takes arguments:
;;;        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
;;;           &optional (TRUTH-VALUE *true*) (SOURCE NIL))
;;;       It asserts that ASSIGNMENT 
;;;       has TRUTH-VALUE (*true* or *false*) in CONSTRAINT-SYSTEM, 
;;;       justified by SOURCE.
;;;       It returns nothing useful.
;;;
;;;      UNASSERT! is a method that takes arguments:
;;;        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
;;;          &optional (TRUTH-VALUE *true*)(SOURCE NIL) )
;;;       It removes the assertion that ASSIGNMENT has TRUTH-VALUE 
;;;       (*true* or *false*) in CONSTRAINT-SYSTEM due to justification SOURCE.
;;;       It returns nothing useful.
;;;
;;;      CONSISTENT-WITH? is a method that takes as arguments:
;;;        (THEORY &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL) (SOURCE NIL)
;;;          (PERFORM-SEARCH? T) (RETURN-CONFLICT? T))
;;;       It tests the consistency of the current constraints and assignments 
;;;       in THEORY, conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.
;;;
;;;       It returns (VALUES CONSISTENT? CONFLICT),
;;;       where CONSISTENT? is true unless theory is proven inconsistent.  
;;;   
;;;       Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;;       Each assignment in ADDED-ASSIGNMENTS is a list:
;;;         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
;;;       and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.
;;;   
;;;       It first tests consistency by propagation.
;;;       If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
;;;       continues with a full DPLL-style satisfiability test.  
;;;
;;;       If RETURN-CONFLICT? is TRUE and an inconsistency is found by unit 
;;;       propagation, then it returns a subset of the current inserted 
;;;       assignments that led to an inconsistency.
;;;  
;;;      INCONSISTENT-WITH? is a method that takes as arguments:
;;;        (CONSTRAINT-SYSTEM &optional &KEY (added-constraints nil) (source nil)
;;;          (perform-search? T) (return-conflict? T)
;;;          (base-theory-consistent-with-all-assignments? T))
;;;       It tests the INCONSISTENCY of the current constraints and 
;;;       assignments in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS.
;;;
;;;       It RETURNS (values INCONSISTENT? CONFLICT).  INCONSISTENT? is 
;;;       false unless constraint-system is proven inconsistent.  It first tests 
;;;       consistency by unit propagation.  Then if constraint-system is consistent
;;;       by unit propagation and PERFORM-SEARCH? is TRUE, 
;;;       it continues with a full DPLL-style satisfiability test.  
;;;
;;;       Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;;       Each assignment in ADDED-ASSIGNMENTS is a list:
;;;         (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))
;;;       and specifies that proposition (= variable value) has TRUTH-VALUE.
;;; 
;;;       If RETURN-CONFLICT? is TRUE and constraint-system is proven necessarily 
;;;       true by unit propagation, then it returns a subset of the current 
;;;       decision variable assignments that is sufficient to guarantee that 
;;;       constraint-system is satisfied.
;;;
;;;       If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
;;;       it assumes that the inserted assignments are consistent with the 
;;;       constraints in constraint-system (minus added-constraints).
;;;       In this case a conflict is a subset of the decision variable 
;;;       assignments that, together with CONSTRAINT-SYSTEM (minus assignments) entail
;;;       ADDED-CONSTRAINTS.
;;;
;;;       A conflict is created by searching over a subset of the current 
;;;       constraints that support the truth of all ADDED-CONSTRAINTS, or all 
;;;       non-unit clauses, and collecting the set of all propositions in
;;;       these clauses such that the proposition denotes an assignment to 
;;;       a decision variable, and the proposition's truth assignment was
;;;       made during search by splitting.

;;; Adding and Removing Sat Constraints

;;; (PUSH-SAT-CONSTRAINTS (SEARCH-PROBLEM OPSAT-PROBLEM) 
;;;     SAT-CONSTRAINT-DESCRIPTIONS &OPTIONAL (SOURCE NIL))
;;;
;;;   Adds the constraints described by SAT-CONSTRAINT-DESCRIPTIONS
;;;   as a NEW top set of sat-constraints for SEARCH-PROBLEM,
;;;   justified by SOURCE.
;;;
;;;   Returns the list of internalized constraints corresponding to 
;;;   SAT-CONSTRAINT-DESCRIPTIONS.
;;;
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;
;;;   SAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
;;;   Each constraint description is a list of state-literals, where
;;;   a state literal is a literal whose proposition is an assignment 
;;;   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
;;;   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
;;;   VARIABLE-NAME's DOMAIN in that description.
;;;
;;;   Recall that for each set of SAT-CONSTRAINTS,the solutions generated
;;;   must satisfy the union of BASE-CONSTRAINTS and 
;;;   SAT-CONSTRAINTS.

;;; (POP-SAT-CONSTRAINTS-FROM-TOP SEARCH-PROBLEM)
;;;
;;;   Removes the top set of sat constraints from SEARCH-PROBLEM.
;;;   Returns the internalized list of constraints that have been
;;;   removed.

;;; (CLEAR-SAT-CONSTRAINTS SEARCH-PROBLEM)
;;;
;;;   Removes ALL sets of sat-constraints for SEARCH-PROBLEM.

;;; Building a Sat Constraints Set Via Multiple Calls:

;;; (START-SAT-CONSTRAINTS-SET (SEARCH-PROBLEM OPSAT-PROBLEM) 
;;;      &OPTIONAL (SOURCE NIL))
;;;
;;;   Starts a NEW top set of sat-constraints for SEARCH-PROBLEM.
;;;   This set contains no constraints.
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;   RETURNS the newly created set.

;;; (ADD-SAT-CONSTRAINTS-TO-TOP (SEARCH-PROBLEM OPSAT-PROBLEM) 
;;;	    SAT-CONSTRAINT-DESCRIPTIONS &OPTIONAL (SOURCE NIL))
;;;
;;;   Adds the constraints described by SAT-CONSTRAINT-DESCRIPTIONS
;;;   to the EXISTING top set of sat-constraints for SEARCH-PROBLEM,
;;;   justified by SOURCE.
;;;
;;;   Returns the complete, top set of internalized sat constraints.
;;;
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;
;;;   SAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
;;;   Each constraint description is a list of state-literals, where
;;;   a state literal is a literal whose proposition is an assignment 
;;;   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
;;;   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
;;;   VARIABLE-NAME's DOMAIN in that description.
;;;
;;;   Recall that for each set of SAT-CONSTRAINTS,the solutions generated
;;;   must satisfy the union of BASE-CONSTRAINTS and 
;;;   SAT-CONSTRAINTS.


;;; Adding and Removing UnSat Constraints

;;; (START-UNSAT-CONSTRAINTS-SET SEARCH-PROBLEM &OPTIONAL (SOURCE NIL))
;;;
;;;   Starts a NEW top set of Unsat-constraints for SEARCH-PROBLEM,
;;;   This set contains no constraints.
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;   Returns the newly created set.

;;; (PUSH-UNSAT-CONSTRAINTS (SEARCH-PROBLEM OPSAT-PROBLEM)
;;;	    UNSAT-CONSTRAINT-DESCRIPTIONS &OPTIONAL (SOURCE NIL))
;;;
;;;   Adds the constraints described by UNSAT-CONSTRAINT-DESCRIPTIONS
;;;   as a NEW top set of unsat-constraints for SEARCH-PROBLEM,
;;;   justified by SOURCE.
;;;
;;;   Returns the list of internalized constraints corresponding to 
;;;   UNSAT-CONSTRAINT-DESCRIPTIONS.
;;;
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;
;;;   UNSAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
;;;   Each constraint description is a list of state-literals, where
;;;   a state literal is a literal whose proposition is an assignment 
;;;   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
;;;   STATE VARIABLE DESCRIPTIONS, and VALUE must be an element of
;;;   VARIABLE-NAME's DOMAIN in that description.
;;;
;;;   Recall that for each set of UNSAT-CONSTRAINTS,the solutions 
;;;   generated must inconsistent with the union of BASE-CONSTRAINTS 
;;;   and UNSAT-CONSTRAINTS.

;;; (ADD-UNSAT-CONSTRAINTS-TO-TOP (SEARCH-PROBLEM OPSAT-PROBLEM)
;;;	   UNSAT-CONSTRAINT-DESCRIPTIONS &OPTIONAL (SOURCE NIL))
;;;
;;;   Adds the constraints described by UNSAT-CONSTRAINT-DESCRIPTIONS
;;;   to the EXISTING top set of unsat-constraints for SEARCH-PROBLEM,
;;;   justified by SOURCE.
;;;
;;;   Returns the complete, top set of internalized unsat constraints.
;;;
;;;   SOURCE is an object that invoked the creation of OPSAT problem.
;;;
;;;   UNSAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
;;;   Each constraint description is a list of state-literals, where
;;;   a state literal is a literal whose proposition is an assignment 
;;;   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
;;;   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
;;;   VARIABLE-NAME's DOMAIN in that description.
;;;
;;;   Recall that for each set of UNSAT-CONSTRAINTS,the solutions 
;;;   generated must inconsistent with the union of BASE-CONSTRAINTS 
;;;   and UNSAT-CONSTRAINTS.

;;; (POP-UNSAT-CONSTRAINTS-FROM-TOP SEARCH-PROBLEM)
;;;
;;;   Removes the top set of unsat constraints from SEARCH-PROBLEM.
;;;   Returns the internalized list of constraints that have been
;;;   removed.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o Allow variables and/or constraints to be removed, and for
;;;   variable values to be permanently fixed.
;;;   - If a constraint can be activated and deactivated,
;;;     Make this explicit, and ensure that the conflict is conditioned
;;;     on the activation variable.
;;;   - If a variable's value is fixed, then remove any clause
;;;     or conflict that is permanently satisfied.
;;;   - If we allow a variable to be removed, then remove 
;;;     all dependent conflicts and constraints.

;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their to do.

;;;  o Distinguish and document all interface functions.
;;;  o Indicate supporting functions clearly.
;;;  o Update terminology to be consistent with
;;;    the JDAM journal article.
;;;  o Update terminology to be consistent with manipulating
;;;    constraints, rather than theories and clauses.
;;;  o Add entailment-constraint as an option.
;;;  o Shorten keyword names.
;;;  o Check other terminology against JDAM paper.
;;;  o Cleanly, separate problem of generating kernel assignments 
;;;    versus generating full assignments.
;;;  o Add in code for systematic (irredundant search).
;;;    ** isn't it already systematic?
;;;       -- gustavop
;;;  o Add in code for generating multiple full assignments.

;;; Fix:
;;; o In new-opsat-problem, the addition of tpn-start-event-name, below is incorrect.
;;;   Start-event is a property specific to OPITC problems, not
;;;   OPSAT problems in general.  OPITC problems should not have
;;;   TPN specific names.  OPITC is independent of TPN.
;;;   Fix this for make-instance of opsat-problem, as well as for
;;;   new-opsat-problem.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface Function Definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o The name cost should probably be changed to objective.

;;;
;;; Creating an Opsat Problem
;;;

;;; To Do:
;;; o Replace calls to new-opsat-problem with calls to make-instance.

;;; o Allow an arbitrary set of initialization parameters to be passed
;;;   in for constraint-system.

;;; o Generalize to arbitrary candidate generators (BFIG),
;;;   similar to our generalization to arbitrary constraint systems.

;;;   - pass in as an argument a candidate generator type.
;;;     - Type must support certain interface methods,
;;;       update these in the documentation.

;;; Changed:
;;; o Updated documentation for state/decision-variable-descriptions.
;;; o Added Fix and To Do:

;;; Fix:
;;; o :conditional? should be added as an optional keyword, 
;;;   similar to new-informed-search-problem.  This should
;;;   determine the type of problem object created.

;;; To Do:
;;; o There is redundant information in the state and decision variable
;;;   descriptions.  For the subset of variables that are decision variables,
;;;   generate the state variable specifications automatically.

;;; Changed:
;;; o Added CONDITIONAL? keyword.

;;; Changed:
;;; o Added CONFLICT-DIRECTED? keyword.
;;; -- gustavop.

;;; To do:
;;; o Opsat-problem now just performs Constraint-based A*.
;;;   Conflict-directed-opsat-problem performs CDA*. Need to update documentation?
;;; o Define conditional-csp w/o conflict-direction.
;;; -- gustavop

;;; Fix:
;;; o The addition of tpn-start-event-name, below is incorrect.
;;;   Start-event is a property specific to OPITC problems, not
;;;   OPSAT problems in general.  OPITC problems should not have
;;;   TPN specific names.  OPITC is independent of TPN.
;;;   Fix this for make-instance of opsat-problem, as well as for
;;;   new-opsat-problem.
(defun new-opsat-problem
    (decision-variable-descriptions
     state-variable-descriptions
     base-constraint-descriptions
     &optional &key
	 (conditional? nil)
	 ;; Defaults to testing consistency on
	 ;; the base constraints alone.
	 (constraint-system-type 'theory)
	 
	 ;; Added by Peng Yu; Define a new slot in opsat-problem, 
	 ;; called tpn-start-event-name, to store the start event name 
	 ;; of an ITC problem.
	 ;; Default to be nil so that it won't affect the existing opsat 
	 ;; implementation.
	 (start-event-name 'start)
	 (sat-constraint-descriptions-list 
	  '(())
	  )
	 (unsat-constraint-descriptions-list 
	  :no-unsat-constraints)

	 (conflict-directed? T)
	 (search-order :best-first)

	 (allow-relaxation? nil)
	 
	 (binary-objective-function #'+)  
	 (cost-offset 0)
	 (cost-order-function #'>=)
	 
	 (source nil)
	 (name (gentemp "opsat-problem-")))
  
  "Creates and returns a combinatorial optimization problem with 
   propositional state logic constraints.  Defaults to a problem that 
   maximizes the sum of the decision variable assignment costs.  The 
   objective function can be modified by calling add-objective-function!.

   State variables are finite domain variables that are mentioned 
   in state constraints.
   Decision variables is a subset of the state variables 
   that are being optimized over.

   STATE-VARIABLE-DESCRIPTIONS is a list:
     (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of state variable.  It currently consists of
    one required keyword pair :DOMAIN DOMAIN where DOMAIN is a list of values.
   
   DECISION-VARIABLE-DESCRIPTIONS is a list:
    (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)*
   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of decision variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *true*])

   where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
   This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable.

   CONDITIONAL?  
     A BOOLEAN.  If true, then decision variables are guarded by assignments, 
     A guard is specified within a variable description by 
       :GUARD '(= <VAR> <VAL>).  
     Use :guard *always-active* (= *true* to specify a variable 
     that is permanently active.

   The multi-attribute objective function of the optimization problem 
   is defined over the decision variables.  The attribute cost of
   each decision variable value is specified in DECISION-VARIABLE-DESCRIPTIONS.

   The composition of attribute costs specified jointly by 
   BINARY-OBJECTIVE-FUNCTION, COST-OFFSET and COST-ORDER-FUNCTION. 
   BINARY-OBJECTIVE-FUNCTION specifies how attribute costs are pair-wise
    combined, and defaults to +.  This function must satisfy preferential
    independence.
   COST-OFFSET specifies a constant value to be added to all costs.
   COST-ORDER-FUNCTION places a total order on values, and is used to
    define the best value.  The default, >=, corresponds to a solution
    that maximizes cost (utility).

   BASE-CONSTRAINT-DESCRIPTIONS is a list of state logic 
    constraint descriptions.  The BASE STATE CONSTRAINTS, should
    be consistent for every decision-variable assignment; the
    modeler violates this rule at his or her own risk.

   A STATE LOGIC CONSTRAINT DESCRIPTION is a disjunctive clause, 
   denoted by a list of STATE LITERALS.

   A STATE LITERAL is a literal whose proposition is an assignment 
    (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
    STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
    VARIABLE-NAME's DOMAIN in that description.

   SAT-CONSTRAINT-DESCRIPTIONS-LIST is either the keyword
   :NO-SAT-CONSTRAINTS, or a list of ONE or more 
   SAT-CONSTRAINT-DESCRIPTIONS, each of which denotes a list of
   ZERO or more constraints.  Each constraint is described by a 
   CONSTRAINT DESCRIPTION, as specified above.  
   If a list is provided, then for each list element SAT-CONSTRAINTS,
   the solutions generated must satisfy the union of BASE-CONSTRAINTS 
   and SAT-CONSTRAINTS.
   Note that this condition will be enforced even when the 
   SAT-CONSTRAINTS list is empty.

   If keyword :NO-SAT-CONSTRAINTS is provided, then no sat constraint
   is imposed, and the base constraints are not tested for consistency
   alone.  This is equivalent to SAT-CONSTRAINT-DESCRIPTIONS = ().

   SAT-CONSTRAINT-DESCRIPTIONS-LIST defaults to '(()), which
   impose the condition that the solution satisfy the base constraints.
  
   UNSAT-CONSTRAINT-DESCRIPTIONS is either a list of 0 or 
   more CONSTRAINT-DESCRIPTIONS or the keyword :NO-UNSAT-CONSTRAINTS.
   If a list is provided, then the solution satisfies the condition
   that the union of BASE-CONSTRAINTS and UNSAT-CONSTRAINTS
   is unsatisfiable.  If the keyword :NO-UNSAT-CONSTRAINT, then this is
   equivalent to providing the empty list.

   SOURCE is an object that invoked the creation of OPSAT problem.
  
   NAME is an sexpression naming the OPSAT problem.

   CONSTRAINT-SYSTEM-TYPE is an object type with slot NAME and
     initarg :NAME. CONSTRAINT-SYSTEM-TYPE defaults to THEORY.
     CONSTRAINT-SYSTEM-TYPE supports the following methods:

      It supports the following methods: 

      NEW-CONSTRAINT is a method that creates a state constraint
       for CONSTRAINT-SYSTEM.  Its arguments are 
       (CONSTRAINT-SYSTEM :SOURCE SOURCE . CONSTRAINT-DESCRIPTION.

      INSERT-CONSTRAINT is a method that inserts a constraint
       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
       CONSTRAINT-OBJECT as arguments.

      GET-ASSIGNMENT is a method that takes arguments
      CONSTRAINT-SYSTEM, VARIABLE AND VALUE, and looks up the object 
       representing assignment within CONSTRAINT-SYSTEM.

      ASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
           &optional (TRUTH-VALUE *true*) (SOURCE NIL))
       It asserts that ASSIGNMENT 
       has TRUTH-VALUE (*true* or *false*) in CONSTRAINT-SYSTEM, 
       justified by SOURCE.
       It returns nothing useful.

      UNASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
          &optional (TRUTH-VALUE *true*)(SOURCE NIL) )
       It removes the assertion that ASSIGNMENT has TRUTH-VALUE 
       (*true* or *false*) in CONSTRAINT-SYSTEM due to justification SOURCE.
       It returns nothing useful.

      CONSISTENT-WITH? is a method that takes as arguments:
        (constraint-system &optional &key (added-constraints nil) 						        (added-assignments nil)(source nil)
          (perform-search? t) (return-conflict? T))
       It tests the consistency of the current constraints and assignments 
       in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.

       It returns (VALUES CONSISTENT? CONFLICT),
       where CONSISTENT? is true unless theory is proven inconsistent.  

       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.

       It first tests consistency by propagation.
       If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
       continues with splitting.

       If RETURN-CONFLICT? is TRUE then it returns a subset of the current inserted 
       assignments that led to an inconsistency.
     
      INCONSISTENT-WITH? is a method that takes as arguments:
        (constraint-system &optional &key (added-constraints nil) 
          (added-assignments nil)(source nil)
          (perform-search? T) (return-conflict? T)
          (base-theory-consistent-with-all-assignments? T))
       It tests the INCONSISTENCY of the current constraints and 
       assignments in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS.

       It RETURNS (values INCONSISTENT? CONFLICT).  INCONSISTENT? is 
       false unless CONSTRAINT-SYSTEM is proven inconsistent.  It first tests 
       consistency by propagation.  Then if CONSTRAINT-SYSTEM is consistent
       by propagation and PERFORM-SEARCH? is TRUE, 
       it continues with splitting.

       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= variable value) has TRUTH-VALUE.
 
       If RETURN-CONFLICT? is TRUE and CONSTRAINT-SYSTEM is proven necessarily 
       true then it returns a subset of the current 
       decision variable assignments that is sufficient to guarantee that 
       constraint-system is satisfied.

       If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
       it assumes that the inserted assignments are consistent with the 
       base constraints in CONSTRAINT-SYSTEM (minus added-constraints).
       In this case a conflict is a subset of the decision variable 
       assignments that, together with CONSTRAINT-SYSTEM (minus assignments) entail 
       ADDED-CONSTRAINTS.

       A conflict is created by searching over a subset of the current 
       constraints that support the truth of all ADDED-CONSTRAINTS, or all 
       non-unit clauses, and collecting the set of all propositions in
       these clauses such that the proposition denotes an assignment to 
       a decision variable, and the proposition's truth assignment was
       made during search by splitting.

   CONFLICT-DIRECTED? is a flag, which defaults to T.  Conflict-directed search
     is performed if and only if this flag is true.

   SEARCH-ORDER specifies the order in which the search tree is
     expolored.  Defaults to :BEST-FIRST.
       "

  (let ((problem-type 
		 (if conditional?
			 
			 'conditional-opsat-problem
			 
			 (if conflict-directed?
				 
				 'conflict-directed-opsat-problem

				 'opsat-problem))))


	(if allow-relaxation? 

		(setf problem-type 'conflict-directed-relaxation-problem))

	(make-instance problem-type
	  :decision-variable-descriptions decision-variable-descriptions
	  :state-variable-descriptions state-variable-descriptions
	  :base-constraint-descriptions base-constraint-descriptions
	  :constraint-system-type constraint-system-type
	  :start-event-name start-event-name
	  :sat-constraint-descriptions-list sat-constraint-descriptions-list
	  :unsat-constraint-descriptions-list unsat-constraint-descriptions-list
	  :search-order search-order
	  :binary-objective-function binary-objective-function  
	  :cost-offset cost-offset
	  :cost-order-function cost-order-function
	  ;;	  :branch-and-bound? t
	  :name name
	  :source source
	  )))

(defmethod add-constraint-system! ((search-problem opsat-problem)
								   constraint-system)

  "Adds CONSTRAINT-SYSTEM to SEARCH-PROBLEM.

   SEARCH-PROBLEM is a search-problem object.

   CONSTRAINT-SYSTEM is a constraint system that is being searched over,
   It supports the following methods: 

      NEW-CONSTRAINT is a method that creates a state constraint
       for CONSTRAINT-SYSTEM.  Its arguments are 
          (CONSTRAINT-SYSTEM :SOURCE SOURCE . CONSTRAINT-DESCRIPTION.

      INSERT-CONSTRAINT is a method that inserts a constraint
       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
       CONSTRAINT-OBJECT as arguments.

      GET-ASSIGNMENT is a method that takes arguments
      THEORY, VARIABLE AND VALUE, and looks up the object representing 
       assignment within CONSTRAINT-SYSTEM.

      ASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
           &optional (TRUTH-VALUE *true*) (SOURCE NIL))
       It asserts that ASSIGNMENT 
       has TRUTH-VALUE (*true* or *false*) in CONSTRAINT-SYSTEM, 
       justified by SOURCE.
       It returns nothing useful.

      UNASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
          &optional (TRUTH-VALUE *true*)(SOURCE NIL) )
       It removes the assertion that ASSIGNMENT has TRUTH-VALUE 
       (*true* or *false*) in CONSTRAINT-SYSTEM due to justification SOURCE.
       It returns nothing useful.

      CONSISTENT-WITH? is a method that takes as arguments:
        (CONSTRAINT-SYSTEM &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL) 
          (ADDED-ASSIGNMENTS NIL)(SOURCE NIL)
          (PERFORM-SEARCH? T) (RETURN-CONFLICT? T))
       It tests the consistency of the current constraints and assignments 
       in CONSTRAINT-SYSTEM, conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.

       It returns (VALUES CONSISTENT? CONFLICT),
       where CONSISTENT? is true unless constraint-system is proven inconsistent.  
   
       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.
   
       It first tests consistency by propagation.
       If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
       continues with a full DPLL-style satisfiability test.  

       If RETURN-CONFLICT? is TRUE and an inconsistency is found by unit 
       propagation, then it returns a subset of the current inserted 
       assignments that led to an inconsistency.
  
      INCONSISTENT-WITH? is a method that takes as arguments:
        (constraint-system &optional &KEY (added-constraints nil) 
          (ADDED-ASSIGNMENTS NIL)(source nil)
          (perform-search? T) (return-conflict? T)
          (base-theory-consistent-with-all-assignments? T))
       It tests the INCONSISTENCY of the current constraints and 
       assignments in constraint-system, conjoined with ADDED-CONSTRAINTS.

       It RETURNS (values INCONSISTENT? CONFLICT).  INCONSISTENT? is 
       false unless constraint-system is proven inconsistent.  It first tests 
       consistency by unit propagation.  Then if constraint-system is consistent
       by unit propagation and PERFORM-SEARCH? is TRUE, 
       it continues with a full DPLL-style satisfiability test.  

       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= variable value) has TRUTH-VALUE.
 
       If RETURN-CONFLICT? is TRUE and constraint-system is proven necessarily 
       true by unit propagation, then it returns a subset of the current 
       decision variable assignments that is sufficient to guarantee that 
       constraint-system is satisfied.

       If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
       it assumes that the inserted assignments are consistent with the 
       constraints in constraint-system (minus added-constraints).
       In this case a conflict is a subset of the decision variable assignments 
       that, together with CONSTRAINT-SYSTEM (minus assignments) entail ADDED-CONSTRAINTS.

       A conflict is created by searching over a subset of the current 
       constraints that support the truth of all ADDED-CONSTRAINTS, or all 
       non-unit clauses, and collecting the set of all propositions in
       these (disjunctive) clauses such that the proposition denotes an assignment to 
       a decision variable, and the proposition's truth assignment was
       made during search by splitting."
  
  ;; Simply adds all information provided by the arguments to SEARCH-PROBLEM.
  (setf (opsat-problem-constraint-system search-problem)
    constraint-system)

  ;; Creates a hash-table used to map from decision variable assignments 
  ;; as encoded by INFORMED SEARCH, to their equivalent assignment as
  ;; encoded by the CONSTRAINT SYSTEM.
  (setf (opsat-problem-convert-constraint-to-candgen-assignment-ht 
		 search-problem)
    (make-hash-table :test #'eql))

  T)


;;;
;;; Incrementally Adding and Removing Sat Constraints
;;;

;;; To Do:  
;;; o Check callers.  If none expect a returned value,
;;    then add ! to its name.
(defmethod start-sat-constraints-set ((search-problem opsat-problem)
									  &optional (source nil))

  "Starts a NEW top set of sat-constraints for SEARCH-PROBLEM,
   This set contains no constraints.

   SOURCE is an object that invoked the creation of OPSAT problem.

   Returns the newly created set."

  (push-sat-constraints search-problem '() source))

(defmethod push-sat-constraints ((search-problem opsat-problem)
								 sat-constraint-descriptions
								 &optional (source nil))

  "Adds the constraints described by SAT-CONSTRAINT-DESCRIPTIONS
   as a NEW top set of sat-constraints for SEARCH-PROBLEM,
   justified by SOURCE.

   Returns the list of internalized constraints corresponding to 
   SAT-CONSTRAINT-DESCRIPTIONS.

   SOURCE is an object that invoked the creation of OPSAT problem.

   SAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
   Each constraint description is a list of state-literals, denoting
   a (disjunctive) clause, where
   a state literal is a literal whose proposition is an assignment 
   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
   VARIABLE-NAME's DOMAIN in that description.

   Recall that for each set of SAT-CONSTRAINTS,the solutions generated 
   must satisfy the union of BASE-CONSTRAINTS and SAT-CONSTRAINTS."

  (let ((interned-constraints 
		 (internalize-constraint-descriptions 
		  search-problem source sat-constraint-descriptions))
		(existing-constraint-list
		 (opsat-problem-sat-constraints-list search-problem)))

	(cond ((or (null existing-constraint-list)
			   (eql existing-constraint-list 
					:no-sat-constraints))
		
		   ;; No sets of sat constraints have been added yet,
		   ;; Add constraints as the first set.
		   (setf (opsat-problem-sat-constraints-list search-problem)
			 (list interned-constraints)))

		  (:otherwise

		   ;; One or more sets of sat constraints exist,
		   ;; add constraints to the top (most recent) set.
		   (push interned-constraints
				 (opsat-problem-sat-constraints-list search-problem))))
	
	interned-constraints))

(defmethod add-sat-constraints-to-top ((search-problem opsat-problem)
									   sat-constraint-descriptions
									   &optional (source nil))

  "Adds the constraints described by SAT-CONSTRAINT-DESCRIPTIONS
   to the EXISTING top set of sat-constraints for SEARCH-PROBLEM,
   justified by SOURCE.

   Returns the complete, top set of internalized sat constraints.

   SOURCE is an object that invoked the creation of OPSAT problem.

   SAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
   Each constraint description is a list of state-literals, denoting
   a (disjunctive) clause, where
   a state literal is a literal whose proposition is an assignment 
   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
   VARIABLE-NAME's DOMAIN in that description.

   Recall that for each set of SAT-CONSTRAINTS,the solutions generated 
   must satisfy the union of BASE-CONSTRAINTS and SAT-CONSTRAINTS."
  
  (let ((interned-constraints 
		 (internalize-constraint-descriptions 
		  search-problem source sat-constraint-descriptions))
		(existing-constraint-list
		 (opsat-problem-sat-constraints-list search-problem)))

	(cond ((or (null existing-constraint-list)
			   (eql existing-constraint-list 
					:no-sat-constraints))

		   ;; No sets of sat constraints have been added yet,
		   ;; Add constraints as the first set.
		   (setf (opsat-problem-sat-constraints-list search-problem)
			 (list interned-constraints))

		   ;; Return the top constraint set.
		   interned-constraints)
		   
		  (:otherwise

		   ;; One or more sets of sat constraints exist,
		   ;; add constraints to the top (most recent) set,
		   ;; and return the augmented top set.
		   (setf (first existing-constraint-list)
			 (nconc interned-constraints
					(first existing-constraint-list)))))))

;;; FIX:
;;; o This will not work correctly.  It can leave in place conflicts
;;;   that are no longer valid.
(defmethod pop-sat-constraints-from-top ((search-problem opsat-problem))

  "Removes the top set of sat constraints from SEARCH-PROBLEM.
   Returns the internalized list of constraints that have been
   removed."

  (error "pop-sat-constraints-from-top will not work correctly.  ~
          It will leave in place conflicts that are no longer valid.")
  
  (pop (opsat-problem-sat-constraints-list search-problem)))

;;; FIX:
;;; o CLEAR-SAT-CONSTRAINTS will not behave correctly, because it will leave in 
;;;   place conflicts that were derived from the current data set.  This is
;;;   also the case for clear-unsat-constraints.  Three
;;;   possible repairs are:
;;;      - Decache all conflicts, and start learning them again.
;;;      - Have the conflicts conditioned on the observables, or more generally
;;;        any variables that activate and deactive constraints.
;;;      - Decache conflicts (and constraints) selectively when these 
;;;        "activation" variables are permanently removed or permanently fixed 
;;;        to a value.

;;; To Do:  
;;; o Check callers.  If none expect a returned value,
;;    then add ! to its name.
(defmethod clear-sat-constraints ((search-problem opsat-problem))

  "Removes ALL sets of sat-constraints for SEARCH-PROBLEM,
   setting its value to :NO-SAT-CONSTRAINTS.
   Returns :NO-SAT-CONSTRAINTS."
  (error "Deprecated becaues of conflict learning.")
  (setf (opsat-problem-sat-constraints-list search-problem)
	:no-sat-constraints))

;;;
;;; Incrementally Adding and Removing UnSat Constraints
;;;

(defmethod start-unsat-constraints-set ((search-problem opsat-problem)
										&optional (source nil))

  "Starts a NEW top set of unsat-constraints for SEARCH-PROBLEM,
   This new set contains no constraints.

   SOURCE is an object that invoked the creation of OPSAT problem.

   Returns the newly created set of unsat-constraints."

  (push-unsat-constraints '() search-problem source))

(defmethod push-unsat-constraints ((search-problem opsat-problem)
								   unsat-constraint-descriptions
								   &optional (source nil))

  "Adds the constraints described by UNSAT-CONSTRAINT-DESCRIPTIONS
   as a NEW top set of unsat-constraints for SEARCH-PROBLEM,
   justified by SOURCE.

   Returns the list of internalized constraints corresponding to 
   UNSAT-CONSTRAINT-DESCRIPTIONS.

   SOURCE is an object that invoked the creation of OPSAT problem.

   UNSAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
   Each CONSTRAINT-DESCRIPTION is a list of STATE-LITERALs, denoting
   a (disjunctive) clause, where
   a STATE LITERAL is a literal whose PROPOSITION is an assignment 
   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
   STATE VARIABLE DESCRIPTIONS, and VALUE must be an element of
   VARIABLE-NAME's DOMAIN in that description.

   Recall that for each set of UNSAT-CONSTRAINTS,the solutions 
   generated must inconsistent with the union of BASE-CONSTRAINTS 
   and UNSAT-CONSTRAINTS."

  (let ((interned-constraints 
		 (internalize-constraint-descriptions 
		  search-problem source UNsat-constraint-descriptions))
		(existing-constraint-list
		 (opsat-problem-unsat-constraints-list search-problem)))

	(cond ((or (null existing-constraint-list)
			   (eql existing-constraint-list 
					:no-unsat-constraints))
		
		   ;; No sets of sat constraints have been added yet,
		   ;; Add constraints as the first set.
		   (setf (opsat-problem-unsat-constraints-list search-problem)
			 (list interned-constraints)))

		  (:otherwise

		   ;; One or more sets of sat constraints exist,
		   ;; add constraints to the top (most recent) set.
		   (push interned-constraints
				 (opsat-problem-unsat-constraints-list
				  search-problem))))
	
	interned-constraints))

(defmethod add-unsat-constraints-to-top ((search-problem opsat-problem)
										 unsat-constraint-descriptions
										 &optional (source nil))

  "Adds the constraints described by UNSAT-CONSTRAINT-DESCRIPTIONS
   to the EXISTING top set of unsat-constraints-list for SEARCH-PROBLEM,
   justified by SOURCE.

   Returns the complete, top set of internalized unsat constraints.

   SOURCE is an object that invoked the creation of OPSAT problem.

   UNSAT-CONSTRAINT-DESCRIPTIONS is a list of constraint descriptions.
   Each constraint description is a list of state-literals, denoting
   a (disjunctive) clause, where
   a state literal is a literal whose proposition is an assignment 
   (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
   STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
   VARIABLE-NAME's DOMAIN in that description.

   Recall that for each set of UNSAT-CONSTRAINTS,the solutions 
   generated must inconsistent with the union of BASE-CONSTRAINTS 
   and UNSAT-CONSTRAINTS."
  
  (let ((interned-constraints 
		 (internalize-constraint-descriptions 
		  search-problem source unsat-constraint-descriptions))
		(existing-constraint-list
		 (opsat-problem-unsat-constraints-list search-problem)))

	(cond ((or (null existing-constraint-list)
			   (eql existing-constraint-list 
					:no-unsat-constraints))

		   ;; No sets of sat constraints have been added yet,
		   ;; Add constraints as the first set.
		   (setf (opsat-problem-unsat-constraints-list search-problem)
			 (list interned-constraints))

		   ;; Return the top constraint set.
		   interned-constraints)
		   
		  (:otherwise

		   ;; One or more sets of sat constraints exist,
		   ;; add constraints to the top (most recent) set,
		   ;; and return the augmented top set.
		   (setf (first existing-constraint-list)
			 (nconc interned-constraints
					(first existing-constraint-list)))))))

;;; FIX:
;;; o This will not work correctly.  It can leave in place conflicts
;;;   that are no longer valid.
(defmethod pop-unsat-constraints-from-top ((search-problem opsat-problem))

  "Removes the top set of unsat constraints from SEARCH-PROBLEM.
   Returns the internalized list of constraints that have been
   removed."
  
  (error "pop-unsat-constraints-from-top will not work correctly.  ~
          It will leave in place conflicts that are no longer valid.")
  (pop (opsat-problem-unsat-constraints-list search-problem)))


;;; FIX:
;;; o CLEAR-SAT-CONSTRAINTS will not behave correctly, because it will leave in 
;;;   place conflicts that were derived from the current data set.  This is
;;;   also the case for clear-unsat-constraints.  Three
;;;   possible repairs are:
;;;      - Decache all conflicts, and start learning them again.
;;;      - Have the conflicts conditioned on the observables, or more generally
;;;        any variables that activate and deactive constraints.
;;;      - Decache conflicts (and constraints) selectively when these 
;;;        "activation" variables are permanently removed or permanently fixed 
;;;        to a value.

;;; To Do:  
;;; o Check callers.  If none expect a returned value,
;;    then add ! to its name.
(defmethod clear-unsat-constraints ((search-problem opsat-problem))

  "Removes ALL sets of unsat-constraints for SEARCH-PROBLEM,
   setting its value to :NO-UNSAT-CONSTRAINTS.
   Returns :NO-UNSAT-CONSTRAINTS."
  (error "Deprecated becaues of conflict learning.")
  (setf (opsat-problem-unsat-constraints-list search-problem)
	:no-unsat-constraints))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; o Check that trace is setup properly, it needs to be doubly linked.

;;; To Do:
;;; o Rename to be consistent with other names.  This should be 
;;;   SETUP-OPTIMAL-CSP, not OPSAT PROBLEM.
;;; o The following should be performed when make-instance is called.

;;; Fix:
;;; Delete the commented out code when this works.

(defmethod setup-opsat-problem! ((search-problem opsat-problem))

  "Given a new SEARCH-PROBLEM, appropriately initialize
   and return it.  NAME is an sexpression denoting the name 
   of SEARCH-PROBLEM, and SOURCE is an sexpression denoting
   the object that created search problem.
   SEARCH-ORDER is the order in which search-tree nodes is expanded,
   and is one of the keywords in *ALLOWED-SEARCH-ORDER-TYPES*."

  ;; Setup the data structures for tracing search.  This
  ;; is used to maintain statistics and perform debugging.
  (setf (opsat-problem-state-trace search-problem)
    (make-instance 'state-trace :search-problem search-problem))
  T)

;;; To Do:
;;; o Extend define-state-variable to accept a keywor list.
;;;   and extend to permit conditional state-variables.
;;; o Should this be a method associated with search-problem
;;;   or constraint system, or not a method?

;;; Changed:
;;; o modified state-variable-descriptions to include variable-keyword-list.
(defun add-state-variables! 
	(constraint-system state-variable-descriptions source)
  
  " Introduce state variables into OPSAT CONSTRAINT-SYSTEM.
    STATE-VARIABLE-DESCRIPTIONS is a list:
     (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
    where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of state variable.  It currently consists of
    one required keyword pair :DOMAIN DOMAIN where DOMAIN is a list of values."
  ;; o Define the STATE VARIABLES and add constraints to
  ;;   CONSTRAINT-SYSTEM that declare that each variable 
  ;;   takes on exactly one value in its respective domain.

  (loop for (variable-name . variable-keyword-list) in state-variable-descriptions
	  for domain = (getf variable-keyword-list :domain)
	  do (define-state-variable constraint-system source variable-name
								domain))
  T)

;;; Change:
;;; o update documentation and name VARIABLE-KEYWORD-LIST.
(defmethod add-decision-variables!
	((search-problem opsat-problem) 
	 decision-variable-descriptions
	 source)
  
  "Introduce the decision variables into OPSAT SEARCH-PROBLEM.
   DECISION-VARIABLE-DESCRIPTIONS is a list:
    (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)*
   where 

   VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *true*])

    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE 
   class.  This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable. "
  
  (declare (ignore source))

  ;; o Introduce the DECISION VARIABLES.  The value in each 
  ;;   domain has an associated cost.

  (loop for (variable-name . variable-keyword-list) 
	  in decision-variable-descriptions
	  do (add-decision-variable search-problem variable-name variable-keyword-list))
  T)

;;; FIX:  
;;; o Check does the order of values matter?

;;; Changed:
;;; o updated documentation.
;;; o Changed value-description to VARIABLE-KEYWORD-LIST
(defmethod add-decision-variable ((search-problem opsat-problem)
								  variable-name variable-keyword-plist)
  
  "Creates a new decision variable with the specified name.

   VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *true*])

    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
   This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable. "
  
  (let ((variable
		 (add-variable search-problem variable-name variable-keyword-plist)))

	;; Creates hash table entries mapping variable's assignments in the constraint
	;; system to variable's assignments in the opsat-search-problem.
    (let ((convert-cs-to-candgen-ht
		   (opsat-problem-convert-constraint-to-candgen-assignment-ht
			search-problem))
		  
		  (constraint-system (opsat-problem-constraint-system
							  search-problem)))
	  
      (loop 
		;; For each assignment in candgen, find its mate in constraint system.
		  for state-assignment in (state-variable-domain variable)
		  for value = (state-assignment-value state-assignment)
		  for constraint-assignment = (get-assignment constraint-system
													  variable-name
													  value)
									  
		  do 
		   (when (typep constraint-assignment 'isat::proposition)
			 (setf  (isat::proposition-searchable constraint-assignment) nil))
			;; Setup two way mapping.
			(setf (get-assignment-from-hash constraint-assignment
											convert-cs-to-candgen-ht)
			  state-assignment)
		   
			(setf (state-assignment-to-constraint-assignment state-assignment)
			  constraint-assignment)))
    
    variable))

(defmethod add-base-constraints! ((search-problem opsat-problem)
								  base-constraint-descriptions
								  source)
  
  "Insert the BASE CONSTRAINTS in CONSTRAINT SYSTEM."
  
  ;;   Create and associate the base constraints.
  (setf (opsat-problem-base-constraints search-problem)
		(internalize-constraint-descriptions search-problem
											 source
											 base-constraint-descriptions))
  

(print (opsat-problem-base-constraints search-problem))
  ;;   Insert the base constraints.
  (loop for base-constraint
	 in (opsat-problem-base-constraints search-problem)
	 do (insert-constraint-into-constraint-system
		 search-problem base-constraint))
  T)

(defmethod add-sat-constraints! ((search-problem opsat-problem) 
								 sat-constraint-descriptions-list
								 source)
      
  "Create and associate the SAT constraints with constraint-system."
  
  (setf (opsat-problem-sat-constraints-list search-problem)
	(if (or (null sat-constraint-descriptions-list)
			(eql sat-constraint-descriptions-list 
				 :no-sat-constraints))
		  
		;; :no-sat-constraints is equivalent to '().
		'()
		
	  (loop for constraint-descriptions
		  in sat-constraint-descriptions-list
		  collecting (internalize-constraint-descriptions 
					  search-problem source constraint-descriptions))))

  T)

(defmethod add-unsat-constraints! 
	((search-problem opsat-problem)
	 unsat-constraint-descriptions-list
	 source)
      
  "Create and associate the UNSAT constraints with constraint-system."
  
  (setf (opsat-problem-unsat-constraints-list search-problem)
	(if (or (null unsat-constraint-descriptions-list)
			(eql unsat-constraint-descriptions-list :no-unsat-constraints))

		;; :no-unsat-constraints is equivalent to '().
		'()

	  (loop for unconstraint-descriptions
		  in unsat-constraint-descriptions-list
		  collecting (internalize-constraint-descriptions 
					  search-problem source unconstraint-descriptions))))
  T)

(defmethod internalize-constraint-descriptions 
	((search-problem opsat-problem)
	 source
	 constraint-descriptions)

  "Given a list of CONSTRAINT-DESCRIPTIONS,
   associates with the constraint-system of SEARCH-PROBLEM
   a set of constraints described by CONSTRAINT-DESCRIPTIONS,
   with SOURCE being the reason for association.

   RETURNS a corresponding list of constraint object for
   constraint-system, in the same order as 
   CONSTRAINT-DESCRIPTIONS.
   
   Does not assert the ocnstraints in constraint-system,
   this is performed through a separate function."

  (unless (null constraint-descriptions)
	(loop for constraint-description in constraint-descriptions
		collecting (new-constraint-for-constraint-system
					search-problem
					source
					constraint-description))))



;;; TO-DO: define standard way to create search-problem
;;; right now the function below defined is similar to new-search-problem
;;; with one aditional key-word argument
;;; *Eliminate duplicated code.

(defun new-vcsp-problem
    (decision-variable-descriptions
     state-variable-descriptions
     base-constraint-descriptions
     &optional &key
			   ;(conditional? nil)
			   ;; Defaults to testing consistency on
			   ;; the base constraints alone.
			   (constraint-system-type 'theory)
			   (sat-constraint-descriptions-list 
				'(())
				)
			   (UNsat-constraint-descriptions-list 
				:no-unsat-constraints)
			   (source nil)
			   (name (gentemp "opsat-problem-"))

	                   ;(conflict-directed? T)
			   
			   (search-order :best-first)
			   
			   (binary-objective-function #'+)  
			   (cost-offset 0)
			   (cost-order-function #'>=)
	                   (soft-constraint-system nil))
  
  "Same as new-opsat-problem with additional keyword argument "

	(make-instance 'conflict-directed-opsat-vcsp
	  :name name
	  :decision-variable-descriptions decision-variable-descriptions
	  :state-variable-descriptions state-variable-descriptions
	  :base-constraint-descriptions base-constraint-descriptions
	  :constraint-system-type constraint-system-type
	  :sat-constraint-descriptions-list sat-constraint-descriptions-list
	  :unsat-constraint-descriptions-list unsat-constraint-descriptions-list
	  :source source
	  :search-order search-order
	  :binary-objective-function binary-objective-function  
	  :cost-offset cost-offset
	  :cost-order-function cost-order-function
	  :soft-constraint-system soft-constraint-system))