;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;            initialize-search-problem!
;;;            find-solutions
;;;            find-next-solutions
;;;            )
;;;          :OPSAT)
;;;  )

;;; This file contains the OpSat search engine that performs a best first 
;;; search to find a leading set of states that satisfy a set of constraints.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;  (INITIALIZE-SEARCH-PROBLEM!
;;;    (SEARCH-PROBLEM SEARCH-PROBLEM)
;;;    &key (SEARCH-ORDER :DEFAULT))
;;;  
;;;   Clears the current search state of SEARCH PROBLEM.
;;;   This includes initializing the candidate generator,
;;;   clearing the constraint system, decaching all recorded conflicts
;;;   and solutions, initializing debugging trace information, and
;;;   setting search back to the start of the state space.
;;;
;;;   Search-order specifies the order in which candidates are generated.
;;;   Must be one of *ALLOWED-SEARCH-ORDER-TYPES* (for example 
;;;   :BEST-FIRST), or :DEFAULT.  If :DEFAULT, then
;;;   search-order is the last one specified.

;;;  (FIND-SOLUTIONS
;;;		(SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*)
;;;     &key
;;;		(CUTOFF-DELTA nil)
;;;     (MAXIMUM-SOLUTION-COUNT 10)
;;;     (MAXIMUM-VISITED-STATE-COUNT NIL)
;;;		(PERFORM-SEARCH-TO-TEST-CONSTRAINTS? T)
;;;     (SEARCH-ORDER :DEFAULT)
;;;		(FIND-ALL-CONSISTENT-MODELS? NIL))

;;;   Starting from the begining of the search space, finds the 
;;;   leading assignments to decision variables that satisfy
;;;   constraints of SEARCH-PROBLEM and have the best cost.
;;;
;;;   Each returned solution is a list of assignments, where
;;;   each assignment is an assignment object of SEARCH-PROBLEM's
;;;   constraint system, denoting a decision variable assignment.
;;;
;;;   If flag find-all-consistent-models? is set, 
;;;   then all consistent models for each solution is generated
;;;   and recorded on the search-object for problem.
;;;
;;;   Generates the sequence of leading best solutions up to 
;;;   MAXIMUM-SOLUTION-COUNT. MAXIMUM-SOLUTION-COUNT defaults to nil, 
;;;   which represents no bound.  
;;;
;;;   Search-order specifies the order in which search-tree nodes
;;;   are expanded.  It is either :DEFAULT or one of *ALLOWED-SEARCH-ORDER-TYPES*.
;;;   If :DEFAULT is specified, then defaults to the search-order specified
;;;   when search-problem was created.
;;;
;;;   Search terminates earlier if the MAXIMUM-VISITED-STATE-COUNT is exceeded 
;;;   or if computed cutoff cost is reached.  
;;;   MAXIMUM-VISITED-STATE-COUNT defaults to 0.  
;;;  
;;;   To determine cutoff cost, if the best first search is in decreasing 
;;;   (increasing) order, the resulting cutoff cost is the least (greatest) 
;;;   cost considered.  The cutoff cost is computed from CUTOFF-DELTA 
;;;   using function COMPUTE-CUTOFF-COST.  
;;;  
;;;   CUTOFF-DELTA specifies a maximum allowed difference (drop off) 
;;;   between the cost of the best solution found and the current 
;;;   search state, or no cutoff, if nil.
;;;   CUTOFF-DELTA defaults to nil.
;;;
;;;   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE, then it searches over
;;;   constraint assignments to see if SAT and UNSAT constraints are
;;;   satisfied.  If nil, then it checks the constraints by performing
;;;   constraint propagation alone.
;;;  
;;;   ASSUMPTIONS:
;;;    - The constraint system has already been setup with all 
;;;      relevant constraints inserted, such as observables, 
;;;      control values and models.
;;;    - Any irrelevant assertions or constraints have been removed
;;;      before invocation.
;;;
;;;   SIDE-EFFECTS:
;;;    - Stores leading solutions in SEARCH-PROBLEM-SOLUTIONS
;;;      as opsat-solution objects.
;;;    - Upon termination, removes any assignments it introduced into 
;;;      CONSTRAINT-SYSTEM during search."

;;; (FIND-NEXT-SOLUTIONS (SEARCH-PROBLEM OPSAT-PROBLEM)
;;;	 &OPTIONAL (CUTOFF-DELTA NIL)
;;;			   (MAXIMUM-SOLUTION-COUNT 10)
;;;			   (MAXIMUM-VISITED-STATE-COUNT NIL)
;;;			   (PERFORM-SEARCH-TO-TEST-CONSTRAINTS? T)
;;;			   (FIND-ALL-CONSISTENT-MODELS? NIL))
;;;
;;;   CONTINUES a previously initiated search, but otherwise, 
;;;   similar to FIND-SOLUTION.  Finds a sequence of 
;;;   decision variable assignments that satisfy
;;;   constraints of SEARCH-PROBLEM, in the order dictated by
;;;   the type of search; 
;;;
;;;   Each solution returned is a list of assignments, where
;;;   each assignment is an assignment object of SEARCH-PROBLEM's
;;;   constraint system, denoting a decision variable assignment.

;;;   ASSUMPTIONS: 
;;;      See find-solutions.

;;;   SIDE-EFFECTS: 
;;;      See find-solutions.

;;; (FIND-SOLUTION-OBJECTS (SEARCH-PROBLEM OPSAT-PROBLEM)
;;;								  &KEY (CUTOFF-DELTA NIL)
;;;									   (MAXIMUM-SOLUTION-COUNT 10)
;;;									   (MAXIMUM-VISITED-STATE-COUNT NIL)
;;;									   (PERFORM-SEARCH-TO-TEST-CONSTRAINTS? T)
;;;									   (SEARCH-ORDER :DEFAULT)
;;;									   (FIND-ALL-CONSISTENT-MODELS? NIL))
;;;  
;;;    Starting from the begining of the search space, finds the 
;;;    leading assignments to decision variables that satisfy
;;;    constraints of SEARCH-PROBLEM and have the best cost.
;;;
;;;   Same as find-solutions but returns the opsat-solution objects for each
;;;   solution, rather than the assignments of the solution.
;;;
;;;   Each opsat-solution object includes:
;;;   o list of assignments, where each ASSIGNMENT is an assignment 
;;;   object of SEARCH-PROBLEM's constraint system, denoting a decision 
;;;   variable assignment,
;;;   o the cost of the solution (when an optimization problem), and
;;;   o when the flag find-all-consistent-models? is set,
;;;	    a list of all models for that solution,
;;;     where a model for an opsat-solution is a full-assignment,
;;;	    that extends the solution's decision variable assignment,
;;;     and satisfies the problem's constraints.

;;;   ASSUMPTIONS: 
;;;      See find-solutions.

;;;   SIDE-EFFECTS: 
;;;      See find-solutions.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; FIX:

;;; o Between tests of successive candidates, 
;;;   OPSAT inserts and removes observations.
;;;   The net effect is to not exploit the common
;;;   inferences between the two cases.  Fix this.
;;;   - Patrick Conrad says that this is fixed in their
;;;     branch of Opsat.  Merge and confirm.

;;; Check:

;;; o Make sure that ISat gives to OpSat only conflicts that involve
;;;   decision variables, not other state variables
;;;   During testing, test1 returned a conflict:
;;;    ((= <(= (OUTFLOW NEWVALVE) POSITIVE)> TRUE)
;;;                        (= <(= (MODE NEWVALVE) OPEN)> :TRUE))
;;;   that involved the non-decision variable Outflow.  This
;;;   may have resulted from a bug in ISAT that has been removed.

;;; o Fn convert-constraint-conflict-to-candgen-conflict filters away any
;;;   assignment in a conflict that doesn't involve a decision
;;;   variable or that is assigned false.  Isn't it an error for
;;;   any such literal to appear?  If so, then an error should be
;;;   flagged, rather than filtering.

;;; Optimization:
;;; o When switching between the current and next assignment does this
;;;   needlessly remove and re-add assignments that are unchanged?
;;;   Likewise, does the maintenance of a complete previous and next
;;;   state waste time, particularly if they usually differ by one
;;;   assignment?

;;; o OpSAT adds and removes sat-constraints and unsat
;;;   constraints after each candidate test.  If only one set of
;;;   constraints is being tested against, then it is more efficient to
;;;   leave them in.

;;; To Do:
;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their local "To Do" comments.

;;; o Change the names of functions to be consistent with the pseudo-code
;;;   in the JDAM paper.

;;; o After variables are given to ISAT (or OPSAT), constraints have been 
;;;   inserted into the constraint system that represent the fact that 
;;;   variable domains are mutually exclusive and collectively exhaustive.  
;;;   This might be redundant information, check this out.  This may be
;;;   an ISAT extension, not an OPSAT extension.

;;; Extensions:

;;; o Extend conflict extraction to leverage the mutual exclusion of
;;;   variable assignments, and the exhaustive pruning of a set of child
;;;   nodes (these two points may not be equivalent).  Is this already
;;;   implemented?

;;; o Augment this to perform Branch and Bound search, as an
;;;   alternative to best first search.  Is this already
;;;   implemented?

;;; o Cleanly separate the generation of kernel assignments 
;;;   (as implicants) from full assignments.

;;; o Augment to Clause-directed A*, similar to Rob Ragno's SM thesis.

;;; o Generalize attribute costs to valued constraints over multiple 
;;;   variables.  Is this now incorporated via Gustavo and MaoKai's work?

;;; o Augment to perform tree decomposition and AND/OR search, 
;;;   similar to joint work with Martin Sachenbacher.
;;;   To what degree is this now incorporated by Henri Badaro's work?

;;; o Add in conditional variables, similar to Bobby Effinger's thesis.
;;;   This has been added in for Op-ITC problem.  
;;;   Has it been completed for all search algorithms?

;;; o Incorporate a better measure of heuristic cost.

;;; o Add in the ability to handle temporal constraints
;;;   and state constraints simultaneously.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-search-problem!
    ((search-problem opsat-problem) &optional 
	 &key (search-order :default))
  
  "Clears the current search state of SEARCH PROBLEM.
   This includes initializing the candidate generator,
   clearing the constraint system, decaching all recorded conflicts
   and solutions, initializing debugging trace information, and
   setting search back to the start of the state space.

   Search-order specifies the order in which candidates are generated.
   Must be one of *ALLOWED-SEARCH-ORDER-TYPES* (for example 
   :BEST-FIRST), or :DEFAULT.  If :DEFAULT, then
   search-order is the last one specified."

  ;; Get rid of information about found solutions.
  (setf (search-problem-solutions search-problem) '())
  (setf (search-problem-best-solution-cost search-problem) nil)
  (setf (search-problem-solution-count search-problem) 0)

  ;; Get rid of any discovered conflicts.
  (if (typep search-problem 'state-space-w-conflicts)
      (remove-all-conflicts! search-problem))
  
  ;; Reset the candidate generator.
  (initialize-candidate-generator! search-problem 
								   :search-order search-order)
  
  ;; Clear the constraint system of the current search state.
  (clear-current-state-from-constraints! search-problem)
  
  ;; Clear the debugging information.
  (initialize-state-trace! search-problem)
  T)

;;; To Do:

;;; o Do we want to eliminate subsumed solutions?  Document and 
;;;   pass flag in as a control parameter, rather than hiding it 
;;;   in the global state.  Perhaps separate this from find-solutions.

;;; o Simplify cutoff costs, based on what is actually used.

;;; o Does state-trace require a separate structure?

;;; o Do we want to clean up any changes we've made to constraint system 
;;;   before returning the results of this function? Or do we want to pass 
;;;   in a flag that specifies whether to clean up?

;;; o Add a tracing flag that determines if tracing information should be 
;;;   recorded.

(defmethod find-solutions ((search-problem opsat-problem)
						   &key
						   (cutoff-delta nil)
						   (maximum-solution-count 10)
						   (maximum-visited-state-count nil)
						   (perform-search-to-test-constraints? T)
						   (search-order :DEFAULT)
						   (find-all-consistent-models? nil))
  
  "Starting from the begining of the search space, finds the 
   leading assignments to decision variables that satisfy
   constraints of SEARCH-PROBLEM and have the best cost.

   Each returned solution is a list of assignments, where
   each assignment is an assignment object of SEARCH-PROBLEM's
   constraint system, denoting a decision variable assignment.

   If flag find-all-consistent-models? is set, 
   then all	consistent models for each solution is generated
   and record on the search-object for problem.

   Generates the sequence of leading best solutions up to 
   MAXIMUM-SOLUTION-COUNT. MAXIMUM-SOLUTION-COUNT defaults to nil, 
   which represents no bound.  

   Search-order specifies the order in which search-tree nodes
   are expanded.  Its :DEFAULT or one of *ALLOWED-SEARCH-ORDER-TYPES*.
   If :DEFAULT is specified, then defaults the the search-order specified
   when search-problem was created.

   Search terminates earlier if the MAXIMUM-VISITED-STATE-COUNT is exceeded 
   or if computed cutoff cost is reached.  
   MAXIMUM-VISITED-STATE-COUNT defaults to 0.  
  
   To determine cutoff cost, if the best first search is in decreasing 
   (increasing) order, the resulting cutoff cost is the least (greatest) 
   cost considered.  The cutoff cost is computed from CUTOFF-DELTA 
   using function COMPUTE-CUTOFF-COST.  
  
   CUTOFF-DELTA specifies a maximum allowed difference (drop off) 
   between the cost of the best solution found and the current 
   search state, or no cutoff, if nil.
   CUTOFF-DELTA defaults to nil.

   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE then it searches over
   constraint assignments to see if SAT and UNSAT constraints are
   satisfied.  If nil, then it checks the constraints by performing
   constraint propagation alone.
  
   ASSUMPTIONS:
    - The constraint system has already been setup with all 
      relevant constraints inserted, such as observables, 
      control values and models.
    - Any irrelevant assertions or constraints have been removed
      before invocation.

   SIDE-EFFECTS:
    - Stores leading solutions in SEARCH-PROBLEM-SOLUTIONS
      as opsat-solution objects.
    - Upon termination, removes any assignments it introduced into 
      CONSTRAINT-SYSTEM during search."
  
  (initialize-search-problem! search-problem
							  :search-order search-order)

  ;;(format t "Search Problem: ~A~%" (opsat-problem-constraint-system search-problem))
  ;;find-next-solution-objects returns opsat-solution objects.
  ;; the following loop is used to map each object to
  ;; a decision variable assignment.
  (let ((solution-objects (find-next-solution-objects search-problem
													  cutoff-delta
													  maximum-solution-count
													  maximum-visited-state-count
													  perform-search-to-test-constraints?
													  find-all-consistent-models?)))
	
	(loop for opsat-solution in solution-objects
		collect (opsat-solution-decision-assignments opsat-solution) into solutions
		finally (return solutions))))

;;; To Do:
;;; o This function is ill named, since it returns more than just costs,
;;;   and has been renamed to find-solution-objects.  
;;;    o Change the name used by all callers and then delete this function.
(defmethod find-solutions-w-costs ((search-problem opsat-problem)
								   &key (cutoff-delta nil)
										(maximum-solution-count 10)
										(maximum-visited-state-count nil)
										(perform-search-to-test-constraints? T)
										(search-order :DEFAULT)
										(find-all-consistent-models? nil))
  
  " Name deprecated, see find-solution-objects."
  
  (find-solution-objects search-problem
						 cutoff-delta
						 maximum-solution-count
						 maximum-visited-state-count
						 perform-search-to-test-constraints?
						 search-order
						 find-all-consistent-models?))

;;; Note:
;;;   o This method RETURNS a list of zero or more best solutions
;;;   found, where each solution is an object of type opsat-solution,
;;;   which encapsulates all information that might be needed
;;;   regarding a solution.
(defmethod find-solution-objects ((search-problem opsat-problem)
								  &key (cutoff-delta nil)
									   (maximum-solution-count 10)
									   (maximum-visited-state-count nil)
									   (perform-search-to-test-constraints? T)
									   (search-order :DEFAULT)
									   (find-all-consistent-models? nil))
  
  " Starting from the begining of the search space, finds the 
    leading assignments to decision variables that satisfy
    constraints of SEARCH-PROBLEM and have the best cost.

   Same as find-solutions but returns the opsat-solution objects for each
   solution, rather than the assignments of the solution.

   Each opsat-solution object includes:
   o list of assignments, where each ASSIGNMENT is an assignment 
   object of SEARCH-PROBLEM's constraint system, denoting a decision 
   variable assignment,
   o the cost of the solution (when an optimization problem), and
   o when the flag find-all-consistent-models? is set,
	 a list of all models of to that solution,
     where a model for an opsat-solution is a full-assignment,
	 that extends the solution's decision variable assignment,
     and satisfies the problem's constraints."
  
  (initialize-search-problem! search-problem
							  :search-order search-order)
  
  (find-next-solution-objects search-problem
							  cutoff-delta
							  maximum-solution-count
							  maximum-visited-state-count
							  perform-search-to-test-constraints?
							  find-all-consistent-models?))

;;; FIX:
;;; o Changed so that each solution is a list of assignments, 
;;;   rather than a solution object.  Check and correct callers, 
;;;   so that they use find-next-solution-objects instead, if needed.

;;; Optimization:
;;; o Should trace-search-state only be called if a debugging
;;;   flag has been set?  Also make sure that test functions
;;;   use this trace information effectively.

;;; o When generating a single solution, experiment with
;;;   systematicity, to determine if it improves or degrades
;;;   performance.
(defmethod find-next-solutions

	((search-problem opsat-problem)
	 &optional (cutoff-delta nil)
			   (maximum-solution-count 10)
			   (maximum-visited-state-count nil)
			   (perform-search-to-test-constraints? T)
			   (find-all-consistent-models? nil))

  "CONTINUES a previously initiated search, but otherwise, 
   similar to FIND-SOLUTION.  Finds a sequence of 
   decision variable assignments that satisfy
   constraints of SEARCH-PROBLEM, in the order dictated by
   the type of search; 

   Each solution returned is a list of assignments, where
   each assignment is an assignment object of SEARCH-PROBLEM's
   constraint system, denoting a decision variable assignment."
  
  (let ((solution-objects (find-next-solution-objects search-problem
													  cutoff-delta
													  maximum-solution-count
													  maximum-visited-state-count
													  perform-search-to-test-constraints?
													  find-all-consistent-models?)))
	
	(loop for opsat-solution in solution-objects
		collect (opsat-solution-decision-assignments opsat-solution) into solutions
		finally (return solutions))))


;;; Fix:
;;; o Changed name 
;;;     find-next-solutions -> find-next-solution-objects
;;;   Correct the relevant callers accordingly.
(defmethod find-next-solution-objects 
	((search-problem opsat-problem)
	 &optional (cutoff-delta nil)
			   (maximum-solution-count 10)
			   (maximum-visited-state-count nil)
			   (perform-search-to-test-constraints? T)
			   (find-all-consistent-models? nil))

  "CONTINUES a previously initiated search, but otherwise, 
   similar to FIND-SOLUTION-OBJECTS.  Finds a sequence of 
   decision variable assignments that satisfy
   constraints of SEARCH-PROBLEM, in the order dictated by
   the type of search; 

   Returns a list of opsat-solution objects, where each object includes:

   o list of assignments, where each ASSIGNMENT is an assignment 
   object of SEARCH-PROBLEM's constraint system, denoting a decision 
   variable assignment,
   o the cost of the solution (when an optimization problem), and
   o when the flag FIND-ALL-CONSISTENT-MODELS? is set,
	 a list of all models of to that solution,
     where a model for an opsat-solution is a full-assignment,
	 that extends the solution's decision variable assignment,
     and satisfies the problem's constraints."

  (unless (or (null maximum-solution-count)
			  (and (numberp maximum-solution-count)
				   (>= maximum-solution-count 0)))
	(error "~A is an invalid maximum-solution-count." maximum-solution-count))
  
  (loop
	 with multiple-solutions? = (or (null maximum-solution-count)
									(> maximum-solution-count 1))
	 with count-of-found-solutions = 0
	 with solution = nil
	 with found? = nil
	 with opsat-solution = nil
	 with cutoff-cost = nil
	   
	 ;; Stop if we've found the maximum number of solutions required.
	 until (and (not (null maximum-solution-count))
				(>= count-of-found-solutions maximum-solution-count))
	   
	 ;; Compute the cutoff cost only once, 
	 ;; right after a best solution has been found.			
	 do (when (null cutoff-cost)
		  (setq cutoff-cost (compute-cutoff-cost search-problem cutoff-delta)))
	   
	 ;; Get the next solution. (Reverted to BCW's original version. -Maokai)
	   (multiple-value-setq (opsat-solution found?)
		 (next-solution search-problem
						cutoff-cost
						maximum-visited-state-count
						perform-search-to-test-constraints?
						multiple-solutions?
						:find-all-consistent-models? find-all-consistent-models?))
	   
	 ;;(format t "OPSAT-Solution: ~A~%" opsat-solution)
	   
	 ;; Terminate if candidates exhausted and no solution found (by next-solution).
	 until (not found?)
	   
	 ;; When a solution is found, 
	 ;;   add it to the founds solutions of search-problem, and...
	 ;;   collect it into solutions.
	 collect opsat-solution into solutions

	 ;; do (break "in find-next-solution-objects")

	 do (setf solution (opsat-solution-decision-assignments opsat-solution))						  
	 do (new-solution! solution search-problem)
	   (incf count-of-found-solutions)

	 ;; Record tracing information of states visited.
	   (trace-search-state 
		(opsat-problem-state-trace search-problem)
		solution
		:accepted)

	 finally

	 ;; Finally, 
	 ;;  Remove those assignments that were most recently 
	 ;;  inserted into CONSTRAINT SYSTEM 
	 ;; (the most recent candidate tested) and
	   (clear-current-state-from-constraints! search-problem)

	 ;;  Return the solutions in the order found (from best down).
	   (return solutions)))



(defmethod find-minimal-relaxations ((search-problem opsat-problem)
						   &key
						   (cutoff-delta nil)
						   (maximum-solution-count 10)
						   (maximum-visited-state-count nil)
						   (perform-search-to-test-constraints? T)
						   (search-order :DEFAULT)
						   (find-all-consistent-models? nil))
  
  "Generating minimal solutions to the search problem using BCDR.
  
   This has similar capabilities compared to the find-solutions function.
   The only difference is that it generated minimal solutions/conflicts during the search.
   The input problem format is the same (opITC with conditional variable and constraints).

"

  (initialize-search-problem! search-problem
							  :search-order search-order)


  (let ((solution-objects (find-next-minimal-relaxation-objects search-problem
													  cutoff-delta
													  maximum-solution-count
													  maximum-visited-state-count
													  perform-search-to-test-constraints?
													  find-all-consistent-models?)))

	(format t "All minimal relaxations  ~A~%" (BFIG::search-problem-feasible-nodes search-problem))

	(loop for minimal-relaxation in (BFIG::search-problem-feasible-nodes search-problem)

		 do (format t "~A   " (BFIG::search-node-ancestor-assignments minimal-relaxation))
		 do (format t "expanded from   ~A  " (BFIG::search-node-source-minimal-relaxations 
												minimal-relaxation))

		 do (format t "with cost  ~A~%"  (search-node-best-cost minimal-relaxation)))
	
	(loop for opsat-solution in solution-objects
		collect (opsat-solution-decision-assignments opsat-solution) into solutions
		finally (return solutions))))



(defmethod find-next-minimal-relaxation-objects 
	((search-problem opsat-problem)
	 &optional (cutoff-delta nil)
			   (maximum-solution-count 10)
			   (maximum-visited-state-count nil)
			   (perform-search-to-test-constraints? T)
			   (find-all-consistent-models? nil))

  "Generate Minimal Solution Objects"

  (declare (ignore cutoff-delta))

  (unless (or (null maximum-solution-count)
			  (and (numberp maximum-solution-count)
				   (>= maximum-solution-count 0)))
	(error "~A is an invalid maximum-solution-count." maximum-solution-count))
  
  (loop
      with multiple-solutions? = (or (null maximum-solution-count)
									 (> maximum-solution-count 1))
      with count-of-found-solutions = 0
      with solution = nil
      with found? = nil
      with opsat-solution = nil
	  with cutoff-cost = nil
			     
						 ;; Stop if we've found the maximum number of solutions required.
      until (and (not (null maximum-solution-count))
				 (>= count-of-found-solutions maximum-solution-count))
        
			;; Compute the cutoff cost only once, 
			;; right after a best solution has been found.			
	  ;; do (when (null cutoff-cost)
	  ;; 	   (setq cutoff-cost (compute-cutoff-cost search-problem cutoff-delta)))
		 
		 ;; Get the next solution. (Reverted to BCW's original version. -Maokai)
		do (multiple-value-setq (opsat-solution found?)
		   (next-minimal-relaxation search-problem
								  cutoff-cost
								  maximum-visited-state-count
								  perform-search-to-test-constraints?
								  multiple-solutions?
								  :find-all-consistent-models? find-all-consistent-models?))
	   
	 ;;(format t "OPSAT-Solution: ~A~%" opsat-solution)
				
		 ;; Terminate if candidates exhausted and no solution found (by next-solution).
	  until (not found?)
        
			;; When a solution is found, 
			;;   add it to the founds solutions of search-problem, and...
			;;   collect it into solutions.
	  collect opsat-solution into solutions

	   ;; do (break "in find-next-solution-objects")

	  do (setf solution (opsat-solution-decision-assignments opsat-solution))						  
      do (new-solution! solution search-problem)
		 (incf count-of-found-solutions)

		 ;; Record tracing information of states visited.
		 ;; (trace-search-state 
		 ;;  (opsat-problem-state-trace search-problem)
		 ;;  solution
		 ;;  :accepted)

	  finally

		;; Finally, 
		 ;;  Remove those assignments that were most recently 
		 ;;  inserted into CONSTRAINT SYSTEM 
		 ;; (the most recent candidate tested) and
		(clear-current-state-from-constraints! search-problem)

		 ;;  Return the solutions in the order found (from best down).
		(return solutions)))


    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Finding the Next Best (Consistent) Solution
;;;
;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o Extend search so that given a kernel, it searches in best 
;;;   first order through the set of ALL states associated with that 
;;;   kernel (similar to the JDAM paper and C++ implementation)
;;;   (isn't this implemented and this comment obsolete?).

;;; o This search routine should exploit exhaustivity of decision
;;;   variables to generate new conflicts, when all child nodes are pruned.
;;;   This is similar to the JDAM paper and C++ implementation,
;;;   and similar to algorithms like Conflict-directed Backjumping.

;;; FIX:
;;; o In the case that the current assignment is the conflict, return
;;;   :CURRENT.  Look at the old code, to determine where this originates.

;;; o Test on first-solution-found? only is correct for Best-first
;;;   search.  Rethink for the case where we are performing Branch and Bound.

;;; o The call to bfig::* violates the defined interfaces.

;;; Changed:
;;; o This method now returns:
;;;   - an opsat-solution object (if found) -- look at (defclass opsat-solution..)
;;;   - whether or not a solution has been found (T or nil)

;;; Changed:
;;; o Added keyword FIND-ALL-CONSISTENT-MODELS? which defaults to nil.
;;;   If FIND-ALL-CONSISTENT-MODELS? is true, then all models (i.e. all
;;;   feasible extensions to the consistent candidate are generated.

;;; Check:
;;; o This used to generate and test only full assignments.
;;;   However, it looks like it has been modified to generate partial
;;;   assignments, (see comment" ;; No - these should be clauses.")
;;;   These partial assignments are tested for consistency, but only
;;;   full assignments are recognized as solutions.
;;;   - Why was this change made?  
;;;   - Will it create problems in terms of systematically generating
;;;     all solutions?
;;;    - Does testing partial assignments for consistency have any
;;;      computational benefit?
;;;    - Will testing partial assignments work correctly for the unsat
;;;      constraints and unsat conflicts?
(defmethod next-solution ((search-problem opsat-problem)
						  cutoff-cost 
						  maximum-visited-state-count
						  perform-search-to-test-constraints?
						  multiple-solutions?
						  &key (find-all-consistent-models? nil))

  " Searches in best-first order the leading unexplored candidates,
    until either a solution is found, the candidates are exhausted or
    a cutoff has been reached.  A solution is the next best
    assignment to decision variables that satisfies the problem constraints.

    RETURNS two values: 
     1) the next best OPSAT-SOLUTION, if found, and 
     2) whether or not a solution has been found (T or NIL).

   where OPSAT-SOLUTION is an object of type opsat-solution,
   and encapsulates properties of each solution, including:
   o list of assignments, where each ASSIGNMENT is an assignment 
   object of SEARCH-PROBLEM's constraint system, denoting a decision 
   variable assignment.
   o the cost of the solution (when an optimization problem) and,
   o when the flag find-all-consistent-models? is set, all
	 a list of all models corresponding to that solution,
     where a model for an opsat-solution is a full-assignment,
	 that extends the solution's decision variable assignment,
     and is consistent with respect to base+sat-description;
	 (where sat-description is an element from
	 sat-constraint-descriptions-list).

   SOLUTION is a list of assignments, where
   each ASSIGNMENT is an assignment object of SEARCH-PROBLEM's
   constraint system, denoting a decision variable assignment.
  
   Search terminates when the set of search states is exhausted,
   CUTOFF-COST is reached, or the MAXIMUM-VISITED-STATE-COUNT is reached.

   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE, then search over
     constraint assignments to see if SAT and UNSAT constraints are
     satisfiable.  Otherwise, perform constraint propagation alone.

   If MULTIPLE-SOLUTIONS? is true, then 2 or more solutions are
      desired, otherwise, at most one solution will be generated.

   If FIND-ALL-CONSISTENT-MODELS? is true, then all models (i.e. all
   feasible extensions to the consistent candidate are generated.
  
   SIDE EFFECTS: Updates, for search problem, the list of found solutions,
   the current search-state and search tree node (called an search-node),
   the queue of search tree nodes to be explored, and a trace of states 
   visited thus far."

  
  (loop with feasible?					; True if the current search state 
										; tests feasible.
			 
	  with conflict						; A conflict extracted when the
										; current search state is found infeasible.


		   ;; at the end of each iteration, opsat-solution is
		   ;; bound to the solution that's just been found.
	  with opsat-solution

		   ;; Used only when FIND-ALL-CONSISTENT-MODELS? is True,
		   ;; and hence we are searching for all logical models of
		   ;; the base + sat constraints.
		   ;; This is a list of elements, <SET-OF-MODELS>.  One
		   ;; is generated for each SAT-CONSTRAINT-DESCRIPTIONS in 
		   ;; DESCRIPTIONS-LIST.
		   ;; SET-OF-MODELS = (all-models of base + sat-constraint-descriptions).
	  with models-for-base-sat-constraints

		   ;; Iterate over the kernel assignments (prime implicants)
		   ;; of the known conflicts in best first order.
		   ;; Each kernel assignment (prime implicant) is denoted by
		   ;; a search node.
	  for search-node = (next-candidate
						 search-problem
						 cutoff-cost
						 maximum-visited-state-count

						 ;; Always want solutions as full assignments.
						 ;; No - these should be clauses.
						 nil

						 ;; If we want multiple solutions without returning
						 ;; the same solution multiple times, we need to ensure 
						 ;; that the set of leaves under two search nodes are 
						 ;; disjoint, so that each candidate is visited at
						 ;; most once (systematicity).
						 
						 ;; If only one solution is desired, then disjointness
						 ;; (systematicity) doesn't affect the correctness of 
						 ;; the returned solution.
						 
						 ;; Systematicity may improve or degrade search time.  
						 ;; This is an open, empirical question that needs to 
						 ;; be evaluated.

						 ;; Choose one of the following:
						 ;; - Passing T enforces disjointness for single and 
						 ;;   multiple solutions.
;;;						 'T
						 ;; 
						 ;; - Passing multiple-solutions? doesn't enforce 
						 ;;   disjointness for single solution, but does 
						 ;;   when generating multiple solutions.

						 multiple-solutions?)

	  until

		;; Terminate when no candidate is returned
		;; (in the form of a search-node).
		;; This occurs when one of the termination 
		;; conditions has been reached.
		(null search-node)
        
		;; Set the current search state to the highest value 
		;; state that is consistent with the assignment of search-node.
		;; A state is a full assignment, and current is an extension
		;; of the assignment of search-node.
	  do (setup-current-search-state-from-search-node
		  search-problem search-node)

	   (if (bfig::all-state-variables-assigned? search-node)
		   (progn (debug-print-for :opsat t "Candidate ~a: ~a." 
								   (search-problem-candidate-count search-problem)
								   (current-search-state search-problem)))
		   (progn (debug-print-for :opsat t "Partial candidate ~a: ~a."
								   (search-problem-candidate-count search-problem)
								   (current-search-state search-problem))))

	   ;;(format t "   Ancester Assignment: ~A~%" (BFIG::search-node-ancestor-assignments search-node))
	   ;;(format t "   Not yet assigned ancestors: ~A~%" (BFIG::search-node-unassigned-ancestor-variables search-node))
	   ;;(format t "   Child Assignments: ~A~%" (BFIG::search-node-child-assignments search-node))
	   ;;(format t "   Not yet assigned: ~A~%" (BFIG::search-node-unassigned-variables search-node))

		 ;; The current state is an assignment to decision variables.
		 ;; Check the feasibility of the current state against
		 ;; search problem's constraint system, and receive
		 ;; a conflict if inconsistent.  
		 
		 ;; A CONFLICT is either a list of state variable assignments 
		 ;; or the keyword :CURRENT, denoting the current search state.

		 ;; Each assignment in CONFLICT is a state assignment object,
		 ;; created by the candidate generator (bfig),
		 ;; not an assignment object created by the constraint system.
		 (multiple-value-setq (feasible? conflict models-for-base-sat-constraints)

		   (if find-all-consistent-models?

			   ;; Find all models consistent with the current state.
			   (check-constraints-and-find-all-models-if-any
				search-problem perform-search-to-test-constraints?)

			 ;; Just check consistency of the current state.
			 (check-current-state-against-constraints 
			  search-problem perform-search-to-test-constraints?)))

	   (format t "Detected Conflicts ~A~%" conflict)
	   
	  if (and feasible? 
			  (bfig::all-state-variables-assigned? search-node))
		 ;; If current state is a full assignment and consistent, 
		 ;; then its a solution.

	  do 
		;; Inform the candidate-generator that SEARCH-NODE
		;; denotes a solution.
		(candidate-is-solution! search-node search-problem)

;;;		(incf common-lisp-user::*model-counter*
;;;			  (length  models-for-base-sat-constraints))

		;; When its the first solution found, 
		;; record its cost as the best found.
		(when (first-solution-found? search-problem)
             
		  (setf (search-problem-best-solution-cost search-problem)
			(search-node-best-cost search-node)))


		;; Record solution as a solution object.

		(setf opsat-solution
		  (make-instance 'opsat-solution
		    :decision-assignments (current-search-constraint-assignments search-problem)
		    :sets-of-consistent-models models-for-base-sat-constraints))

	   


		(setf (gethash 'metric-name
					   (opsat-solution-objective-fns-costs-ht opsat-solution))
		  
		  (search-node-best-cost search-node))
		
		;; Return the current state as a solution,
		;; in the form of a list of assignment objects 
		;; from search-problem's constraint system.

	  and return (values opsat-solution T)
	   
	  else
	  if (not feasible?)
	   
	  ;; When the current state is inconsistent,
	  ;; a) record state's conflict, b) record tracing information
	  ;; for debugging, and c) repeat the loop to get the next 
	  ;; search state.
	  do (record-conflict-for-inconsistent-search-state!
		  search-problem :conflict conflict)

	   ;; (break "next-solution, solution not found")
	   
	   (debug-print-for :opsat t "Recording conflict ~a" conflict)

		 (trace-search-state
		  (opsat-problem-state-trace search-problem)
		  (current-search-constraint-assignments search-problem)
		  :rejected)
		 
	  end
		;; When all states are exhausted or a cutoff has been reached, 
		;; return that no solution was found.
		
	  finally (return (values nil nil))))



(defmethod next-minimal-relaxation ((search-problem opsat-problem)
								  cutoff-cost 
								  maximum-visited-state-count
								  perform-search-to-test-constraints?
								  multiple-solutions?
								  &key (find-all-consistent-models? nil))

  " Find next minimal solution to the search problem"

  (declare (ignore find-all-consistent-models?))
  (loop with feasible?					; True if the current search state 
										; tests feasible.
			 
	  with conflict						; A conflict extracted when the
										; current search state is found infeasible.


		   ;; at the end of each iteration, opsat-solution is
		   ;; bound to the solution that's just been found.
	  with opsat-solution

		   ;; Used only when FIND-ALL-CONSISTENT-MODELS? is True,
		   ;; and hence we are searching for all logical models of
		   ;; the base + sat constraints.
		   ;; This is a list of elements, <SET-OF-MODELS>.  One
		   ;; is generated for each SAT-CONSTRAINT-DESCRIPTIONS in 
		   ;; DESCRIPTIONS-LIST.
		   ;; SET-OF-MODELS = (all-models of base + sat-constraint-descriptions).
	  with models-for-base-sat-constraints

		   ;; Iterate over the kernel assignments (prime implicants)
		   ;; of the known conflicts in best first order.
		   ;; Each kernel assignment (prime implicant) is denoted by
		   ;; a search node.
	  for search-node = (next-minimal-relaxation-candidate
						 search-problem
						 cutoff-cost
						 maximum-visited-state-count

						 ;; Always want solutions as full assignments.
						 ;; No - these should be clauses.
						 nil

						 ;; If we want multiple solutions without returning
						 ;; the same solution multiple times, we need to ensure 
						 ;; that the set of leaves under two search nodes are 
						 ;; disjoint, so that each candidate is visited at
						 ;; most once (systematicity).
						 
						 ;; If only one solution is desired, then disjointness
						 ;; (systematicity) doesn't affect the correctness of 
						 ;; the returned solution.
						 
						 ;; Systematicity may improve or degrade search time.  
						 ;; This is an open, empirical question that needs to 
						 ;; be evaluated.

						 ;; Choose one of the following:
						 ;; - Passing T enforces disjointness for single and 
						 ;;   multiple solutions.
;;;						 'T
						 ;; 
						 ;; - Passing multiple-solutions? doesn't enforce 
						 ;;   disjointness for single solution, but does 
						 ;;   when generating multiple solutions.

						 multiple-solutions?)

	  until

		;; Terminate when no candidate is returned
		;; (in the form of a search-node).
		;; This occurs when one of the termination 
		;; conditions has been reached.
		(null search-node)
        
		;; Set the current search state to the highest value 
		;; state that is consistent with the assignment of search-node.
		;; A state is a full assignment, and current is an extension
		;; of the assignment of search-node.
	  ;; do (setup-current-search-state-from-search-node
	  ;; 	  search-problem search-node)

	   do (if (bfig::all-state-variables-assigned? search-node)
		   (progn (debug-print-for :opsat t "Candidate ~a: ~a." 
								   (search-problem-candidate-count search-problem)
								   (current-search-state search-problem))
				  ;;				  (format t "Candidate ~a: ~a.~%" 
				  ;;		  (search-problem-candidate-count search-problem)
				  ;;		  (current-search-state search-problem)				  )
				  )
		   (progn (debug-print-for :opsat t "Partial candidate ~a: ~a."
								   (search-problem-candidate-count search-problem)
								   (current-search-state search-problem))
				  ;;				  (format t "Partial candidate ~a: ~a.~%"
					;;	  (search-problem-candidate-count search-problem)
					;;	  (current-search-state search-problem)				  )
				  ))

	   ;;(format t "   Ancester Assignment: ~A~%" (BFIG::search-node-ancestor-assignments search-node))
	   ;;(format t "   Not yet assigned ancestors: ~A~%" (BFIG::search-node-unassigned-ancestor-variables search-node))
	   ;;(format t "   Child Assignments: ~A~%" (BFIG::search-node-child-assignments search-node))
	   ;;(format t "   Not yet assigned: ~A~%" (BFIG::search-node-unassigned-variables search-node))

		 ;; The current state is an assignment to decision variables.
		 ;; Check the feasibility of the current state against
		 ;; search problem's constraint system, and receive
		 ;; a conflict if inconsistent.  
		 
		 ;; A CONFLICT is either a list of state variable assignments 
		 ;; or the keyword :CURRENT, denoting the current search state.

		 ;; Each assignment in CONFLICT is a state assignment object,
		 ;; created by the candidate generator (bfig),
		 ;; not an assignment object created by the constraint system.
		 (multiple-value-setq (feasible? conflict models-for-base-sat-constraints)
		   
			 ;; Just check consistency of the current state.
			 (check-current-state-against-constraints 
			  search-problem perform-search-to-test-constraints?))

	   (format t "Feasible? ~A~%" feasible?)

	   (setf (BFIG::search-problem-current-search-node-feasible? search-problem) feasible?)



	   ;;(push (BFIG::search-node-ancestor-assignments search-node) (BFIG::state-space-conflicts search-problem))

	   (when feasible?

		 (push (BFIG::search-problem-current-search-node search-problem)
			   (BFIG::search-problem-feasible-nodes search-problem))

		 (push (BFIG::search-node-source-minimal-relaxations 
				(BFIG::search-problem-current-search-node search-problem))
			   (BFIG::search-problem-feasible-minimal-relaxations search-problem))
		 

		 ;; (push (BFIG::search-node-source-minimal-relaxations 
		 ;; 		(BFIG::search-problem-current-search-node search-problem))
		 ;; 	   (BFIG::state-space-conflicts search-problem))

		 ;; (BFIG::insert-new-conflict search-problem 
		 ;; 							:inconsistent-assignments 
		 ;; 							(BFIG::search-node-source-minimal-relaxations 
		 ;; 							 (BFIG::search-problem-current-search-node search-problem)))
		 ;; (format t "Conflict? ~A~%" (BFIG::search-node-source-minimal-relaxations 
		 ;; 							 (BFIG::search-problem-current-search-node search-problem)))
		 )

	   (format t "Conflict? ~A~%" conflict)
	   
	  if (and feasible? 
			  (bfig::all-state-variables-assigned? search-node))
		 ;; If current state is a full assignment and consistent, 
		 ;; then its a solution.

	  do 
		;; Inform the candidate-generator that SEARCH-NODE
		;; denotes a solution.
		(candidate-is-solution! search-node search-problem)

;;;		(incf common-lisp-user::*model-counter*
;;;			  (length  models-for-base-sat-constraints))

		;; When its the first solution found, 
		;; record its cost as the best found.
		(when (first-solution-found? search-problem)
             
		  (setf (search-problem-best-solution-cost search-problem)
			(search-node-best-cost search-node)))


		;; Record solution as a solution object.

		(setf opsat-solution
		  (make-instance 'opsat-solution
		    :decision-assignments (current-search-constraint-assignments search-problem)
		    :sets-of-consistent-models models-for-base-sat-constraints))

	   


		(setf (gethash 'metric-name
					   (opsat-solution-objective-fns-costs-ht opsat-solution))
		  
		  (search-node-best-cost search-node))
		
		;; Return the current state as a solution,
		;; in the form of a list of assignment objects 
		;; from search-problem's constraint system.

	  and return (values opsat-solution T)
	   
	  else
	  if (not feasible?)
	   
	  ;; When the current state is inconsistent,
	  ;; a) record state's conflict, b) record tracing information
	  ;; for debugging, and c) repeat the loop to get the next 
	  ;; search state.
	  do (record-conflict-for-inconsistent-search-state!
		  search-problem :conflict conflict)

	   ;; (break "next-solution, solution not found")
	   
		 (trace-search-state
		  (opsat-problem-state-trace search-problem)
		  (current-search-constraint-assignments search-problem)
		  :rejected)
		 
	  end
		;; When all states are exhausted or a cutoff has been reached, 
		;; return that no solution was found.
		
	  finally (return (values nil nil))))


;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Record a Newly Found Consistent Candidate as a Solution
;;;
;;;;;;;;;;;;;;;;;;;;;
    
;;; Optimization:

;;; o The use of LAST in the following is inefficient.
;;;   Instead, append to the end of solutions more 
;;;   efficiently wih a second pointer.
(defmethod new-solution! 
	(solution (search-problem opsat-problem))
  
  "Records SOLUTION as one of the found solutions of
   SEARCH-PROBLEM.  Ensures that the record is in the order 
   that the solutions were discovered.  
  
   SOLUTION is expected to a be list of assignments
   objects from SEARCH-PROBLEM's constraint system.
   Called for effect, not value."

  (let ((count (incf (search-problem-solution-count 
					  search-problem))))

	;; Inform us of the found solution.
	(debug-print-for :opsat t "Solution ~a: ~a." count solution)
	
	(if (null (search-problem-solutions search-problem))
		
	  ;; First solution found, start a new list of solutions.
		(setf (search-problem-solutions search-problem)
          (list solution))

	  ;; Else, add to the end of an existing list of solutions.
	  (setf (rest (last (search-problem-solutions
						 search-problem)))
		(list solution))))
  T)


;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Record a Conflict from an Inconsistent Candidate
;;;
;;;;;;;;;;;;;;;;;;;;;

(defmethod record-conflict-for-inconsistent-search-state!
    (problem-without-conflicts &key &allow-other-keys)
  
  ;; Problem doesn't maintain conflicts, ignore the conflict.
  
  problem-without-conflicts)


(defmethod record-conflict-for-inconsistent-search-state!
    ((search-problem informed-search-problem-w-conflicts) &key conflict)
  
  " Records CONFLICT, derived from the CURRENT search-state of 
   SEARCH-PROBLEM.  
   Called for effect, not value.
  
   CONFLICT is a list of assignments or the keyword :CURRENT.  
   If CONFLICT is a list of assignments, then its conjunction is 
   inconsistent with SEARCH-PROBLEM's constraint system.  
   If CONFLICT is :CURRENT, then the complete set of 
   assignments for the current search state form the conflict.
   Each ASSIGNMENT is a state assignment object of candgen
   (not an assignment object of the constraint system)."
  
  (cond ((not (null conflict))
		 ;; When a conflict is supplied, 
		 ;; record it in the search problem's conflict database.

		 (when (eql conflict :current)
		   
		   ;; If Conflict is the keyword :CURRENT,
		   ;; replace with the complete current search state as the conflict.
		   (setq conflict (remove nil (current-search-state search-problem))))
		 
		 (unless (null conflict)

		   ;; Inform us of the conflict,
		   (debug-print-for :opsat t "Conflict ~a." conflict)

		   ;; and record it.
		   (insert-new-conflict search-problem
								:inconsistent-assignments conflict))) ;;TODO maybe a hack?
    
		;; No conflict was returned, even though search-state was inconsistent,
		;; this shouldn't happen.
		(:else
		 
		 ;; Wasn't given a conflict.
		 (cerror "Continue" "Inconsistent search state, but no conflict returned.")))
  T)



;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Computing the Cutoff Cost
;;;
;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o DECISION-SPACE-BINARY-OBJECTIVE-FUNCTION should probably 
;;;   instead be a method call.
(defmethod compute-cutoff-cost ((search-problem opsat-problem) cutoff-delta)
  
  "Computes the cutoff cost for SEARCH-PROBLEM.   If there is 
   no solution or if CUTOFF-DELTA is NIL, then the result is NIL,
   signifying that there is no cutoff cost.  Otherwise, the cost
   function of the search problem (DECISION-SPACE-BINARY-OBJECTIVE-FUNCTION)
   is applied to the combination of the cost of the first solution found 
   and CUTOFF-DELTA.  
   By default DECISION-SPACE-BINARY-OBJECTIVE-FUNCTION returns a delta off the 
   best cost."
  
  ;; Only call if a best solution has been found and a cutoff-delta
  ;; has been provided.
  (when (and (search-problem-best-solution-cost search-problem)
			 
			 cutoff-delta)
	
	;; Compute the cutoff from the best cost and cutoff-delta
	;; using the function DECISION-SPACE-BINARY-OBJECTIVE-FUNCTION,
	;; associated with search problem.
    (funcall (decision-space-binary-objective-function search-problem)
			 (search-problem-best-solution-cost search-problem)
			 cutoff-delta)))
