;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

(in-package :OPSAT)

(eval-when (compile load eval)
  (export 
   '(
;;;	 new-constraint-based-trellis-diagram
	 constraint-based-trellis-diagram
	 constraint-based-trellis-search-problem
	 )
   :OPSAT)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Creating, Printing an Finding Solutions to 
;;; a Constraint-based Trellis Diagram
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (MAKE-INSTANCE CONSTRAINT-BASED-TRELLIS-DIAGRAM
;;;			&KEY
;;;			(STATE-VARIABLE-DESCRIPTIONS '())
;;;			(INITIAL-STATE-DESCRIPTIONS '())
;;;			(STATE-CONSTRAINT-DESCRIPTIONS '())
;;;			(TRANSITION-DESCRIPTIONS '())
;;;			
;;;			(BINARY-TRANSITION-OBJECTIVE-FUNCTION #'+)  
;;;			(TRANSITION-COST-OFFSET 0)
;;;			(COST-ORDER-FUNCTION #'>=)
;;;			
;;;			(CONSTRAINT-SYSTEM-TYPE 'THEORY)
;;;			(CONFLICT-DIRECTED? T)
;;;			(SEARCH-ORDER :BEST-FIRST))
;;;
;;;   Creates a constraint encoding of a trellis diagram and 
;;;   frames the problem of enumerating the consistent trajectories
;;;   through the Trellis in best first order of cost.
;;;
;;;   Trellis is named NAME and is generated by SOURCE, where NAME and 
;;;   SOURCE are lisp objects.
;;;
;;;   A trajectory through the trellis is a sequences of 
;;;   TIME-STEP-COUNT states, numbered from 0, where count must be 
;;;   at least 2, and defaults to 2.
;;;
;;;   A trellis specifies the set of all consistent states at each 
;;;   time step up to TIME-STEP-COUNT, and all consistent transitions
;;;   between these sucessive states.
;;;
;;;   Each candidate state in the trellis is expressed as an assignment 
;;;   to a set of finite domain state variables, specified in 
;;;   STATE-VARIABLE-DESCRIPTIONS.
;;;
;;;   Consistent states are restricted to those that satsfy the state
;;;   constraints specified in STATE-CONSTRAINT-DESCRIPTIONS.
;;;
;;;   The set of allowed initial states and their corresponding cost is
;;;   specified in INITIAL-STATE-DESCRIPTIONS.
;;;
;;;   The set of consistent state transitions are those satisfying the 
;;;   constraints specified in TRANSITION-DESCRIPTIONS.
;;;
;;;   Each candidate transition in the trellis corresponds to an assignment 
;;;   to a set of finite domain transition variables, where the transition
;;;   variables are specified in TRANSITION-DESCRIPTIONS.
;;;
;;;   A consistent trajectory is an assignment to state and transition 
;;;   variables from time steps 1 to NUMBER-OF-TIME-STEPS, such that 
;;;   the state and transition constraints for each time step are satisfied.
;;;
;;;   The problem defaults to enumerating trajectories in decreasing
;;;   order of cost.
;;;
;;;   Trajectory cost defaults to the sum of the cost of the global
;;;   transitions between successive states in the trajectory.  The cost
;;;   of a transition is the sum of the costs of concurrent primitive
;;;   transitions that comprise the (composite) transition.  
;;;
;;;   The objective function is specified through variables
;;;     BINARY-TRANSITION-OBJECTIVE-FUNCTION (defaults to '>=),
;;;     TRANSITION-COST-OFFSET (defaults to 0), and 
;;;     COST-ORDER-FUNCTION (defaults to '>=).
;;;
;;;   STATE-VARIABLE-DESCRIPTIONS is a list:
;;;     (STATE-VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
;;;    where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
;;;    supplied to make-instance of state variable.  It currently consists of
;;;    one required keyword pair :DOMAIN DOMAIN, where DOMAIN is a list of values.
;;;
;;;   INITIAL-STATE-DESCRIPTIONS is a list of elements:
;;;     (STATE-VARIABLE-NAME :DOMAIN ((INITIAL-VALUE :COST COST)+))
;;;
;;;     where STATE-VARIABLE-NAME is a name specified in 
;;;       STATE-VARIABLE-DESCRIPTIONS, and
;;;     where there is an (INITIAL-VALUE :COST COST) description for each 
;;;       INITIAL-VALUE in the domain of STATE-VARIABLE-NAME.
;;;
;;;     An INITIAL-STATE-DESCRIPTION needs to be provided for every state
;;;     variable declared in STATE-VARIABLE-DESCRIPTIONS.
;;;
;;;   STATE-CONSTRAINT-DESCRIPTIONS is a list of state logic constraint 
;;;   descriptions, each of the form:
;;;     (:CLAUSE STATE-LITERAL+) where
;;;
;;;     A STATE LITERAL is a STATE-PROPOSITION or (NOT STATE-PROPOSITION). 
;;;     A STATE-PROPOSITION is an assignment:
;;;       (= STATE-VARIABLE-NAME VALUE),   
;;;
;;;       where STATE-VARIABLE-NAME is defined in STATE-VARIABLE-DESCRIPTIONS, and
;;;       where VALUE is a value in the domain of STATE-VARIABLE-NAME.
;;;
;;;   TRANSITION-DESCRIPTIONS describes the legal transitions for
;;;   the state variables.  TRANSITION-DESCRIPTIONS is a list of
;;;   STATE-VARIABLE-TRANSITION-DESCRIPTION elements of the form:
;;;
;;;     (SOURCE-STATE-VARIABLE-NAME
;;;       (SOURCE-VALUE TRANSITION-VARIABLE-NAME 
;;;                     SOURCE-TRANSITION-DESCRIPTION+)+)
;;;
;;;       where SOURCE-STATE-VARIABLE-NAME is a name specified in 
;;;          STATE-VARIABLE-DESCRIPTIONS, and

;;;       where each SOURCE-TRANSITION-DESCRIPTION is of the form:
;;;         (TRANSITION-VALUE COST TARGET-DESCRIPTION+)
;;;
;;;         where TARGET-DESCRIPTION is of the form:
;;;           (GUARD TARGET-ASSIGNMENT)
;;;
;;;           where TARGET-ASSIGNMENT is a list of state variable assignments
;;;             that hold at the next time point if conditions are satisfied.
;;;
;;;             where GUARD is a list of zero or more state assignment literals whose conjunction
;;;                enables the transition when satisfied in the preceding time point. 
;;;
;;;   CONSTRAINT-SYSTEM-TYPE is an object type with slot NAME and
;;;     initarg :NAME. CONSTRAINT-SYSTEM-TYPE defaults to THEORY.
;;;     CONSTRAINT-SYSTEM-TYPE must support the methods specified in the
;;;     NEW-OPSAT-PROBLEM documentation string.
;;;
;;;   CONFLICT-DIRECTED? is a flag that defaults to T.  Conflict-directed search
;;;     is performed by OPSAT on the trellis if and only if this flag is true.
;;;
;;;   SEARCH-ORDER specifies the order in which the search tree is
;;;     explored.  This defaults to :BEST-FIRST. 


;;; (PRINT-DESCRIPTION  
;;;	 (TRELLIS CONSTRAINT-BASED-TRELLIS-DIAGRAM) STREAM)
;;;  
;;;   Prints to STREAM a description of the constraint encoding
;;;   of trellis.

;;;(FIND-SOLUTIONS (SEARCH-PROBLEM CONSTRAINT-BASED-TRELLIS-DIAGRAM)
;;;						   &KEY
;;;						   (CUTOFF-DELTA NIL)
;;;						   (MAXIMUM-SOLUTION-COUNT 10)
;;;						   (MAXIMUM-VISITED-STATE-COUNT NIL)
;;;						   (PERFORM-SEARCH-TO-TEST-CONSTRAINTS? T)
;;;						   (SEARCH-ORDER :DEFAULT)
;;;						   (FIND-ALL-CONSISTENT-MODELS? NIL))
;;;  
;;;   Starting from the begining of the search space, finds the 
;;;   leading assignments to decision variables that satisfy
;;;   constraints of SEARCH-PROBLEM and have the best cost.
;;;
;;;   Each solution returned is a list of indexed assignments, 
;;;   in increasing order of index, where
;;;   each assignment is an assignment object of SEARCH-PROBLEM's
;;;   constraint system, denoting an indexed decision variable assignment.
;;;
;;;   Generates the sequence of leading best solutions up to 
;;;   MAXIMUM-SOLUTION-COUNT. MAXIMUM-SOLUTION-COUNT defaults to nil, 
;;;   which represents no bound.  
;;;
;;;   Search-order specifies the order in which search-tree nodes
;;;   are expanded.  Its :DEFAULT or one of *ALLOWED-SEARCH-ORDER-TYPES*.
;;;   If :DEFAULT is specified, then defaults the the search-order specified
;;;   when search-problem was created.
;;;
;;;   Search terminates earlier if the MAXIMUM-VISITED-STATE-COUNT is exceeded 
;;;   or if computed cutoff cost is reached.  
;;;   MAXIMUM-VISITED-STATE-COUNT defaults to 0.  
;;;  
;;;   To determine cutoff cost, if the best first search is in decreasing 
;;;   (increasing) order, the resulting cutoff cost is the least (greatest) 
;;;   cost considered.  The cutoff cost is computed from CUTOFF-DELTA 
;;;   using function COMPUTE-CUTOFF-COST.  
;;;  
;;;   CUTOFF-DELTA specifies a maximum allowed difference (drop off) 
;;;   between the cost of the best solution found and the current 
;;;   search state, or no cutoff, if nil.
;;;   CUTOFF-DELTA defaults to nil.
;;;
;;;   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE then it searches over
;;;   constraint assignments to see if constraints are satisfied.  
;;;   If nil, then it checks the constraints by performing
;;;   constraint propagation alone.
;;;  
;;;   ASSUMPTIONS:
;;;    - The constraint system has already been setup with all 
;;;      relevant constraints inserted, such as observables, 
;;;      control values and models.
;;;    - Any irrelevant assertions or constraints have been removed
;;;      before invocation.
;;;
;;;   SIDE-EFFECTS:
;;;    - Stores leading solutions in SEARCH-PROBLEM-SOLUTIONS.
;;;    - Upon termination, removes any assignments it introduced into 
;;;      CONSTRAINT-SYSTEM during search.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o Sort out the encoding of initial state.
(defclass constraint-based-trellis-diagram ()

  ((name
	:initarg :name
	:initform (gentemp "CONSTRAINT-BASED-TRELLIS-")
	:reader constraint-based-trellis-name
	:documentation "Name of a Constraint-based Trellis Diagram.")
   
   (source
	:initarg :source
	:initform nil
	:reader constraint-based-trellis-source
	:documentation "Ojbect that generated the trellis.")
   
   (time-step-count
	:initarg :time-step-count
	:initform 2
	:reader constraint-based-trellis-time-step-count
	:documentation "Name of a Constraint-based Trellis Diagram.")

   ;;; Fix:
   ;;; - How should the initial state values be structured?
   (initial-state-variable
	:reader constraint-based-trellis-initial-state-variable
	:documentation "A variable denoting assignments to the state
                    variables at the first time step.")
   
   (state-variables
	:reader constraint-based-trellis-state-variables
	:documentation "An array of n state variable descriptions.
                    An assignment to variables denotes the states 
                    in an N step trajectory.  An 
                    assignment to the ith variables in the array denotes 
                    the ith state in the trajectory.")

   (transition-variables
	:reader constraint-based-trellis-transition-variables
	:documentation "An array of n transition variable descriptions.
                    An assignment to these variables denotes the transitions 
                    of an N step trajectory.  An assignment to the 
                    ith variable in the array denotes 
                    a transition from the ith to the i+1th state 
                    in the trajectory.")
   
   (constraints
	:reader constraint-based-trellis-constraints
	:documentation "An array of n constraint descriptions.  The
                    constraints in position i denote the state constraints
                    at time i and the transition constraints from i to i+1. ")
   
   (search-problem
	:reader constraint-based-trellis-search-problem
	:documentation "The OPSAT Constraint optimization problem used to compute
                    a minimum cost trajectory through Trellis.")

   (soft-constraints
	:initarg :soft-constraints
	:reader constraint-based-trellis-soft-constraints
	:initform '()
	:documentation ""))

  (:documentation "Encoding of the problem of computing an optimal trajectory
                   as a Constraint-based Trellis Diagram (a kind of OPSAT formulation)."))

;;; Fix:

;;; o One transition should be taken with each variable at each time point,
;;;   rather than each assignment.  Hence transition decision variables 
;;;   should be associated with each variable, not each assignment.  

;;; o Think through and correct the setup of the initial state variable.
;;;   See notes on the function below.

;;; o Think through n-step-estimation for 1 step.  The documentation says that
;;;   this isn't allowed, but it appears to be equivalent to sherlock-style estimation.
;;;   Is that correct?  If so, update the documentation accordingly, and the tests for valid
;;;   step length.

;;; o Make clear what an estimate is.  Does it specify a state sequence or a sequence
;;;   of transition variable assignments.  Make sure that it outputs what we want,
;;;   and then update the documentation.

;;; o When this is finished, get rid of new-constraint-based-trellis-diagram,
;;;   and update header documentation for interface functions.

;;; o setup initial-state-variable, state-variables
;;;   and transition-variables if needed, currently they are descriptions.  ???

;;; To Do:

;;; o Develop the transition relation encoding that captures the marking of primitive
;;;   states underlying start states, found in PHCA.

;;; o Generalize initial state description to sets of valued CSPs.
;;;   For now, encode each valued CSP by adding an initial state variable
;;;   that selects one of the tuples of the valued constraint.
;;;   - represent each tuple as a list of assignments.
;;;   - index the assignment variables with Time = 0.

;;; o To implement online filtering, the initial state constraints
;;;   need to be expressed as a dynamic constraint, which can be
;;;   removed and replaced with another valued constraint.
;;;   To do this, for each update, remove all constraints and learned
;;;   conflicts that involve assignments used to select an initial state assignment.

;;; o Add methods for 
;;;   - asserting observed values, and
;;;   - enumerating trajectories and for returning solutions.  
;;;   - We may record solutions found on the trellis object.

;;; o Do we want the trellis class to be a specialization of a problem?

;;; o Allow the distribution over initial states to be modified,
;;;   this is needed to implement n-step estimation with a moving
;;;   time window.

;;; o Upon input, test that the descriptions are well formed.

;;; o Change transition-descriptions to use keywords, in order 
;;;   to be consistent with the other types of descriptions.

;;; o Generalizations:
;;;   - Generalize cost functions.  
;;;     For example, divide cost into a function 
;;;     for combining concurrent (simultaneous) transitions and 
;;;      a function for combining successive transitions.
;;;   - Permit state constraints specified by
;;;     different kinds of constraint systems (e.g, linear eqns or inequalities ala JHU-APL).
;;;     Nominally this is permitted by OPSAT, but won't work for
;;;     transition encodings, since they are hard wired to state-clauses.
;;;   - Permit state constraints expressed as conditional CSPs (is this useful?).
(defmethod initialize-instance :after 
		   ((trellis constraint-based-trellis-diagram) 
			&key
			(state-variable-descriptions '())
			(initial-state-descriptions '())
			(state-constraint-descriptions '())
			(transition-descriptions '())
			
			(binary-transition-objective-function #'+)  
			(transition-cost-offset 0)
			(cost-order-function #'>=)
			
			(constraint-system-type 'theory)
			(conflict-directed? T)
			(search-order :best-first))

  "Creates a constraint encoding of a trellis diagram and 
   frames the problem of enumerating the consistent trajectories
   of this trellis in best first order.

   Trellis is named NAME and is generated by SOURCE, where NAME and 
   SOURCE are lisp objects.

   A trajectory through the trellis is a sequences of 
   TIME-STEP-COUNT states, numbered from 0, where count must be 
   at least 2, and defaults to 2.

   A trellis specifies the set of all consistent states at each 
   time step up to TIME-STEP-COUNT, and all consistent transitions
   between these sucessive states.

   Each candidate state in the trellis is expressed as an assignment 
   to a set of finite domain state variables, specified in 
   STATE-VARIABLE-DESCRIPTIONS.

   Consistent states are restricted to those that satisfy the state
   constraints specified in STATE-CONSTRAINT-DESCRIPTIONS.

   The set of allowed initial states and their corresponding cost is
   specified in INITIAL-STATE-DESCRIPTIONS.

   The set of consistent state transitions are those satisfying the 
   constraints specified in TRANSITION-DESCRIPTIONS.

   Each candidate transition in the trellis corresponds to an assignment 
   to a set of finite domain transition variables, one for each state
   variable and time step, where the transition
   variables are specified in TRANSITION-DESCRIPTIONS.

   A consistent trajectory is an assignment to state and transition 
   variables from time step 0 to NUMBER-OF-TIME-STEPS, such that 
   the state and transition constraints for each time step are satisfied.

   The problem defaults to enumerating trajectories in decreasing
   order of cost.

   Trajectory cost defaults to the sum of the cost of the global
   transitions between successive states in the trajectory.  The cost
   of a transition is the sum of the costs of concurrent primitive
   transitions that comprise the (composite) transition.  

   The objective function is specified through variables
     BINARY-TRANSITION-OBJECTIVE-FUNCTION (defaults to '>=),
     TRANSITION-COST-OFFSET (defaults to 0), and 
     COST-ORDER-FUNCTION (defaults to '>=).

   STATE-VARIABLE-DESCRIPTIONS is a list:
     (STATE-VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
    where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of state variable.  It currently consists of
    one required keyword pair :DOMAIN DOMAIN, where DOMAIN is a list of values.

   INITIAL-STATE-DESCRIPTIONS is a list of elements:
     (STATE-VARIABLE-NAME :DOMAIN ((INITIAL-VALUE :COST COST)+))

     where STATE-VARIABLE-NAME is a name specified in 
       STATE-VARIABLE-DESCRIPTIONS, and
     where there is an (INITIAL-VALUE :COST COST) description for each 
       INITIAL-VALUE in the domain of STATE-VARIABLE-NAME.

     An INITIAL-STATE-DESCRIPTION needs to be provided for every state
     variable declared in STATE-VARIABLE-DESCRIPTIONS.

   STATE-CONSTRAINT-DESCRIPTIONS is a list of state logic constraint 
   descriptions, each of the form
     (:CLAUSE STATE-LITERAL+) where

     A STATE LITERAL is a STATE-PROPOSITION or (NOT STATE-PROPOSITION). 
     A STATE-PROPOSITION is an assignment:
       (= STATE-VARIABLE-NAME VALUE),   

       where STATE-VARIABLE-NAME is defined in STATE-VARIABLE-DESCRIPTIONS, and
       where VALUE is a value in the domain of STATE-VARIABLE-NAME.

   TRANSITION-DESCRIPTIONS describes the legal transitions for
   the state variables.  TRANSITION-DESCRIPTIONS is a list of
   STATE-VARIABLE-TRANSITION-DESCRIPTION elements of the form:

     (SOURCE-STATE-VARIABLE-NAME TRANSITION-VARIABLE-NAME
       (SOURCE-VALUE  
           SOURCE-TRANSITION-DESCRIPTION+)+)

       where SOURCE-STATE-VARIABLE-NAME is a name specified in 
          STATE-VARIABLE-DESCRIPTIONS, and

       where each SOURCE-TRANSITION-DESCRIPTION is of the form:

         (TRANSITION-VALUE COST TARGET-DESCRIPTION+)

         where TARGET-DESCRIPTION is of the form:
           (GUARD TARGET-ASSIGNMENT)

           where TARGET-ASSIGNMENT is a list of state variable assignments
             that hold at the next time point if conditions are satisfied.

           where GUARD is a list of zero or more state assignment literals whose conjunction
             must be satisfied by the system state in the preceding time point, 
            in order for the transition to be enabled.

   CONSTRAINT-SYSTEM-TYPE is an object type with slot NAME and
     initarg :NAME. CONSTRAINT-SYSTEM-TYPE defaults to THEORY.
     CONSTRAINT-SYSTEM-TYPE must support the methods specified in the
     NEW-OPSAT-PROBLEM documentation string.

   CONFLICT-DIRECTED? is a flag that defaults to T.  Conflict-directed search
     is performed by OPSAT on the trellis if and only if this flag is true.

   SEARCH-ORDER specifies the order in which the search tree is
     explored.  This defaults to :BEST-FIRST. "

  (with-slots (name source time-step-count
			   initial-state-variable
			   state-variables
			   transition-variables
			   constraints
			   search-problem
			   soft-constraints)
	  trellis
	
	(setq state-variables (make-array time-step-count)
		  transition-variables (make-array time-step-count)
		  constraints (make-array time-step-count))

	(loop 
	   with initial-state-opsat-variable-descriptions =
		 
	   ;; o Add to OPSAT variables, variables corresponding to the set 
	   ;;   of possible initial states.
		 (add-index-to-variable-descriptions initial-state-descriptions 0)

	   with indexed-opsat-transition-variable-descriptions
	   with indexed-isat-transition-variable-descriptions
	   with indexed-isat-state-variable-descriptions
	   with indexed-isat-constraint-descriptions
	  
	   for time from 0 to (- time-step-count 1)
		 
	   do (multiple-value-setq (indexed-opsat-transition-variable-descriptions
								indexed-isat-transition-variable-descriptions
								indexed-isat-state-variable-descriptions
								indexed-isat-constraint-descriptions)
			
			;; Create descriptions for state variables and constraints at i
			;; and transition variables and constraints for i to i+1.
			
			(create-trellis-segment-for-time
			 state-variable-descriptions
			 state-constraint-descriptions
			 transition-descriptions
			 time
			 time-step-count))
		 
		 (setf (aref state-variables time)
			   indexed-isat-state-variable-descriptions)
		 
		 (setf (aref transition-variables time)
			   indexed-opsat-transition-variable-descriptions)
		 
		 (setf (aref constraints time)
			   indexed-isat-constraint-descriptions)
		 
	   appending indexed-opsat-transition-variable-descriptions
	   into opsat-variable-descriptions
		 
	   appending (append
				  indexed-isat-transition-variable-descriptions
				  indexed-isat-state-variable-descriptions) 
	   into isat-variable-descriptions
		 
	   appending indexed-isat-constraint-descriptions 
	   into isat-constraint-descriptions
		 
	   finally
		 (debug-print-progn-for :opsat
								(format t "~% initial-state-opsat-variable-descriptions = ~s" 
										initial-state-opsat-variable-descriptions)
								
								(format t "~% opsat-variable-descriptions = ~s" 
										opsat-variable-descriptions)
								(format t "~% isat-variable-descriptions = ~s" 
										isat-variable-descriptions)
								(format t "~% isat-constraint-descriptions = ~s" 
										isat-constraint-descriptions))
		 
;;;		  (format t "~% constraint-system-type = ~s" constraint-system-type)
;;;		  (format t "~% search-order = ~s" search-order)
;;;		  (format t "~% conflict-directed? = ~s" conflict-directed?)
;;;		  (format t "~% binary-transition-objective-function = ~s" binary-transition-objective-function)
;;;		  (format t "~% transition-cost-offset = ~s" transition-cost-offset)
;;;		  (format t "~% cost-order-function = ~s" cost-order-function)
		  
		 (setq initial-state-variable
			   initial-state-opsat-variable-descriptions)
		 
	     ;; Create an OPSAT problem based on the defined variables 
	     ;; and constraints.
		 ;; if observation function is not allowed
		 ;; frame to a OPSAT object search problem
		 ;; else frame to a VCSP object search problem
		 (if (null soft-constraints)
			 (setq search-problem
				   (new-opsat-problem (nconc
									   initial-state-opsat-variable-descriptions
									   opsat-variable-descriptions)
									  isat-variable-descriptions
									  isat-constraint-descriptions
									  
									  :constraint-system-type constraint-system-type
									  :search-order search-order
									  :conflict-directed? conflict-directed?
									  
									  :binary-objective-function
									  binary-transition-objective-function
									  :cost-offset transition-cost-offset
									  :cost-order-function cost-order-function
									  
									  :source trellis
									  :name (gentemp (format nil "~a-PROBLEM"
															 name))))
			 (progn
			   (setq search-problem
					 (new-vcsp-problem (nconc
										initial-state-opsat-variable-descriptions
										opsat-variable-descriptions)
									   isat-variable-descriptions
									   isat-constraint-descriptions
									   :soft-constraint-system
									   soft-constraints
									   
									   :constraint-system-type constraint-system-type
									   :search-order search-order
						               ;;; TODO add the tag back when Gusta fix VCSP code
									   ;;; :conflict-directed? conflict-directed?
									   
									   :binary-objective-function
									   binary-transition-objective-function
									   :cost-offset transition-cost-offset
									   :cost-order-function cost-order-function
									   
									   :source trellis
									   :name (gentemp (format nil "~a-PROBLEM"
															  name))))))
	   ;; FIX:
	   ;; o After the opsat problem is generated,
	   ;;   replace state and transition variable 
	   ;;   descriptions with their corresponding,
	   ;;   instantiated variables.
	   ;; Or do we want to keep these descriptions around
	   ;; for reference and debugging?
	   ;; For example, we can create an association list of variables
	   ;; and their corresponding descriptions.
		 )))

(defmethod print-description  
	((trellis constraint-based-trellis-diagram) stream)
  
  "Prints to STREAM a description of the constraint encoding
   of trellis."
  
  (with-slots (name source time-step-count
			   initial-state-variable
			   state-variables
			   transition-variables
			   constraints) 
	  trellis
	(format stream "Trellis ~a, created by ~a, with ~a steps:" 
			name source time-step-count)
	(loop for i from 0 to (- time-step-count 1)
		do (format stream "~%   Step ~a:" i)
		   (format stream "~%      state variables = ~a" 
				   (aref state-variables i))
		   (format stream "~%      transition variables = ~a" 
				   (aref transition-variables i))
		   (format stream "~%      constraints = ~a" 
				   (aref constraints i)))))

(defmethod find-solutions ((search-problem constraint-based-trellis-diagram)
						   &key
						   (cutoff-delta nil)
						   (maximum-solution-count 10)
						   (maximum-visited-state-count nil)
						   (perform-search-to-test-constraints? T)
						   (search-order :DEFAULT)
						   (find-all-consistent-models? nil))
  
  "Starting from the beginning of the search space, finds the 
   best cost assignments to decision variables that satisfy
   constraints of SEARCH-PROBLEM.

   Each solution returned is a list of indexed assignments
   in increasing order of index, where
   each assignment is an assignment object of SEARCH-PROBLEM's
   constraint system, denoting an assignment to one of the 
   indexed decision variables.

   Generates the sequence of leading best solutions down to
   MAXIMUM-SOLUTION-COUNT. MAXIMUM-SOLUTION-COUNT defaults to nil, 
   which represents no bound.  

   Search-order specifies the order in which search-tree nodes
   are expanded.  Its :DEFAULT or one of *ALLOWED-SEARCH-ORDER-TYPES*.
   If :DEFAULT is specified, then defaults the the search-order specified
   when search-problem was created.

   Search terminates earlier if the MAXIMUM-VISITED-STATE-COUNT is exceeded 
   or if computed cutoff cost is reached.  
   MAXIMUM-VISITED-STATE-COUNT defaults to 0.  
  
   To determine cutoff cost, if the best first search is in decreasing 
   (increasing) order, the resulting cutoff cost is the least (greatest) 
   cost considered.  The cutoff cost is computed from CUTOFF-DELTA 
   using function COMPUTE-CUTOFF-COST.  
  
   CUTOFF-DELTA specifies a maximum allowed difference (drop off) 
   between the cost of the best solution found and the current 
   search state, or no cutoff, if nil.
   CUTOFF-DELTA defaults to nil.

   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE then it searches over
   constraint assignments to see if constraints are satisfied.  
   If nil, then it checks the constraints by performing
   constraint propagation alone.
  
   ASSUMPTIONS:
    - The constraint system has already been setup with all 
      relevant constraints inserted, such as observables, 
      control values and models.
    - Any irrelevant assertions or constraints have been removed
      before invocation.

   SIDE-EFFECTS:
    - Stores leading solutions in SEARCH-PROBLEM-SOLUTIONS.
    - Upon termination, removes any assignments it introduced into 
      CONSTRAINT-SYSTEM during search."

  (find-solutions (constraint-based-trellis-search-problem search-problem)
				  :cutoff-delta cutoff-delta
				  :maximum-solution-count maximum-solution-count 
				  :maximum-visited-state-count maximum-visited-state-count
				  :perform-search-to-test-constraints? perform-search-to-test-constraints?
				  :search-order search-order
				  :find-all-consistent-models? find-all-consistent-models?))

(defmethod find-next-solutions 
	((search-problem constraint-based-trellis-diagram)
	 &optional (cutoff-delta nil)
			   (maximum-solution-count 10)
			   (maximum-visited-state-count nil)
			   (perform-search-to-test-constraints? T)
			   (find-all-consistent-models? nil))

  "CONTINUES a previously initiated search, finding 
   a sequence of decision variable assignments that satisfy
   constraints of SEARCH-PROBLEM, in the order dictated by
   the type of search.
   Otherwise similar to find-solutions.
  
   RETURNS a list of zero or more solutions found,
   where each solution is an object of type opsat-solution,
   which encapsulates all information that might be needed
   regarding a solution. Currently the relevant info is:
   o List of constraint system assignment objects,
		(each denotes an assignment to a decision variable)
   o cost of solution (if an optimization problem) and,
   o if flag find-all-consistent-models? was set, all
	consistent models are also sent.
    (a model for an opsat-solution is the full-assignment, 
	 extended from partial assign made to the decision variables,
     which is consistent with respect to base+sat-description;
	 where sat-description is an element from
	 sat-constraint-descriptions-list).
	 
   Generates a sequence of leading solutions of length up to 
   MAXIMUM-SOLUTION-COUNT. MAXIMUM-SOLUTION-COUNT defaults to nil, 
   which represents no bound.  

   Search terminates earlier if the MAXIMUM-VISITED-STATE-COUNT 
   is exceeded or if for an optimization problem the computed cutoff 
   cost is reached.  MAXIMUM-VISITED-STATE-COUNT defaults to nil.  
  
   To determine cutoff cost, if the best first search is in decreasing 
   (increasing) order, the resulting cutoff cost is the least (greatest) 
   cost considered.  The cutoff cost is computed from CUTOFF-DELTA 
   using function COMPUTE-CUTOFF-COST.  
  
   CUTOFF-DELTA specifies a maximum allowed difference (drop off) 
   between the cost of the best solution found and the current 
   search state, or no cutoff, if nil.
   CUTOFF-DELTA defaults to nil.

   If PERFORM-SEARCH-TO-TEST-CONSTRAINTS? is TRUE then it searches over
   constraint assignments to see if SAT and UNSAT constraints are
   satisfied.  If nil, then it checks the constraints by performing
   constraint propagation alone.

   ASSUMPTIONS:
    - The constraint system has already been setup with all 
      relevant constraints inserted, such as observables, 
      control values and models.
    - Any irrelevant assertions or constraints have been removed
      before invocation.

   SIDE-EFFECTS:
    - Stores leading solutions in SEARCH-PROBLEM-SOLUTIONS.
    - Upon termination, removes any assignments it introduced into 
      CONSTRAINT-SYSTEM during search."

  (find-next-solutions 
   (constraint-based-trellis-search-problem search-problem)
   cutoff-delta
   maximum-solution-count
   maximum-visited-state-count
   perform-search-to-test-constraints?
   find-all-consistent-models?))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding a Slice of the Trellis for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun create-trellis-segment-for-time

    (state-variable-descriptions state-constraint-descriptions
	 transition-descriptions time number-of-time-steps)
  
  "Given a constraint automata described by 

      STATE-VARIABLE-DESCRIPTIONS 
      STATE-CONSTRAINT-DESCRIPTIONS and
      TRANSITION-DESCRIPTIONS,

   as described in INITIALIZE-INSTANCE for CONSTRAINT-BASED-TRELLIS-DIAGRAM,

   creates elements of an OPSat description corresponding to
   a segment of the constraint-based TRELLIS from TIME to TIME + 1.
   Here time ranges from 0 to NUMBER-OF-TIME-STEPS - 1.

   The segment generated is comprised of state and transition 
   variables at TIME, state constraints at TIME, and transition
   constraints from TIME to TIME + 1.  

   Returns the relevant variable and constraint descriptions for 
   OPSAT as:

   (values OPSAT-VARIABLE-DESCRIPTIONS
    	   ISAT-TRANSITION-VARIABLE-DESCRIPTIONS
           ISAT-STATE-VARIABLE-DESCRIPTIONS
    	   ISAT-CONSTRAINT-DESCRIPTIONS)"

  (check-clause-descriptions! state-constraint-descriptions)
  
  (let ((indexed-state-variable-descriptions
		 
		 ;; Generate state variables for TIME.
		 (add-index-to-variable-descriptions state-variable-descriptions
											 time))
	
		;; Generate state constraints for TIME.
		(indexed-state-constraint-descriptions
		 (add-index-to-state-constraint-descriptions
		  state-constraint-descriptions time)))

	;; o Encode transitions from TIME to TIME + 1, 
	;;   - by specifying transition variables at TIME and 
	;;     introducing them both as OPSAT and ISAT variables, and
	;;   - by defining state constraints that instantiate each transition
	;;     from TIME to TIME + 1.
	    
	;; o Encoding the transitions from TIME to TIME+1 as elements
	;;   of an Opsat Description.
	;;   if TIME is NOT the last time step, then make the description;
	;;   otherwise, ignore the transitions.
    (multiple-value-bind (transition-opsat-variable-descriptions
						  transition-isat-variable-descriptions
						  transition-constraint-descriptions)

		(if (eql time (- number-of-time-steps 1))
			
			;; At the last time step, we ignore the transitions
			;; and hence add nothing to the OPSAT descriptions
			;; wrt the transitions.
			(values '() '() '())

		  ;; Creates opsat constraints, decision and state variables,
		  ;; for each transition.
		  (convert-transition-descriptions-to-variables-and-constraints
		   transition-descriptions time))

	  ;; o Compose OPSAT Encodings of the State Constraints, Transitions
	  ;;   and variables into an OPSAT description.

	  (let (
;;; The following is now constructed by the caller:			
;;;			(ISAT-variable-descriptions
;;;
;;;			 ;; o Introduce as OPSAT state variables, 
;;;			 ;;   state variables indexed at TIME.
;;;			 (nconc transition-isat-variable-descriptions
;;;					indexed-state-variable-descriptions))

			(ISAT-constraint-descriptions
			 (nconc transition-constraint-descriptions
					indexed-state-constraint-descriptions)))
		
		(values transition-opsat-variable-descriptions
				transition-isat-variable-descriptions
				indexed-state-variable-descriptions
;;;				ISAT-variable-descriptions
				ISAT-constraint-descriptions)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding All Transitions for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-transition-descriptions-to-variables-and-constraints

    (transition-descriptions time)
    
  "Given TRANSITION-DESCRIPTIONS of all transitions of all state variables, 
   creates opsat and isat variables and constraints that represent all 
   possible transitions from TIME to TIME + 1.

   Each TRANSITION-DESCRIPTION is of the form:
       (SOURCE-STATE-VARIABLE-NAME TRANSITION-VARIABLE-NAME
           VARIABLE-TRANSITION-DESCRIPTION+),

   where STATE-VARIABLE-NAME is a defined state variable, 
   TRANSITION-VARIABLE-NAME is a decision variable denoting
   a transition of state variable and
   where VARIABLE-TRANSITION-DESCRIPTION is documented under 
   INITIALIZE-INSTANCE of CONSTRAINT-BASED-TRELLIS-DIAGRAM.

   Returns (values OPSAT-VARIABLE-DESCRIPTIONS
                   ISAT-VARIABLE-DESCRIPTIONS
		           TRANSITION-CONSTRAINT-DESCRIPTIONS)"

  (loop 
	;; These are the accumulated elements of the OpSat description
	;; for the transitions.
	  with variable-isat-variable-descriptions
      with variable-opsat-variable-descriptions
      with variable-transition-constraint-descriptions
      
      for (state-variable-name transition-variable-name
							   . variable-transition-descriptions) 
	  in transition-descriptions
		 
		 ;; Create the Opsat Encoding for all transitions for each
		 ;; state-variable in TRANSITION-DESCRIPTIONS.

      do (multiple-value-setq (variable-isat-variable-descriptions
							   variable-opsat-variable-descriptions
							   variable-transition-constraint-descriptions)
	   
		   (convert-variable-transition-description-to-variables-and-constraints
			state-variable-name transition-variable-name
			variable-transition-descriptions time))
	   
		 ;; Accumulate each element of the Opsat Encoding in its
		 ;; respective description.
		 
      nconcing variable-isat-variable-descriptions 
	  into isat-variable-descriptions

      nconcing variable-opsat-variable-descriptions 
	  into opsat-variable-descriptions

      nconcing variable-transition-constraint-descriptions 
	  into transition-constraint-descriptions

      finally (return (values isat-variable-descriptions	 
							  opsat-variable-descriptions
							  transition-constraint-descriptions))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding All Transitions out of a single Variable for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-variable-transition-description-to-variables-and-constraints

    (variable-name transition-variable-name variable-transition-descriptions time)
  
  "Let VARIABLE-NAME be the name of a state variable, let TRANSITION-VARIABLE-NAME
   denoting a transition from state variable, and let
   VARIABLE-TRANSITION-DESCRIPTIONS describe all transitions associated with variable.

   Create variables and constraints representing 
   all possible transitions of state variable from TIME to TIME + 1.  

   The form of each transition description in VARIABLE-TRANSITION-DESCRIPTIONS
   is documented under the INITIALIZE-INSTANCE method of 
   CONSTRAINT-BASED-TRELLIS-DIAGRAM.

   RETURNS (values OPSAT-VARIABLE-DESCRIPTIONS
                   ISAT-VARIABLE-DESCRIPTIONS
		           TRANSITION-CONSTRAINT-DESCRIPTIONS)"
  
  (let (
		;; Source state variable indexed at TIME.
		(indexed-variable-name 
		 (add-index-to-variable-name variable-name time))

		;; transition variable indexed by TIME. Indicates
		;; which transition from source has been made at TIME.
		(indexed-transition-variable-name
		 (add-index-to-variable-name transition-variable-name time)))

	(loop 

	  ;; Elements of an OpSat description for
	  ;; transitions for a particular source state 
	  ;; variable assignment.
								 
		with source-assignment-transition-isat-domain
		with source-assignment-transition-opsat-domain
		with transition-constraint-descriptions

			 ;; Iterate over each source-value assignment to
			 ;; source VARIABLE-NAME.
		for (source-value . transition-descriptions-from-state)
		in variable-transition-descriptions

		   ;; Index the source assignment.
		for indexed-source-assignment = 
		  (make-assignment :variable-name indexed-variable-name
						   :value source-value)
		
		  ;; Generate the data needed to create elements of the OpSAT description for
		  ;; al transitions from VARIABLE-NAME assigned SOURCE-VALUE.
		do (multiple-value-setq (source-assignment-transition-opsat-domain
								 source-assignment-transition-isat-domain
								 transition-constraint-descriptions)

			 (convert-state-transition-descriptions-to-variables-and-constraints 
			  indexed-source-assignment 
			  indexed-transition-variable-name
			  transition-descriptions-from-state
			  time))

		nconcing source-assignment-transition-opsat-domain
		into transition-variable-opsat-domain
			 
		nconcing source-assignment-transition-isat-domain
		into transition-variable-isat-domain
			 
		   ;; Accumulate the constraints encoding the transitions.
		nconcing transition-constraint-descriptions
		into variable-transition-constraint-descriptions

			 ;; Once all transitions for VARIABLE-NAME at TIME
			 ;; have been created, return the accumulated OPSAT descriptions.
		finally (let (;; Create both OPSAT and ISAT descriptions of transition-variable,
					  ;; where the domain of transition-variable denotes all transitions from 
					  ;; source with source-variable assigned source-value.
					  (opsat-variable-descriptions
					   (list (make-decision-variable-domain-description
							  :variable-name indexed-transition-variable-name
							  :domain 
							  transition-variable-opsat-domain)))

					  (isat-variable-descriptions
					   (list (make-state-variable-domain-description
							  :variable-name indexed-transition-variable-name
							  :domain 
							  transition-variable-isat-domain))))
		  
				  (return (values opsat-variable-descriptions
								  isat-variable-descriptions
								  variable-transition-constraint-descriptions))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding the Transitions from a Single Variable Assignment for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-state-transition-descriptions-to-variables-and-constraints 

    (indexed-source-assignment 
     indexed-transition-variable-name
     transition-descriptions-from-state
     time)
  
  "Given state INDEXED-SOURCE-ASSIGNMENT at TIME, 
         a transition variable INDEXED-TRANSITION-VARIABLE-NAME, and  
         a TRANSITION-DESCRIPTIONS-FROM-STATE of all transitions from state, 
   creates variables and constraints representing these transitions from TIME
   to TIME + 1.  

   Returns (values OPSAT-VARIABLE-DESCRIPTIONS
                   ISAT-VARIABLE-DESCRIPTIONS
  		           TRANSITION-CONSTRAINT-DESCRIPTIONS)."
  
  (loop 
	;; Encode iteratively each of the non-deterministic (probabilistic) transitions
	;; from state (location) INDEXED-TRANSITION-VARIABLE-NAME
	  for (transition-value transition-cost . target-descriptions)
      in transition-descriptions-from-state
		 
      for indexed-transition-assignment = 
		(make-assignment :variable-name indexed-transition-variable-name
						 :value transition-value)

		;; Each non-deterministic transition is labeled by a TRANSITION-VALUE,
		;; Collectively these transition-values, together with those for
		;; all other assignments to the variable of INDEXED-SOURCE-ASSIGNMENT
		;; constitute the domain of the transition variable, whose assignment 
		;; indicates which transition is taken at TIME.
		
		;; Accumulate the OPSAT State Variable transition variable domain.
      collecting transition-value into transition-variable-isat-domain
	
									   ;; Accumulate the OPSAT Decision Variable transition variable domain.
      collecting (make-decision-variable-domain-element-description
				  :value transition-value 
				  :cost transition-cost) 
	  into transition-variable-opsat-domain

		   ;; Accumulate the constraint encoding for each non-deterministic
		   ;; transition, enabled by its corresponding transition variable assignment.
      nconcing (convert-state-transition-target-descriptions-to-constraints
				indexed-source-assignment indexed-transition-assignment
				target-descriptions time)
      into transition-constraint-descriptions
		   
		   ;; Return the accumulated fragments of an OPSAT description.
		   
      finally (return (values transition-variable-opsat-domain
							  transition-variable-isat-domain
							  transition-constraint-descriptions))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding All Transitions from a Single Variable Assignment 
;;; for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-state-transition-target-descriptions-to-constraints

    (indexed-source-assignment 
	 indexed-transition-assignment
     target-descriptions time)

  "Create and return an OPSAT encoding of a set of transitions.

   Transition encodings are returned as PROPOSITIONAL-STATE-CLAUSES
   of an Opsat problem formulation. 

   All transitions occur from TIME to TIME + 1, 
   They start at INDEXED-SOURCE-ASSIGNMENT (at TIME), and 
   are selected by INDEXED-TRANSITION-ASSIGNMENT (at TIME).

   TARGET-DESCRIPTIONS describe a set of target assignments
   that hold at TIME + 1 when their corresponding GUARD assignment literals
   hold at TIME.

   In particular, TARGET-DESCRIPTIONS is a list of elements:

      (GUARD TARGET-ASSIGNMENT)

    where TARGET-ASSIGNMENT is a list of state assignments that
          holde at TIME + 1 when transition is taken, and
          GUARD is a list of zero or more state assignment literals whose conjunction
          must be satisfied by the system state at TIME, in order for the 
          transition to be enabled and taken."

  (loop 
	  for (guard target-assignment) in target-descriptions

										;; Parameterize the guard condition by TIME and the 
										;; target assignment by TIME + 1.
      for indexed-guard = 
		(add-index-to-state-literal-descriptions guard time)

      for indexed-target-assignment = 
		(add-index-to-state-literal-descriptions target-assignment (1+ time))
		      
		;; Build and return the clauses denoting the transition
		;; from source-assignment to target-assignments,
		;; enabled by transition-assignment and guard.
      nconcing (create-transition-constraints 
				indexed-source-assignment
				indexed-transition-assignment
				indexed-guard 
				indexed-target-assignment)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Encoding a Single Guarded Transition from a Single Variable Assignment 
;;; for Time to Time + 1
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Be consistent about how clauses are indicated.  Should guard
;;;   start with :clause?  If so we need to remove this keyword
;;;   upon input.

;;; o Do we want to handle other types of state constraints and guard
;;;   conditions?  For example, JHU-APL used network flow equations as
;;;   state constraints.  Do we want to permit something similar?
(defun create-transition-constraints

    (indexed-source-assignment

     indexed-transition-assignment
     indexed-guard 

     indexed-target-assignment)
  
  "Let INDEXED-SOURCE-ASSIGNMENT and INDEXED-TRANSITION-ASSIGNMENT each be a single
   assignment that is indexed by some TIME, let INDEXED-TARGET-ASSIGNMENT be a list of
   INDEXED ASSIGNMENTS that are indexed by TIME + 1, and let GUARD be a list of zero or more
   INDEXED ASSIGNMENT LITERALS indexed by TIME (denoting a conjunction of the literals).
   An indexed assignment is an assignment to a defined state variable in which
   its state variable is a function of a time step.

   Returns a list of constraints (CURRENTLY PROPOSITIONAL-STATE-CLAUSES) encoding 
   a transition from INDEXED-SOURCE-ASSIGNMENT at TIME to next state assignments 
   INDEXED-TARGET-ASSIGNMENT at TIME + 1.

   A transition is SELECTED if INDEXED-TRANSITION-ASSIGNMENT is satisfied.
   A transition is ENABLED if it is SELECTED and INDEXED-GUARD is satisfied.
   If a transition is ENABLED, then its INDEXED-TARGET-ASSIGNMENT holds at
   the next time step.

   TRANSITION-ASSIGNMENT can represent a decision-theoretic,
   probabilistic or non-deterministic choice of that transtion.

   Creates and returns a constraint encoding of transition as a list of
   PROPOSITIONAL-STATE-CLAUSES."

  ;; A transition is captured through the following two implications,
  ;; (which are converted to clauses):
  (nconc

   ;; 1) IF the transition is selected (transition-assignment holds) 
   ;;      AND the guard is satisfied,
   ;;    THEN the target assignments holds.
   (implication-to-clause-descriptions
    (cons indexed-transition-assignment
		  indexed-guard)
    indexed-target-assignment)
   
   ;; 2) The transition can not be selected (transition-assignment holds)
   ;;    UNLESS the source-assignment holds.
   (implication-to-clause-descriptions
    (list indexed-transition-assignment)
    (list indexed-source-assignment))
   
   ;; 3) The current semantics allows for a transition to be selected,
   ;;    but for its guard not to hold, in which case the target assignment
   ;;    does not hold.
   ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Adding a Time Step Index to the Variables of Each Description
;;; corresponding to time INDEX
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun add-index-to-variable-descriptions 
	
	(variable-descriptions index)

  "Given a list of VARIABLE-DESCRIPTIONS, each of the form
     (VARIABLE-NAME . ATTRIBUTES), 

   Returns an equivalent list, with INDEX added to the
   VRIABLE-NAME of each description, that is:

     ((VARIABLE-NAME INDEX) . ATTRIBUTES)"
  
  (loop for (variable-name . attributes) in variable-descriptions
      for indexed-variable-name = 
		(add-index-to-variable-name variable-name index) 
      collecting (cons indexed-variable-name attributes)))

(defun add-index-to-state-constraint-descriptions

    (state-constraint-descriptions time)

  "Let STATE-CONSTRAINT-DESCRIPTIONS be a list of state clauses,
   then create a list of indexed STATE-CONSTRAINT DESCRIPTIONS,
   by adding the time index TIME to all state variables in
   STATE-CONSTRAINT-DESCRIPTION."
    
  (loop 
	  for state-constraint-description in state-constraint-descriptions

      for indexed-state-constraint-description
      = (add-index-to-state-constraint-description state-constraint-description
												   time)

      collecting indexed-state-constraint-description))

(defun add-index-to-state-constraint-description

    (state-constraint-description time)

  "Given a list of state literals STATE-CONSTRAINT-DESCRIPTION,
   Returns an indexed STATE-CONSTRAINT-DESCRIPTION
   that adds an index of TIME to all state variables in
   STATE-CONSTRAINT-DESCRIPTION."

  ;; Deconstruct each literal into its elements ...
  ;; and reassemble as an indexed literal.
  (let* ((literals (clause-description-literals state-constraint-description))
		 (indexed-literals (add-index-to-state-literal-descriptions literals time)))
	(make-clause-description indexed-literals)))

(defun add-index-to-state-literal-descriptions

    (state-literal-descriptions time)

  "Given a list of STATE-LITERAL-DESCRIPTIONs,
   Returns a lsit o indexed STATE-LITERAL-DESCRIPTIONs
   that adds an index of TIME to the state variable in
   each STATE-LITERAL-DESCRIPTION."
  
  (loop for literal in state-literal-descriptions
	  collecting (add-index-to-state-literal-description
				  literal
				  time)))

(defun add-index-to-state-literal-description

    (state-literal-description time)

  "Given a STATE-LITERAL-DESCRIPTION,
   Returns an indexed STATE-LITERAL-DESCRIPTION
   that adds an index of TIME to the state variable in
   STATE-LITERAL-DESCRIPTION."
  
  ;; Deconstruct literal into its elements ...
  ;; and reassemble as an indexed literal.
  (let* ((state-assignment-description (proposition-of-literal state-literal-description))
		 (indexed-assignment-description  (add-index-to-assignment
										   state-assignment-description time)))
	(if (negation? state-literal-description)
		(make-negation 
		 indexed-assignment-description)
	  indexed-assignment-description)))

;;; FIX:
;;; - This function is redundant with add-index-to-assignment,
;;;   remove when we are sure that it has no callers.
;;;(defun add-index-to-state-assignment-description
;;;
;;;    (state-assignment-description time)
;;;
;;;  "Given a STATE-ASSIGNMENT-DESCRIPTION,
;;;   Returns an indexed STATE-ASSIGNMENT-DESCRIPTION
;;;   that adds an index of TIME to the state variable in
;;;   STATE-ASSIGNMENT-DESCRIPTION.  A STATE-ASSIGNMENT-DESCRIPTION
;;;   is of the form (= variable-name value)."
;;;  
;;;  ;; Deconstruct assignment into its elements ...
;;;  ;; and reassemble as an indexed assignment.
;;;	
;;;  (make-indexed-assignment :VARIABLE-NAME (assignment-variable 
;;;										   state-assignment-description)
;;;							 
;;;						   :VALUE (assignment-value 
;;;								   state-assignment-description)
;;;							 
;;;						   :index time))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Convert an Implication to an Equivalent Clause
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun implication-to-clause-descriptions

    (antecedent-literals consequent-literals)

  "Given an implication:
      ANTECEDENT-LITERALS IMPLIES CONSEQUENT-LITERALS, 

   Returns an equivalent list of CLAUSE-DESCRIPTIONS,
   where each CLAUSE-DESCRIPTION is a list of state-literals.  

   ANTECEDENT-LITERALS and CONSEQUENT-LITERALS, are both lists of 
   state literals, each denoting a conjunctive clause.
   The proposition of each state literal is an assignment 
   of a state variable to an element of its domain.

   A CLAUSE-DESCRIPTION is generated for for each CONSEQUENT-LITERAL, 
   and includes the negation of all the ANTECEDENT-LITERALS."

  (let ((negated-antecedent-literals
		 (loop for antecedent-literal in antecedent-literals
			 collecting (negate-literal antecedent-literal))))
	
    (loop for consequent-literal in consequent-literals
		collecting (make-clause-description
					(cons consequent-literal negated-antecedent-literals)))))
