;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

(in-package :OPSAT)

;;; Copyright (c) 1999 - 2009 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;                     OPSAT Testing with                        ;;;
;;;                   Constraint-based Trellis                    ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Add logging functions.
;;;(deftest-suite :OPITC
;;;;;;    (log-OpSAT-statistics)
;;;;;;  (stop-and-print-OpSAT-statistics)
;;;    )

(defvar *test* nil "Test problem")

;;; FIX:
;;; o For some reason there are two events named START.
;;;   One of these is placed on the queue, presumably during
;;;   initialization.  The second has many successors and predecessors,
;;;   and is created when mentioned in a constraint.  
;;;   o Why isn't this event hashed to the first Start?
(deftest (test20 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; constraint to satisfy.  The best candidate satisfies 
  ;; the constraint.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))))
		 (constraints '((:start start :end e1 :lb 6 :ub 6)
						(:start start :end e1 :lb 4 :ub 5
						 :guard ((= a hi)))
						(:start start :end e1 :lb 0 :ub 20
						 :guard ((= a low)))))

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  constraints
						  :constraint-system-type 'ITC-STN
						  :source 'test20
						  :name 'test20
						  )))
	(setq *test* search-problem)
;;;	(let ((soln (find-solutions search-problem)))
;;;	  (format t "~% Expects ((= A low)), found ~a.  Check correctness."
;;;			  soln)
;;;	  (setq *soln* soln))
    (test '(((= A low)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)},
                      constraints: [e1,e2] for 6
                                   A=hi  implies [e1,e2] for [4,5]
                                   A=low implies [e1,e2] for [0,20]")
	
	))

(deftest (test20-1 :opsat) ()

  ;; An optimal CSP for explanation generation tests.
  
  (let* ((decision-variable-descriptions 
          '(            
            (e012 :domain ((IN :cost 90) (OUT :cost 10)))
            (e1213 :domain ((IN :cost 80) (OUT :cost 10)))
            (e1314 :domain ((IN :cost 70) (OUT :cost 10))) 
            
            ))
         
         (state-variable-descriptions 
          '(                        
            (e012 :domain (IN OUT))
            (e1213 :domain (IN OUT))
            (e1314 :domain (IN OUT))
 
            ))
         
         (constraints '((:start start :end e1 :lb 0 :ub 10)                                               
                        
                        (:start start :end e12 :lb 2 :ub 2 :guard ((= e012 IN)))
                        (:start start :end e12 :lb 0 :ub 10000 :guard ((= e012 OUT)))
                        
                        (:start e12 :end e13 :lb 2 :ub 2 :guard ((= e1213 IN)))
                        (:start e12 :end e13 :lb 0 :ub 10000 :guard ((= e1213 OUT)))
                        
                        (:start e13 :end e1 :lb 2 :ub 2 :guard ((= e1314 IN)))
                        (:start e13 :end e1 :lb 0 :ub 10000 :guard ((= e1314 OUT)))
                                                
                        ))

         (search-problem (new-opsat-problem
                          decision-variable-descriptions
                          state-variable-descriptions
                          constraints
                          :constraint-system-type 'ITC-STN
                          :source 'test20-1
                          :name 'test20-2
                          :conditional? T
                    )))
    
    (test '(((= E1314 IN) (= E1213 IN) (= E012 IN)))
		  (find-solutions search-problem :maximum-solution-count 1)
		  :test #'equal-model-sequences
		  :fail-info "Best solution: ((E1314=IN E1213=IN E012=IN))")
	
))

(deftest (test20-2 :opsat) ()

  ;; An optimal CSP for explanation generation tests.
 
  (let* ((decision-variable-descriptions 
          `((UorD :domain ((UP :cost 300) (MID :cost 100) (DOWN :cost 10)) :guard ,*always-active*)
            (e01 :domain ((IN :cost 10000) (OUT :cost 0)) :guard ,*always-active*)
            
            (e012 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD UP))
            (e1213 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD UP))
            (e131 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD UP))
            
            (e02 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD DOWN))
            (e23 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD DOWN))
            (e31 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD DOWN))
            
            (e022 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD MID))
            (e2223 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD MID))
            (e231 :domain ((IN :cost 10) (OUT :cost 0)) :guard (= UorD MID))
            

            
            ))
         
         (state-variable-descriptions 
          '((UorD :domain (DOWN MID UP))
            (e01 :domain (IN OUT))
            
            (e02 :domain (IN OUT))
            (e23 :domain (IN OUT))
            (e31 :domain (IN OUT))
            
            (e012 :domain (IN OUT))
            (e1213 :domain (IN OUT))
            (e131 :domain (IN OUT))
            
            (e022 :domain (IN OUT))
            (e2223 :domain (IN OUT))
            (e231 :domain (IN OUT))
 
            ))
         
         (constraints '((:start start :end e1 :lb 0 :ub 4 :guard ((= e01 IN)))
                        (:start start :end e1 :lb 0 :ub 100 :guard ((= e01 OUT)))
                        
                        (:start start :end e2 :lb 2 :ub 2 :guard ((= e02 IN)))
                        (:start start :end e2 :lb 0 :ub 10000 :guard ((= e02 OUT)))
                        
                        (:start e2 :end e3 :lb 2 :ub 2 :guard ((= e23 IN)))
                        (:start e2 :end e3 :lb 0 :ub 10000 :guard ((= e23 OUT)))
                        
                        (:start e3 :end e1 :lb 2 :ub 2 :guard ((= e31 IN)))
                        (:start e3 :end e1 :lb 0 :ub 10000 :guard ((= e31 OUT)))
                        
                        
                        
                        (:start start :end e12 :lb 2 :ub 2 :guard ((= e012 IN)))
                        (:start start :end e12 :lb 0 :ub 10000 :guard ((= e012 OUT)))
                        
                        (:start e12 :end e13 :lb 2 :ub 2 :guard ((= e1213 IN)))
                        (:start e12 :end e13 :lb 0 :ub 10000 :guard ((= e1213 OUT)))
                        
                        (:start e13 :end e1 :lb 2 :ub 2 :guard ((= e131 IN)))
                        (:start e13 :end e1 :lb 0 :ub 10000 :guard ((= e131 OUT)))
                        
                        
                        
                        (:start start :end e22 :lb 2 :ub 2 :guard ((= e022 IN)))
                        (:start start :end e22 :lb 0 :ub 10000 :guard ((= e022 OUT)))
                        
                        (:start e22 :end e23 :lb 2 :ub 2 :guard ((= e2223 IN)))
                        (:start e22 :end e23 :lb 0 :ub 10000 :guard ((= e2223 OUT)))
                        
                        (:start e23 :end e1 :lb 2 :ub 2 :guard ((= e231 IN)))
                        (:start e23 :end e1 :lb 0 :ub 10000 :guard ((= e231 OUT)))
                        
                        
                        ))
         

         
         (search-problem (new-opsat-problem
                          decision-variable-descriptions
                          state-variable-descriptions
                          constraints
                          :constraint-system-type 'ITC-STN
                          :source 'test20-2
                          :name 'test20-2
                          :conditional? T
                          :conflict-directed? T
                          )))

    
  (test '(((= E01 IN) (= UORD UP) (= E131 IN) (= E1213 IN) (= E012 OUT))
		  ((= E01 IN) (= UORD UP) (= E131 IN) (= E1213 OUT) (= E012 IN)))
		(find-solutions search-problem :maximum-solution-count 2)
		:test #'equal-model-sequences
		:fail-info "Top 2 solutions: ((= E01 IN) (= UORD UP) (= E131 IN) (= E1213 IN) (= E012 OUT)),
((= E01 IN) (= UORD UP) (= E131 IN) (= E1213 OUT) (= E012 IN))")
	
))

(deftest (test40 :opsat) ()

  ;; A conditional optimal CSP with two unconditional 
  ;; decision variables and one constraint to satisfy.  
  ;; The best candidate satisfies the constraint.

  (let* ((decision-variable-descriptions 
		  `((a :domain ((hi :cost 3)(low :cost 2))
			   :guard ,*always-active*)))
		 (state-variable-descriptions 
		  '((a :domain (hi low))			))
		 (constraints '((:start start :end e1 :lb 6 :ub 6)
						(:start start :end e1 :lb 4 :ub 5
						 :guard ((= a hi)))
						(:start start :end e1 :lb 0 :ub 20
						 :guard ((= a low)))))

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  constraints
						  :constraint-system-type 'ITC-STN
						  :conditional? t
						  :source 'test40
						  :name 'test40
						  )))
	(setq *test* search-problem)

    (test '(((= A low)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)},
                      constraints: [e1,e2] for 6
                                   A=hi  implies [e1,e2] for [4,5]
                                   A=low implies [e1,e2] for [0,20]")
	
	))


(deftest (test41 :opsat) ()

  ;; A conditional optimal CSP, generated from a simple TPN test.tpn in mtk;tpn.

  (let* ((decision-variable-descriptions 
		  '((22 :DOMAIN ((1 :cost 0) (2 :cost 0) (3 :cost 0)) 
				:GUARD TRUE)))
		 (state-variable-descriptions 
		  '((22 :DOMAIN (1 2 3))))
		 (constraints
		  '((:START 3 :END 1 :LB 0 :UB 0 :GUARD TRUE)
			(:START 5 :END 3 :LB 0 :UB 0 :GUARD TRUE)
			(:START 4 :END 5 :LB 0 :UB :infinity :GUARD TRUE)
			(:START 15 :END 3 :LB 0 :UB 0 :GUARD TRUE)
			(:START 14 :END 15 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 9 :END 14 :LB 0 :UB 0 :GUARD TRUE)
			(:START 11 :END 9 :LB 0 :UB 0 :GUARD TRUE)
			(:START 10 :END 11 :LB 0 :UB 0 :GUARD TRUE)
			(:START 13 :END 9 :LB 0 :UB 0 :GUARD TRUE)
			(:START 12 :END 13 :LB 0 :UB 1 :GUARD TRUE)
			(:START 12 :END 13 :LB 0 :UB 1 :GUARD TRUE)
			(:START 8 :END 9 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 8 :END 12 :LB 0 :UB 0 :GUARD TRUE)
			(:START 8 :END 10 :LB 0 :UB 0 :GUARD TRUE)
			(:START 7 :END 8 :LB 0 :UB 0 :GUARD TRUE)
			(:START 6 :END 7 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 2 :END 6 :LB 0 :UB 0 :GUARD TRUE)
			(:START 2 :END 4 :LB 0 :UB 0 :GUARD TRUE)
			(:START 17 :END 1 :LB 0 :UB 0 :GUARD TRUE)
			(:START 19 :END 17 :LB 0 :UB 0 :GUARD TRUE)
			(:START 18 :END 19 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 29 :END 17 :LB 0 :UB 0 :GUARD TRUE)
			(:START 28 :END 29 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 23 :END 28 :LB 0 :UB 0 :GUARD TRUE)
			(:START 21 :END 22 :LB 0 :UB 0 :GUARD TRUE)
			(:START 20 :END 21 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 16 :END 20 :LB 0 :UB 0 :GUARD TRUE)
			(:START 16 :END 18 :LB 0 :UB 0 :GUARD TRUE)
			(:START 0 :END 1 :LB 0 :UB 10 :GUARD TRUE)
			(:START 0 :END 16 :LB 0 :UB 0 :GUARD TRUE)
			(:START 0 :END 2 :LB 0 :UB 0 :GUARD TRUE)
			(:START 25 :END 23 :LB 0 :UB 0 :GUARD ((= 22 1)))
			(:START 24 :END 25 :LB 0 :UB 0 :GUARD ((= 22 1)))
			(:START 22 :END 24 :LB 0 :UB 0 :GUARD ((= 22 1)))
			(:START 27 :END 23 :LB 0 :UB 0 :GUARD ((= 22 2)))
			(:START 26 :END 27 :LB 1 :UB 2 :GUARD ((= 22 2)))
			(:START 26 :END 27 :LB 1 :UB 2 :GUARD ((= 22 2)))
			(:START 22 :END 26 :LB 0 :UB 0 :GUARD ((= 22 2)))
			(:START 22 :END 23 :LB 0 :UB :INFINITY :GUARD ((= 22 3)))))

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  constraints
						  :constraint-system-type 'ITC-STN
						  :conditional? t
						  :source 'test41
						  :name '(:CHOICE-SELECTION <TPN   main.run   [0 1]>)
						  )))

    (test '(((= 22 3))
			((= 22 2))
			((= 22 1)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "22 = {(1 0),(2 0), (3 0)},
                      constraints: see code")
	))

;;; FIX:
;;; o Returns nil when maximum-count is one, but works for 2 or more.
;;;;  Fix for count of 1.
(deftest (test42 :opsat) ()

  ;; A conditional optimal CSP, generated from a simple TPN.

  (let* ((decision-variable-descriptions 
		  '((22 :DOMAIN ((1 :COST 0) (2 :COST 0) (3 :COST 0)) :GUARD TRUE)))
		 (state-variable-descriptions 
		  '((22 :DOMAIN (1 2 3))))
		 (constraints
		  '((:START 17 :END 1 :LB 0 :UB 0 :GUARD TRUE)
			(:START 29 :END 17 :LB 0 :UB 0 :GUARD TRUE)
			(:START 28 :END 29 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 23 :END 28 :LB 0 :UB 0 :GUARD TRUE)
			(:START 21 :END 22 :LB 0 :UB 0 :GUARD TRUE)
			(:START 20 :END 21 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 19 :END 17 :LB 0 :UB 0 :GUARD TRUE)
			(:START 18 :END 19 :LB 0 :UB :INFINITY :GUARD TRUE)
			(:START 16 :END 18 :LB 0 :UB 0 :GUARD TRUE)
			(:START 16 :END 20 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 3 :END 1 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 15 :END 3 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 14 :END 15 :LB 0 :UB :INFINITY :GUARD TRUE) 
			(:START 9 :END 14 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 13 :END 9 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 12 :END 13 :LB 0 :UB 1 :GUARD TRUE) 
			(:START 12 :END 13 :LB 0 :UB 1 :GUARD TRUE) 
			(:START 11 :END 9 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 10 :END 11 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 8 :END 10 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 8 :END 12 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 8 :END 9 :LB 0 :UB :INFINITY :GUARD TRUE) 
			(:START 7 :END 8 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 6 :END 7 :LB 0 :UB :INFINITY :GUARD TRUE) 
			(:START 5 :END 3 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 4 :END 5 :LB 0 :UB :INFINITY :GUARD TRUE) 
			(:START 2 :END 4 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 2 :END 6 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 0 :END 2 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 0 :END 16 :LB 0 :UB 0 :GUARD TRUE) 
			(:START 0 :END 1 :LB 0 :UB 10 :GUARD TRUE) 
			(:START 22 :END 23 :LB 0 :UB :INFINITY :GUARD ((= 22 1))) 
			(:START 27 :END 23 :LB 0 :UB 0 :GUARD ((= 22 2))) 
			(:START 26 :END 27 :LB 1 :UB 2 :GUARD ((= 22 2))) 
			(:START 26 :END 27 :LB 1 :UB 2 :GUARD ((= 22 2))) 
			(:START 22 :END 26 :LB 0 :UB 0 :GUARD ((= 22 2))) 
			(:START 25 :END 23 :LB 0 :UB 0 :GUARD ((= 22 3))) 
			(:START 24 :END 25 :LB 0 :UB 0 :GUARD ((= 22 3))) 
			(:START 22 :END 24 :LB 0 :UB 0 :GUARD ((= 22 3))) 
			))

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  constraints
						  :constraint-system-type 'ITC-STN
						  :conditional? t
						  :source '(TPN   main.run (0 1))
						  :name (LIST :CHOICE-SELECTION '(TPN   main.run 0 1))
						  )))
	
	(setq *test* search-problem)
    (test '(((= 22 3))
			((= 22 2))
			((= 22 1))
			)
		  (find-solutions search-problem :maximum-solution-count 3)
		  :test #'equal-model-sequences
		  :fail-info "Top 3 solutions,((= 22 3)),((= 22 2)),((= 22 1))
                      constraints: see code")
	))


;; (deftest (test43 :opsat) ()

;;   ;; A conditional optimal CSP with two unconditional 
;;   ;; decision variables and one constraint to satisfy.  
;;   ;; The best candidate satisfies the constraint.

;;   (let* ((decision-variable-descriptions 
;; 		  `((a :domain ((hi :cost 3)(low :cost 2))
;; 			   :guard ,*always-active*)))
;; 		 (state-variable-descriptions 
;; 		  '((a :domain (hi low))			))
;; 		 (constraints '((:start start :end e1 :lb 6 :ub 6)
;; 						(:start start :end e1 :lb 4 :ub 5
;; 						 :guard (= a hi) :lb-relaxable? t :lb-cost-ratio 10)
;; 						(:start start :end e1 :lb 0 :ub 20
;; 						 :guard (= a low))))



;; 		 (search-problem (new-opsat-problem
;; 						  decision-variable-descriptions
;; 						  state-variable-descriptions
;; 						  constraints
;; 						  :constraint-system-type 'ITC-STN
;; 						  :conditional? t
;; 						  :conflict-directed? t
;; 						  :source 'test43
;; 						  :name 'test43
;; 						  :allow-relaxation? t
;; 						  )))

;; 	(format t "Complete setting up problem defs ~A~%" (type-of search-problem))

;; 	;; (loop for constraint in constraints do
;; 	;; 	 (format t "Can be relaxed? ~A~%" (CS::constraint-can-be-relaxed? constraint)))

;; 	(format t "Solutions ~A~%" (find-solutions search-problem))

	
;; 	))
