(in-package :observer)

(defclass event-publisher ()
  ((listeners
	:initform (make-hash-table :test #'eql)
	:reader event-listeners
	:documentation "A hash table representing all the listeners for events on
                    this object. Key is event name, value is a list of 
                    function handles."))
  (:documentation "A superclass for an object that can fire events to listeners."))

(defmethod fire ((obj event-publisher) event &rest args)
  "First looks up all listeners that are registered to EVENT.
   Then, calls the given function handle with ARGS."
  (loop for listener in (gethash event (event-listeners obj))
	 do (apply listener obj args)))

(defmethod register-listener ((obj event-publisher) event listener)
  "Add LISTENER to the list of listeners called for EVENT on OBJ."
  (pushnew listener (gethash event (event-listeners obj) '())))

(defmethod remove-listener ((obj event-publisher) event listener)
  "Remove LISTENER from the list of listeners for EVENT on OBJ."
  (setf (gethash event (event-listeners obj)) 
		(delete listener (gethash event (event-listeners obj) '()))))