;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-qsp)

(defun make-qsp-bus ()
  (let* ((qsp (make-instance 'qsp)))
    (add-constraints (qsp-constraint-description-bus) qsp)
    qsp))

(defun make-qsp-bus-driver ()
  (let* ((qsp (make-instance 'qsp)))
    (add-constraints (qsp-constraint-description-bus-driver) qsp)
    qsp))

(defun make-qsp-bus-driver-valve ()
  (let* ((qsp (make-instance 'qsp)))
    (add-constraints (qsp-constraint-description-bus-driver-valve) qsp)
    qsp))


(defun qsp-constraint-description-bus ()
  `((:scope (0 1) :relation ((:lower-bound (0) :upper-bound (10))))
    (:scope (1 2) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (bus) :relation ((= bus on))))
    (:scope (2 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 4) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (bus) :relation ((= bus off))))
    (:scope (4 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (0 5) :relation ((:lower-bound (100) :upper-bound (250))))
    ))

(defun qsp-constraint-description-bus-driver ()
  `((:scope (0 1) :relation ((:lower-bound (0) :upper-bound (10))))
    (:scope (1 2) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (driver) :relation ((= driver on))))
    (:scope (2 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 4) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (driver) :relation ((= driver off))))
    (:scope (4 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (0 5) :relation ((:lower-bound (100) :upper-bound (250))))
    (:scope (1 6) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (6 7) :relation ((:lower-bound (0) :upper-bound (,*inf*)))
	    :state-constraint (:scope (bus) :relation ((= bus off))))
    (:scope (6 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (7 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (8 9) :relation ((:lower-bound (0 :inequality ,#'>) 
					   :upper-bound (,*inf*)))
	    :state-constraint (:scope (bus) :relation ((= bus off))))
    (:scope (9 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    ))

(defun qsp-constraint-description-bus-driver-valve ()
  `((:scope (0 1) :relation ((:lower-bound (0) :upper-bound (10))))
    (:scope (1 2) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (valve) :relation ((= valve opened))))
    (:scope (2 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 4) :relation ((:lower-bound (100) :upper-bound (101)))
	    :state-constraint (:scope (valve) :relation ((= valve closed))))
    (:scope (4 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (0 5) :relation ((:lower-bound (100) :upper-bound (250))))
    (:scope (1 6) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (6 7) :relation ((:lower-bound (0) :upper-bound (,*inf*)))
	    :state-constraint (:scope (driver) :relation ((= driver off))))
    (:scope (7 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (6 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (8 9) :relation ((:lower-bound (0 :inequality ,#'>) 
					   :upper-bound (,*inf*)))
	    :state-constraint (:scope (driver) :relation ((= driver off))))
    (:scope (9 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (1 10) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (10 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (10 11) :relation ((:lower-bound (0) :upper-bound (,*inf*)))
	    :state-constraint (:scope (bus) :relation ((= bus off))))
    (:scope (11 12) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (3 12)  :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    (:scope (12 13) :relation ((:lower-bound (0 :inequality ,#'>) 
					     :upper-bound (,*inf*)))
	    :state-constraint (:scope (bus) :relation ((= bus off))))
    (:scope (13 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
    ))


