;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :TPN)

;;; ***********************************************************************
;;;
;;;                Testing Incremental Temporal Consistency
;;;
;;; ***********************************************************************

;;; Last update: September 29th, 2010.

;;; ***********************************************************************
;;;
;;;                Problems to Fix
;;;
;;; ***********************************************************************

;;; Fix:
;;; o For some reason parse-tpn-file doesn't work when
;;;   run-tests is called from :cl-user, unless the package
;;;   is forced to :tpn.

;;;   It could be that routines in parse-tpn-file are dependent
;;;   on *package*, for example during reading or interning, and
;;;   need to be in the :tpn package.  Track down and make package
;;;   independent.

;;; ***********************************************************************
;;;
;;;                TPN Choice Selection Tests
;;;
;;; ***********************************************************************

(deftest-suite :TPN
;;;    (log-TPN-statistics)
;;;  (stop-and-print-TPN-statistics)
	)

(deftest (test1 :TPN) ()

  ;; Simple TPN Test.

  (let ((original-package *package*))

	;; parse-tpn-file only seems to work when the package is forced to TPN.
	;;
	;; Reply...
	;; DAVID WANG: This is because instead of checking that an XML tag has a given name,
	;;             the code checks if an XML tag matches a symbol. For example,
	;;             (eql (made-up-get-xml-tag-function xml) 'TPN)
	;;             This is bad because symbols are associated with packages.
	;;             If the "xml" variable is made by parsing a file in a package outside of
    ;;             package #:tpn, then we will be checking that these two symbols:
	;;             "some-package:TPN"  and  "tpn:TPN" are "eql".
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file "mtk:TPN;Examples;test.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))

	  ;; (initialize-tpn-choice-selection! tpn)
	  ;; (format t "TPN best choice ~A~%" (tpn-best-choice tpn))
	  (initialize-tpn-choice-selection! tpn)

	  (test
	   `((= 22 3))
	   (tpn-best-choice tpn)
	   :multiple-values nil
	   :test #'equal-models
	   :fail-info "choice selection -> ((= 22 3)).")
	  )

	;; Restore the calling package.
	(setq *package* original-package)
	))


(deftest (test2 :TPN) ()

  ;; Boeing Demo Scenario 1 (w/o vehicle paths) October, 2010.
  ;; Solution: (= <evt 2> 1)  Go home via Yarmouth.
  ;;    Both options are feasible, its best to go to Yarmouth, then home.
  ;;    Both routes can be traversed in 0 to inf minutes.
  
    ;;   Must get home in 9 mins.

  ;;   Choice 1: Cost = 1     [0,inf] mins:
  ;;     Current -> Yarmouth: [0,inf] mins,
  ;;     Yarmouth -> home:    [0,inf] mins.

  ;;   Choice 2: Cost = 2
  ;;     Current -> home:     [0,inf] mins.

  (let ((original-package *package*))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;BoeingOct2010demo2.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   `((= 2 1))

	   (tpn-best-choice tpn)
   
	   :multiple-values nil
	   :test #'equal-models
	   :fail-info "choice selection -> ((= 2 2)).")

	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))


(deftest (test3 :TPN) ()

  ;; Boeing Demo Scenario 1 (with vehicle paths) October 2010.
  ;; Solution: (= <evt 2> 2)  Must go directly home.
  
  ;;   Must get home in 9 mins.

  ;;   Choice 1: Cost = 1     11 mins:
  ;;     Current -> Yarmouth: 5 mins,
  ;;     Yarmouth -> home:    6 mins.

  ;;   Choice 2: Cost = 2
  ;;     Current -> home:     8 mins.

  (let ((original-package *package*))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;BoeingOct2010demo3.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   `((= 2 2))

	   (tpn-best-choice tpn)
   
	   :multiple-values nil
	   :test #'equal-models
	   :fail-info "choice selection -> ((= 2 2)).")

	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))

(deftest (test4 :TPN) ()

  ;; Boeing Demo Scenario 1 (with vehicle paths) October 2010.
  ;; No solution.  Both choices are temporally infeasible.

  ;;   Must get home in 7 mins.

  ;;   Choice 1: Cost = 1     11 mins:
  ;;     Current -> Yarmouth: 5 mins,
  ;;     Yarmouth -> home:    6 mins.

  ;;   Choice 2: Cost = 2
  ;;     Current -> home:     8 mins.
  
  (let ((original-package *package*))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;BoeingOct2010demo4.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   '()								; no-solution
	   (tpn-best-choice tpn)
   
	   :multiple-values nil
	   :test #'equal
	   :fail-info "choice selection -> no feasible choice.")

	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))



(deftest (test5 :TPN) ()

  ;; Boeing Demo Scenario 1 (with vehicle paths) October 2010.
  ;; Solution: (= <evt 2> 1)  Go home via Yarmouth.
  ;;    Both options are feasible, its best to go to Yarmouth, then home.

  ;;   Must get home in 13 mins.

  ;;   Choice 1: Cost = 1     11 mins:
  ;;     Current -> Yarmouth: 5 mins,
  ;;     Yarmouth -> home:    6 mins.

  ;;   Choice 2: Cost = 2
  ;;     Current -> home:     8 mins.
  
  (let ((original-package *package*))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;BoeingOct2010demo5.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   `((= 2 1))

	   (tpn-best-choice tpn)
   
	   :multiple-values nil
	   :test #'equal-models
	   :fail-info "choice selection -> ((= 2 1)).")

	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))



(deftest (test6 :TPN) ()

 ;; Boeing Demo Scenario 1 (w/o vehicle paths) October, 2010.

 (let ((original-package *package*))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;BoeingOct2010demo1.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   `((= 2 2))

	   (tpn-best-choice tpn)
   
	   :multiple-values nil
	   :test #'equal-models
	   :fail-info "choice selection -> ((= 2 2)).")

	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; Tests for generating relaxations for over-constrained TPNs


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(deftest (test10 :TPN) ()

  ;; Explanation Test
  ;; Use OPSAT to generate partial solution to over constrained problem

  (let ((original-package *package*)
		(explanations nil))
	
	;; parse-tpn-file only seems to work when the package is forced to TPN.
	(in-package :tpn)

	;; File name explanations:
	;; choice-x-length-y-n.tpn: x indicates the number of choices in the TPN, s for small(~5), m for medium(~10) and l for large(~20)
	;;                          y indicates the number of constraints inside the TPN, s for small(~5 per choice), m for medium(~10 per choice) and l for large(~20 per choice)
	;;                          n indicates the level of overconstraint: 1 for slightlt over constraint, 2 for moderate and 3 for heavy.

	(let ((tpn (first
				(parse-tpn-file #P"mtk:TPN;Examples;PerformanceTest;choice-s-length-s-1.tpn"
								'tpn::get-event 'tpn::make-episode
								'tpn::add-episode 'tpn::make-tpn))))


	  (initialize-tpn-choice-selection! tpn)
	  (setf explanations (tpn-explanations tpn :exp-count 2))

	  (format t "~3% Best Explanations ~a  ~%" (first explanations))
	  
;;	  (format t "Continuous Relaxation ~a   ~3%" (continuous-constraint-relaxation tpn (first explanations)))


	  ;; Restore the calling package.
	  (setq *package* original-package)
	  )))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Tests for trans-decision constraints.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-trans-decision-1 :TPN) ()

  ;; This is a simple test to make sure that constraints that cross
  ;; decision boundaries work.

  ;; Choice 0:
  ;;
  ;;   Option 1: Cost = 100 [5, 10]
  ;;   Option 2: Cost = 10  [5, 10]

  ;; Choice 4:
  ;;
  ;;   Option 1: Cost = 50  [5, 10]
  ;;   Option 2: Cost = 10  [5, 10]

  ;; Constraint:
  ;;
  ;;   [1, 2] Between Beginning of 0=2 and 4=2



  ;; parse-tpn-file only seems to work when the package is forced to TPN.

  (let ((tpn (first
			  (parse-tpn-file #P"mtk:TPN;Examples;spec3;trans-decision-constraints.tpn"))))

	(initialize-tpn-choice-selection! tpn)

	(test
	 `((= 4 1) (= 0 2))

	 (tpn-best-choice tpn)

	 :multiple-values nil
	 :test #'equal-models
	 :fail-info "trans decision constraints not working.")))

(deftest (test-only-activated-decisions-made-simple :tpn) ()
  ;; This tests the Kirk->Opsat->BFIG chain to make sure that Kirk
  ;; reports values for only decisions that are activated. (There was
  ;; a bug that caused some non-activated decision variables to be
  ;; assigned a value.
  (let ((tpn (first
				(parse-tpn-file "mtk:TPN;tests;regression;test-only-activated-decisions-made-simple.tpn"))))
	
	  (initialize-tpn-choice-selection! tpn)
	
	  (test
	   `((= 4 6))

	   (tpn-best-choice tpn)

	   :multiple-values nil
	   :test #'equal-models)))



(deftest (test-loop-flattening :tpn) ()
  ;; Tests the flattening of a looping episode into multiple episodes in series.
  (let ((tpn (first
			  (parse-tpn-file "mtk:TPN;tests;simple-looping.tpn"))))
	
	(let ((episodes-with-loops '() ))
	  (loop for episode in (tpn-episodes tpn)
		 do 
		   (when (episode-contains-loops? episode)
			 (push (list episode 5) episodes-with-loops)))
	  
	  ;; Call function to flatten loops:
	  (with-slots (number-of-episodes)
		  (flatten-tpn-loops! tpn episodes-with-loops)
		

		;; Test:
		(test
		 `8 number-of-episodes
		 
		 :multiple-values nil
		 :test #'eql)))))
