;;;; Copyright (c) 2013

;;;; This file contains methods used to copy an existing TPN into a
;;;; new TPN structure.

(in-package :TPN)

(defun copy-tpn (tpn)
  "Copy the TPN object into a new TPN object. Greedily copies
  everything."
  (copy-thing tpn))

(define-copy-method copy-inner-class progn ((self temporal-plan-network)
											copy-object
											copy-htable)
					(with-slot-copying (copy-object copy-htable self)
					  (copy-slots name
								  start
								  end
								  events
								  event-ht
								  decisions
								  episodes
								  number-of-episodes)
					  (copy-set-slot labeled? nil))
					;; LAST-OF-EPISODES requires some care because
					;; it points to a cons cell that is part of
					;; EPISODES.
					(setf (slot-value copy-object 'last-of-episodes)
						  (loop for c on (slot-value copy-object 'episodes)
							 while (cdr c)
							 finally (return c)))
					)

(define-copy-method copy-inner-class progn ((self tpn-w-selection)
											copy-object
											copy-htable)
					(declare (ignore copy-htable))
					(with-slot-copying (copy-object copy-htable self)
					  (copy-set-slot activation-episode-ht (make-hash-table))
					  (copy-set-slot best-choice nil)
					  (copy-set-slot selection-problem nil)
					  (copy-set-slot temporal-network nil)))

(define-copy-method copy-inner-class progn ((self event)
											copy-object
											copy-htable)
					(with-slot-copying (copy-object copy-htable self)
					  (copy-slots tp
								  name
								  decision?
								  successor-episodes
								  predecessor-episodes)))

(define-copy-method copy-inner-class progn ((self event-w-selection)
											copy-object
											copy-htable)
					(declare (ignore copy-htable))
					(with-slot-copying (copy-object copy-htable self)
					  (copy-set-slot activation-assignment nil)
					  (copy-set-slot activation-label nil)
					  (copy-set-slot itc-event nil)))

(define-copy-method copy-inner-class progn ((self episode)
											copy-object
											copy-htable)
					(with-slot-copying (copy-object copy-htable self)
					  (copy-slots tp
								  index
								  start
								  end
								  lowerbound
								  upperbound
								  temporal-duration
								  cost
								  description
								  support
								  relaxable
								  isstateconstraint
								  istemporalconstraint
								  isactivity
								  isdummy
								  disabled
								  executed
								  controllable
								  mean
								  variance)))

(define-copy-method copy-inner-class progn ((self episode-w-activation-labeling)
											copy-object
											copy-htable)
					(declare (ignore copy-htable))
					(with-slot-copying (copy-object copy-htable self)
					  (copy-set-slot activation-assignment nil)
					  (copy-set-slot activation-assignment-is-conjunction? nil)
					  (copy-set-slot activation-label nil)))


