(in-package #:tpn)

(defvar *dot-path*
  #+(or win32 mswindows) "dot.exe"
  #-(or win32 mswindows) "dot"
  "Path to `dot`")

(defmethod tpn2dot ((tpn-stream stream) (filename t)
					&key 
					(format "pdf") 
					(size "7.5,10") 
					(orientation "portrait")
					(rankdir "LR")
					(dot-path *dot-path*) )
  (tpn2dot (first (parse-tpn-stream tpn-stream
									'tpn::get-event 'tpn::make-episode
									'tpn::add-episode 'tpn::make-tpn))
		   filename
		   :format format
		   :size size
		   :orientation orientation
		   :rankdir rankdir
		   :dot-path dot-path))

(defmethod tpn2dot ((tpn-string string)(filename t)
					&key 
					(format "pdf") 
					(size "7.5,10") 
					(orientation "portrait")
					(rankdir "LR")
					(dot-path *dot-path*) 
					(verbose nil) )
  (tpn2dot (first (parse-tpn-string tpn-string
									'tpn::get-event 'tpn::make-episode
									'tpn::add-episode 'tpn::make-tpn))
		   filename
		   :format format
		   :size size
		   :orientation orientation
		   :rankdir rankdir
		   :dot-path dot-path
		   :verbose verbose))

(defmethod tpn2dot ((tpn tpn::temporal-plan-network) (outfile t)
					&key 
					(format "pdf") 
					(size "10,7.5") 
					(orientation "portrait")
					(rankdir "LR")
					(dot-path *dot-path*)
					(verbose nil))
  "Prints the TPN to the OUTFILE. Where OUTFILE is a pathname."
  (let ((dot-file-name (make-pathname :type "dot" 
									  ;;:directory (pathname-directory outfile)
									  ;;:name (pathname-name outfile)
									  :defaults outfile))
		(command "")
		(exit-code 0))
	;; create the dot file
    (with-open-file (stream dot-file-name :direction :output :if-exists :supersede
			    :if-does-not-exist :create)
	  (dot-print-tpn tpn :stream stream :rankdir rankdir))
	;; create the command
	(setf command
		  (format nil 
				  "\"~0A\" -T~0A -Gdpi=300 -Gsize=\"~0A\" -Gorientation=\"~0A\" -Grankdir=\"~0A\" -o \"~0A\" \"~0A\""
				  dot-path format size orientation rankdir outfile dot-file-name))
	;; print the command
	(when verbose
	  (format t "~a~%" command))
	;; run the command
	(setf exit-code (asdf:run-shell-command command))
	;; print errors
    (when (not (= 0 exit-code))
	  (format t "error: failed to generate dot output for tpn.~%"))
	exit-code))

(defmethod dot-print-tpn ((tpn tpn::temporal-plan-network) &key (stream t) (rankdir nil))
  (format stream "digraph G {~%")
  (when rankdir
	  (format stream "    rankdir=~a~%" rankdir))
  ;;(format stream "    node [shape = circle];~%")
  (loop for event in (tpn::tpn-events tpn) do
	   (format stream "    \"~a\" [shape = ~a, fixedsize = true, width = 0.75, height = 0.75]~%"
			   (tpn::event-name event)
			   (if (find event (tpn::tpn-decisions tpn))
				   "doublecircle"
				   "circle")))

  (loop for episode in (tpn::tpn-episodes tpn) do
	   (let ((primitive (tpn::episode-primitive episode)))
		 (when primitive
		   ;; add in escape characters
		   (setf primitive (format nil "~s" primitive))
		   (setf primitive (subseq primitive 1 (- (length primitive) 1)))
		   ;; format
		   (format stream "    \"~a\" -> \"~a\" [label = \"~a [~a,~a]\"]~%"
				   (tpn::event-name (tpn::episode-start episode))
				   (tpn::event-name (tpn::episode-end episode))
				   primitive
				   (get-pretty-number (tpn::episode-lowerbound episode))
				   (get-pretty-number (tpn::episode-upperbound episode))))
		 (unless primitive
		   ;; format
		   (format stream "    \"~a\" -> \"~a\" [label = \"[~a,~a]\"]~%"
				   (tpn::event-name (tpn::episode-start episode))
				   (tpn::event-name (tpn::episode-end episode))
				   (get-pretty-number (tpn::episode-lowerbound episode))
				   (get-pretty-number (tpn::episode-upperbound episode))))))
  (format stream "}~%"))

(defun get-pretty-number (val)
  "Helper function to pretty-print infinities"
  (cond ((equalp val :INFINITY) "∞")
		((equalp val :-INFINITY) "-∞")
		(t (format nil "~a" val))))
