;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Why do we need a 'size' slot, if we could just have a macro that
;;; read the length of the elements list?

(defclass discrete-domain (domain)
  ((size :initform '()
		 :reader discrete-domain-size
		 :documentation "The number of elements in this domain.")
   (elements		 
			 :accessor discrete-domain-elements
			 :documentation "The discrete elements of this domain.")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructors
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((domain discrete-domain) &key elements)
  "Set the slot size to the length of elements."
  (with-slots (size) domain
	(setf size (length elements))))

(defmacro make-discrete-domain (elements)
  "Syntatic sugar for creation of discrete domains"
  `(make-instance 'discrete-domain :elements ,elements))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod domain-member? ((domain discrete-domain) value)
  "Determine if VALUE is a member of DOMAIN.

   ========================================
   Preconditions
   ========================================
   None.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   T iff VALUE is a member of DOMAIN."
  (with-slots (elements) domain
	(if (member value elements :test #'equal)
		T
		nil)))

(defmethod domain-volume ((domain discrete-domain))
  (slot-value domain 'size))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Print Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod print-object ((domain discrete-domain) stream)
  (print-unreadable-object (domain stream :identity T :type T)
	(princ "Size: " stream)
	(princ (discrete-domain-size domain) stream)
	(princ " Elements: " stream) 
	(princ (discrete-domain-elements domain) stream)))