;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; This file describes the interface for domains.
;;; It contains the base class for distributions and the
;;; definition of generic functions.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defclass domain ()
  ((dimensions :initform 0
			   :initarg :dims
			   :accessor dimensions
			   :documentation "The number of dimensions in this domain.")   
   (elements :initform '()
			 :initarg :elements
			 :accessor domain-elements
			 :documentation "The elements of this domain."))  
  (:documentation "The base class for any domain."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Generic Function Definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric domain-member? (domain value)
  (:documentation
   "Determine if VALUE is a member of the DOMAIN.

    ======================================================================
    Preconditions
    ======================================================================
    None.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    First value: T iff VALUE is a member of DOMAIN. NIL else.
    Second value: Reason for the return value. One of:
        :MEMBER
        :NOT-A-MEMBER
        :OUT-OF-BOUNDS (only for bounded domains)"))

(defgeneric domain-volume (domain)
  (:documentation
   "Calculate the volume covered by DOMAIN.

    ======================================================================
    Preconditions
    ======================================================================
    None.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    Volume covered by DOMAIN. Can be float or integer."))


