;; TO BUILD AN EXECUTABLE (using AllegroCL):
;; 1.) load the package to which this deliver.lisp file belongs
;;     >> (asdf:load-system 'tburton)
;; 3.) Decide which function you want to execute below, and modify its parameters accordingly.
;; 4.) Save this file.
;; 5.) Load this file.
;;     * If you are using emacs with slime: Open this file in emacs. 
;;       Make sure it is selected by clicking on it (active). 
;;       Then use the keyboard commands  Ctrl-c, Ctrl-k to compile the file.
;;     * If on a REPL:
;;       >> (load (mars-pathname "tburton/deliver.lisp"))
;; 6.) On the REPL, evaluate the function you are interested in. i.e.:
;;     >> (tburton::deliver-win)
;;     >> (tburton::deliver-source)
;; Note: There is no cross-compilation capability. 
;;   (deliver-win) must be evaluated within common-lisp on a windows machine.
;;   (deliver-linux64) must be evaluated within common-lisp on a linux 64 bit machine.

;; GENERIC INSTRUCTIONS FOR DELIVER.LISP
;; TO BUILD AN EXECUTABLE (using AllegroCL):
;; 1.) load the package to which this deliver.lisp file belongs
;;     >> (asdf:load-system 'this-package-name)
;; 3.) Decide which function you want to execute below, and modify its parameters accordingly.
;; 4.) Save this file.
;; 5.) Load this file.
;;     * If you are using emacs with slime: Open this file in emacs. 
;;       Make sure it is selected by clicking on it (active). 
;;       Then use the keyboard commands  Ctrl-c, Ctrl-k to compile the file.
;;     * If on a REPL:
;;       >> (load "String with the absolute path to this file.lisp")
;; 6.) On the REPL, evaluate the function you are interested in. i.e.:
;;     >> (this-package-name::deliver-win)
;;     >> (this-package-name::deliver-source)
;; Note: There is no cross-compilation capability. 
;;   (deliver-win) must be evaluated within common-lisp on a windows machine.
;;   (deliver-linux64) must be evaluated within common-lisp on a linux 64 bit machine.

(in-package #:tburton)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (asdf:load-system 'make-executable))

(defun deliver-win ()
  (let (;; PARAMETERS YOU SHOULD SET!
		(;; The name of the executable you wish to build.
		 exec-name "tburton.exe")
		(;; The path where all the of the binaries should be put.
		 exec-path (mars-pathname "tburton/bin/windows/"))
		(;; Path to your common lisp init file - the file that is loaded and executed 
		 ;; automatically by your running lisp instance.
		 init-path "C:/Users/David Wang/clinit.cl")
		;; PARAMETERS YOU SHOULD NOT TOUCH!
		(closopt-fasl (mars-pathname "tburton/closopt.fasl")))
	
	;; Build the executable
	(make-executable:make-executable-from-asdf 
	 exec-name
	 exec-path
	 'tburton
	 'tburton::tburton-main
	 :debug-build nil
	 :init-file-pathname init-path
	 :extra-input-files (if (probe-file closopt-fasl) (list :trace :constructor :foreign closopt-fasl) '())
	 :lisp-heap-start "1000m"
	 ;; :lisp-heap-size "4000m"
	 :aclmalloc-heap-start "4000m"
	 ;; :aclmalloc-heap-size "2000m"
	 )
	))

(defun deliver-linux64 ()
  (let (; PARAMETERS YOU SHOULD SET!
		(;;The name of the executable you wish to build.
		 exec-name "tburton.exe" )
		(;; The path where all the of the binaries should be put.
		 exec-path (mars-pathname "tburton/bin/linux64/"))
		(;; "Path to your common lisp init file - 
		 ;; the file that is loaded and executed 
		 ;; automatically by your running lisp instance.
		 init-path "~/clinit.cl")
		;; PARAMETERS YOU SHOULD NOT TOUCH!
		(closopt-fasl (mars-pathname "tburton/closopt.fasl")))

	;; Build the executable
	(make-executable:make-executable-from-asdf 
	 exec-name
	 exec-path
	 'tburton
	 'tburton::tburton-main
	 :debug-build nil
	 :init-file-pathname init-path
	 :extra-input-files (if (probe-file closopt-fasl) (list :trace :constructor :foreign closopt-fasl) '())
	 :lisp-heap-start "1000m"
	 ;; :lisp-heap-size "4000m"
	 :aclmalloc-heap-start "4000m"
	 ;; :aclmalloc-heap-size "2000m"
	 )
	))

(defun deliver-source ()
  ;; PARAMETERS YOU SHOULD SET!
  (let ((;; The path where all of the source files should be copied to.
		 src-path (mars-pathname "tburton/src/"))
		(;;Path to your common lisp init file -
		 ;; the file that is loaded and executed 
		 ;; automatically by your running lisp instance.
		 init-path #+linux "~/clinit.cl"
				   #+mswindows "C:/Users/David Wang/clinit.cl"))

	;; Build the executable
	(make-executable:make-source-folder src-path 'tburton :include-make-executable t :verbose t)
	;; copy the deliver file
	(make-executable:copy-file-to-source-folder (mars-pathname "tburton/deliver.lisp") src-path)
	;; copy the clinit.cl file
	(make-executable:copy-file-to-source-folder init-path src-path :keep-pathname-function #'make-executable:default-keep-pathname)
	;; copy the load-MARS-systems.lisp file
	(make-executable:copy-file-to-source-folder (mars-pathname "load-MARS-systems.lisp") src-path)
	))

#+allegro
(defun train ()
  (let* ((closopt-lisp (mars-pathname "tburton/closopt.cl"))
		 ;; (closopt-fasl (merge-pathnames (make-pathname :type "fasl") closopt-lisp))
		 (packages (get-package-names 'tburton)))
	(with-open-file 
		(st closopt-lisp :direction :output :if-exists :supersede)
	  (format st "(excl::preload-forms)~%")
	  (format st "(excl::preload-constructors ~s)~%" (append '(list :CL-USER :LISP) packages))
	  (format st "(excl::precache-generic-functions ~s)" (append '(list :CL-USER :LISP) packages)))
	;; TODO: Put a test here that uses tBurton on a hard search problem.
	;; (test-feature-extractor-satellite-with-sas-to-file)
	(compile-file closopt-lisp)
	))


;; This is convenience function, intended to be run manually to get a 
;; quick a dirty listing of all the packages your system depends on.
(defun get-package-names (asdf-system-name)
  "Try to get the names of packages, assuming they are named identically to their asdf systems."
  (let* ((poss-pkg-names (mapcar (lambda (system) (intern (string-upcase (asdf::component-name system)) :keyword)) 
								 (make-executable:get-all-systems asdf-system-name))))
	(mapcan #'(lambda (package-name) 
				(if (find-package package-name)
					(progn
					  (list package-name))
					(progn
					  (format t "~a is NOT a package.~%" package-name)
					  '())
					)) 
			poss-pkg-names)))
