

;; make sure the SNARK package is loaded
(in-package #:cl-user)
(eval-when (:compile-toplevel :load-toplevel :execute)
  (unless (find-package 'snark)
	(load (mars-pathname "3rdparty/snark-20120808r022/snark-system.lisp"))
	(make-snark-system t)
	;; (in-package #:tburton)
	;; (use-package :common-lisp)
	;; (use-package :snark-lisp)
	;; (use-package :snark-deque)
	;; (use-package :snark-sparse-array)
	;; (use-package :snark-dpll)
	;; (use-package :snark)
	))


(in-package #:tburton)


(defun pddl-to-snark (domain problem)
  "Returns a SNARK representation of the domain and problem."
  (let ((snark-rep (make-smart-list :initial-contents '(progn))))
	(flet ((add! (&rest instrs) (coll-add-all! snark-rep instrs)))

	  ;; initialization
	  (add! '(initialize :verbose nil))
	  (add! '(use-hyperresolution))

	  ;; delare types
	  (with-slots (types) domain
		;; add the default supertype type
		(add! `(declare-sort 'object :subsorts-incompatible t))
		;; add all the types in the type map as subsorts.
		(docoll ((subtypes (pddl::default-supertype types)) type-obj)
				(let ((type-name (type-name type-obj)))
				  ;; (domap (types type-name type-obj)
				  ;; add the direct subsorts of each type
				  (docoll ((direct-supertypes type-obj) sup-type-obj)
						  (add! `(snark-user::declare-subsort 
								  ',type-name 
								  ',(type-name sup-type-obj) :subsorts-incompatible t))
						  )))
			   ;; )
		)
	  
	  ;; declare constants & objects
	  (do-domain-constants (domain constant)
		(add-pddl-object-to-snark! snark-rep constant))
	  (do-problem-objects (problem object)
		(add-pddl-object-to-snark! snark-rep object))

	  ;; declare predicates
	  (do-domain-predicates (domain predicate)
		(let ((pred-name (predicate-name predicate))
			  (pred-args (rest predicate))
			  (num-pred-args (length (rest predicate)))
			  (sorts (make-smart-list)))
		  ;; loop over predicate arguments and 
		  ;; build up types list
		  (dolist (arg pred-args)
			(let ((types (rest arg))
				  (num-types (length (rest arg))) )
			  (cond (;; if there are no types, declare as type 'object
					 (= num-types 0)
					 (coll-add! sorts 'object) )
					(;; if there is one type, declare as that type
					 (= num-types 1)
					 (coll-add! sorts (first types)) )
					(;; if there is more than one type, then create a new sort that 
					 ;; combines existing sorts.
					 :otherwise
					 ;; TODO: we currently do not support the use of either
					 ;; we just add the first type of an either expression
					 (coll-add! sorts (first types)))
					 ;; (dolist (type (rest arg))))
					)
			  )
			;; add the relation with sorts
			(add!  `(declare-relation 
					 ',pred-name ,num-pred-args 
					 :sort ',(coll-tolist sorts))) )
		  ))

	  ;;test add action to snark
	  ;; (let ((first t))
	  ;; 	(do-domain-actions (domain action)
	  ;; 	  (when first
	  ;; 		(setf first nil)
	  ;; 		(coll-add-all! snark-rep (translate-pddl-variables-to-snark (action-parameters action))))))
	 
	  snark-rep)))

(defun add-pddl-object-to-snark! (snark-rep object)
  (flet ((add! (&rest instrs) (coll-add-all! snark-rep instrs)))
	(let* ((obj-name (first object))
		   (obj-types (rest object))
		   (num-types (length obj-types)))
	  (cond (;; if there are no types, declare as type 'object
			 (= num-types 0)
			 (add! `(declare-constant
					 ',obj-name
					 :sort 'object)) )
			(;; if there is one type, declare as that type
			 (= num-types 1)
			 (add! `(declare-constant
					 ',obj-name
					 :sort ',(first obj-types))) )
			(;; if there is more than one type, declare a temporary type
			 :otherwise
			 ;; TODO: we currently don't support more than one type for an object 
			 )
			)
	  )))


(defun translate-pddl-typed-list-to-snark-vars (typed-var-list &optional (snark-rep (make-smart-list)))
  (dolist (typed-var typed-var-list)
	(if (second typed-var)
		(coll-add! snark-rep `(declare-variable 
							   ,(first typed-var)
								;; :sort
								,(second typed-var))) 
		(coll-add! snark-rep `(declare-variable
							   ,(first typed-var)
								;; :sort
								object)) 
		)
	)
  snark-rep)

(defun translate-pddl-typed-list-to-snark-types (typed-var-list &optional (snark-rep (make-smart-list)))
  (dolist (typed-var typed-var-list)
	(if (second typed-var)
		(coll-add! snark-rep `(,(first typed-var)
								;; :sort
								,(second typed-var))) 
		(coll-add! snark-rep `(,(first typed-var)
								;; :sort
								object)) 
		)
	)
  snark-rep)

(defun translate-pddl-condition-to-snark (condition &optional (snark-rep (make-smart-list)))
  (flet ((add! (&rest instrs) (coll-add-all! snark-rep instrs)))
	(case (first condition)
	  ('pddl::and
	   (add! 'and)
	   (dolist (gd (rest condition))
		 (add! (translate-pddl-condition-to-snark gd))) )
	  ('pddl::or
	   (add! 'or)
	   (dolist (gd (rest condition))
		 (add! (translate-pddl-condition-to-snark gd))) )
	  ('pddl::not
	   (add! 'not)
	   (add! (translate-pddl-condition-to-snark (second condition))) )
	  ('pddl::imply
	   (add! 'imply)
	   (add! (translate-pddl-condition-to-snark (second condition)))
	   (add! (translate-pddl-condition-to-snark (third condition))) )
	  ('pddl::exists
	   (add! 'exists)
	   ;; add variables
	   (add! (loop for typed-var in (second condition)
				collect `(,(first typed-var) 
						   ;; :sort
						   ,(second typed-var)) into snark-vars
				  finally (return snark-vars) ))
	   ;; add conditions
	   (add! (translate-pddl-condition-to-snark (third condition))) )
	  ('pddl:forall
   	   (add! 'forall)
	   ;; add variables
	   (add! (loop for typed-var in (second condition)
				collect `(,(first typed-var) 
						   ;; :sort
						   ,(second typed-var)) into snark-vars
				finally (return snark-vars) ))
	   ;; add conditions
	   (add! (translate-pddl-condition-to-snark (third condition))) )
	  (;; deal with predicate
	   t 
	   (loop for elem in condition do
			(add! elem)) )
	  ))
  (coll-tolist snark-rep))


(defun enable-snark-output (verbose)
  (snark-user::print-rows-when-given nil)
  (snark-user::print-rows-when-derived verbose)
  (snark-user::print-rows-when-processed nil)
  (snark-user::print-final-rows verbose)
  (snark-user::print-unorientable-rows verbose)
  (snark-user::print-pure-rows nil)
  (snark-user::print-irrelevant-rows nil)
  (snark-user::print-rewrite-orientation nil)

  (snark-user::print-summary-when-finished verbose)
  (snark-user::print-clocks-when-finished verbose)
  (snark-user::print-term-memory-when-finished verbose)
  (snark-user::print-agenda-when-finished verbose)
  (snark-user::print-rows-when-finished verbose)
  (snark-user::print-options-when-starting verbose)
  (snark-user::print-assertion-analysis-notes verbose)
  (snark-user::print-symbol-table-warnings verbose))


(defun eval-snark (snark-problem &key (verbose nil))
  ;; set snark verbosity
  (enable-snark-output verbose)
  (let ((curr-pkg *package*)
		(result nil))
	;; convert adt-collections to a standard lisp list
	(when (typep snark-problem 'adt-collection)
	  (setf snark-problem (coll-tolist snark-problem :copy nil)))
	;; ensure all symbols in the list are in the snark-user package.
	(setf snark-problem (snarkify snark-problem))
	;; call snark on the expression
	(unwind-protect
		 (progn (in-package #:snark-user)
				(when verbose
				  (format t "Solving SNARK problem: ~s~%" snark-problem))
				(setf result (eval snark-problem)))
	  (setf *package* curr-pkg)
	  result)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HELPER
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun snark-symbol (sym)
  (intern sym 'snark-user))

(defun snarkify (list)
  "Given a recursively nested list of symbols, 
   returns a structurally identical list with 
   symbols internalized to te snark-user package."
  (cond ((null list)
		 nil)
		((keywordp list)
		 list)
		((symbolp list)
		 (snark-symbol list))
		((listp list)
		 (mapcar #'(lambda (elem) (snarkify elem)) list))
		(:otherwise
		 list)
		))


(in-package #:snark-user)

(defun steamroller-example2 ()

  (initialize)
  (use-hyperresolution)
  
  (declare-sort 'object)
  (declare-subsort 'animal 'object)
  (declare-subsort 'plant 'object)
  ;; (declare-sort 'animal :subsorts-incompatible t)
  ;; (declare-sort 'plant :subsorts-incompatible t)
  (declare-subsort 'bird 'animal)
  (declare-subsort 'caterpillar 'animal)
  (declare-subsort 'fox 'animal)
  (declare-subsort 'snail 'animal)
  (declare-subsort 'wolf 'animal)
  (declare-subsort 'grain 'plant)
  
  (declare-relation 'e 2 :sort '((1 animal)))	;animal*true
  (declare-relation 'm 2 :sort '((t animal)))	;animal*animal
  
  (declare-variable '?a1 :sort 'animal)
  (declare-variable '?a2 :sort 'animal)

  (assertion (forall ((?s1 snail) (?b1 bird))		;KIF-style sort specification
                     (m ?s1 ?b1))			;all KIF variables begin with ?
             :name snails-are-smaller-than-birds)
  (assertion (forall ((b1 :sort bird) (f1 :sort fox))	;SNARK-preferred sort specification
                     (m b1 f1))
             :name birds-are-smaller-than-foxes)
  (assertion (forall ((f1 true :sort fox)
                      (w1 wolf :sort wolf))		;this works too
                     (m f1 w1))
             :name foxes-are-smaller-than-wolves)
  (assertion (forall ((w1 wolf) (f1 fox))
                     (not (e w1 f1)))
             :name wolves-dont-eat-foxes)
  (assertion (forall ((w1 :sort wolf) (g1 :sort grain))
                     (not (e w1 g1)))
             :name wolves-dont-eat-grain)
  (assertion (forall ((b1 :sort bird) (c1 :sort caterpillar))
                     (e b1 c1))
             :name birds-eat-caterpillars)
  (assertion (forall ((b1 :sort bird) (s1 :sort snail))
                     (not (e b1 s1)))
             :name birds-dont-eat-snails)
  (assertion (forall ((c1 :sort caterpillar))
                     (exists ((p1 :sort plant))
                             (e c1 p1)))
             :name caterpillars-eat-some-plants)
  (assertion (forall ((s1 :sort snail))
                     (exists ((p1 :sort plant))
                             (e s1 p1)))
             :name snails-eat-some-plants)
  (assertion (forall ((p1 :sort plant) (p2 :sort plant))
                     (implied-by (or (e ?a1 ?a2) (e ?a1 p1)) (and (m ?a2 ?a1) (e ?a2 p2)))))
  
  (prove '(and (e ?x.animal ?y.animal) (e ?y.animal ?z.grain))
		 :answer '(values ?x.animal ?y.animal ?z.grain))

)