;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Seung H. Chung, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities

(in-package #:tburton)

(defstruct cca
  state-variables
  causal-graph
  planning-function)

(defmethod plan-test-bus ()
  (let* ((qsp (make-qsp-bus))
	 (causal-graph (make-causal-graph-bus))
	 (policy-function-ht-component (make-hash-table :test #'equal))
	 (state-variables '(mtk-qsp::bus))
	 (s0 '((= mtk-qsp::bus mtk-qsp::off)
	       (= mtk-qsp::bus-cmd mtk-qsp::no-cmd))))
    (setf (gethash '(mtk-qsp::bus) policy-function-ht-component)
	  (make-policy-function #'policy-description-bus))
    (let* ((planning-function (make-planning-function policy-function-ht-component))
	   (cca (make-cca :state-variables state-variables
			  :causal-graph causal-graph
			  :planning-function planning-function)))
      (tburton:plan qsp cca s0))))

(defmethod plan-test-bus-driver ()
  (let* ((qsp (make-qsp-bus-driver))
	 (causal-graph (make-causal-graph-bus-driver))
	 (policy-function-ht-component (make-hash-table :test #'equal))
	 (state-variables '(mtk-qsp::bus mtk-qsp::driver))
	 (s0 '((= mtk-qsp::bus mtk-qsp::off)
	       (= mtk-qsp::bus-cmd mtk-qsp::no-cmd)
	       (= mtk-qsp::driver mtk-qsp::off)
	       (= mtk-qsp::driver-cmd mtk-qsp::no-cmd)
	       )))
    (setf (gethash '(mtk-qsp::bus) policy-function-ht-component)
	  (make-policy-function #'policy-description-bus))
    (setf (gethash '(mtk-qsp::driver) policy-function-ht-component)
	  (make-policy-function #'policy-description-driver-compiled))
    (let* ((planning-function (make-planning-function policy-function-ht-component))
	   (cca (make-cca :state-variables state-variables
			  :causal-graph causal-graph
			  :planning-function planning-function)))
      (tburton:plan qsp cca s0))))

(defmethod plan-test-bus-driver-valve ()
  (let* ((qsp (make-qsp-bus-driver-valve))
	 (causal-graph (make-causal-graph-bus-driver-valve))
	 (policy-function-ht-component (make-hash-table :test #'equal))
	 (state-variables '(mtk-qsp::bus mtk-qsp::driver mtk-qsp::valve))
	 (s0 '((= mtk-qsp::bus mtk-qsp::off)
	       (= mtk-qsp::bus-cmd mtk-qsp::no-cmd)
	       (= mtk-qsp::driver mtk-qsp::off)
	       (= mtk-qsp::driver-cmd mtk-qsp::no-cmd)
	       (= mtk-qsp::valve mtk-qsp::closed)
	       (= mtk-qsp::valve-cmd mtk-qsp::no-cmd)
	       )))
    (setf (gethash '(mtk-qsp::bus) policy-function-ht-component)
	  (make-policy-function #'policy-description-bus))
    (setf (gethash '(mtk-qsp::driver) policy-function-ht-component)
	  (make-policy-function #'policy-description-driver-compiled))
    (setf (gethash '(mtk-qsp::valve) policy-function-ht-component)
	  (make-policy-function #'policy-description-valve-compiled))
    (let* ((planning-function (make-planning-function policy-function-ht-component))
	   (cca (make-cca :state-variables state-variables
			  :causal-graph causal-graph
			  :planning-function planning-function)))
      (tburton:plan qsp cca s0))))

(defmethod plan-test-bus-driver-valve-expanded ()
  "This method is identical to the one above except some function calls have been inlined, 
   so you can better see how all the pieces come together."
  (let* ((qsp (let* ((qsp (make-instance 'qsp)))
				(mtk-qsp::add-constraints `((:scope (0 1) :relation ((:lower-bound (0) :upper-bound (10))))
								   (:scope (1 2) :relation ((:lower-bound (100) :upper-bound (101)))
										   :state-constraint (:scope (valve) :relation ((= valve opened))))
								   (:scope (2 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (3 4) :relation ((:lower-bound (100) :upper-bound (101)))
										   :state-constraint (:scope (valve) :relation ((= valve closed))))
								   (:scope (4 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (0 5) :relation ((:lower-bound (100) :upper-bound (250))))
								   (:scope (1 6) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (6 7) :relation ((:lower-bound (0) :upper-bound (,*inf*)))
										   :state-constraint (:scope (driver) :relation ((= driver off))))
								   (:scope (7 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (6 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (3 8) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (8 9) :relation ((:lower-bound (0 :inequality ,#'>) 
																		  :upper-bound (,*inf*)))
										   :state-constraint (:scope (driver) :relation ((= driver off))))
								   (:scope (9 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (1 10) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (10 3) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (10 11) :relation ((:lower-bound (0) :upper-bound (,*inf*)))
										   :state-constraint (:scope (bus) :relation ((= bus off))))
								   (:scope (11 12) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (3 12)  :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   (:scope (12 13) :relation ((:lower-bound (0 :inequality ,#'>) 
																			:upper-bound (,*inf*)))
										   :state-constraint (:scope (bus) :relation ((= bus off))))
								   (:scope (13 5) :relation ((:lower-bound (0) :upper-bound (,*inf*))))
								   ) qsp)
				qsp))
		 (causal-graph   (let ((causal-graph (make-instance 'mtk-graph:directed-graph)))
						   (mtk-graph:add-edge '(bus-cmd bus) causal-graph)
						   (mtk-graph:add-edge '(driver-cmd driver) causal-graph)
						   (mtk-graph:add-edge '(valve-cmd valve) causal-graph)
						   (mtk-graph:add-edge '(bus driver) causal-graph)
						   (mtk-graph:add-edge '(bus valve) causal-graph)
						   (mtk-graph:add-edge '(driver valve) causal-graph)
						   causal-graph))
		 (state-variables '(bus driver valve))
		 (s0 '((= bus off)
			   (= bus-cmd no-cmd)
			   (= driver off)
			   (= driver-cmd no-cmd)
			   (= valve closed)
			   (= valve-cmd no-cmd)
			   )))
    (let ((policy-function-ht-component (make-hash-table :test #'equal)))
	  (setf (gethash '(bus) policy-function-ht-component)
			(make-policy-function #'policy-description-bus))
	  (setf (gethash '(driver) policy-function-ht-component)
			(make-policy-function #'policy-description-driver-compiled))
	  (setf (gethash '(valve) policy-function-ht-component)
			(make-policy-function #'policy-description-valve-compiled))
	  (let* ((planning-function (make-planning-function policy-function-ht-component))
			(cca (make-cca :state-variables state-variables
						   :causal-graph causal-graph
						   :planning-function planning-function)))
		(tburton:plan qsp cca s0)))))


;; A planning function takes an argument of (from to component). Given the
;; "from" state and the "to" state of the component "component", the function
;; returns a finite trajectory of the form (a0 s1 a1 s2 ...), where a and s
;; denote action and state, respectively.
(defun make-planning-function (policy-function-ht-component)
  #'(lambda (from to component)
      (plan-using-policy-function
       from to (gethash component policy-function-ht-component))))

(defun make-causal-graph-bus ()
  (let ((causal-graph (make-instance 'mtk-graph:directed-graph)))
    (mtk-graph:add-edge '(mtk-qsp::bus-cmd mtk-qsp::bus) causal-graph)
    causal-graph))

(defun make-causal-graph-bus-driver ()
  (let ((causal-graph (make-causal-graph-bus)))
    (mtk-graph:add-edge '(mtk-qsp::driver-cmd mtk-qsp::driver) causal-graph)
    (mtk-graph:add-edge '(mtk-qsp::bus mtk-qsp::driver) causal-graph)
;;    (mtk-graph:add-edge '(mtk-qsp::bus mtk-qsp::driver-cmd) causal-graph)
    causal-graph))

(defun make-causal-graph-bus-driver-valve ()
  (let ((causal-graph (make-causal-graph-bus-driver)))
    (mtk-graph:add-edge '(mtk-qsp::valve-cmd mtk-qsp::valve) causal-graph)
    (mtk-graph:add-edge '(mtk-qsp::driver mtk-qsp::valve) causal-graph)
;;    (mtk-graph:add-edge '(mtk-qsp::driver mtk-qsp::valve-cmd) causal-graph)
    (mtk-graph:add-edge '(mtk-qsp::bus mtk-qsp::valve) causal-graph)
    causal-graph))

(defun make-policy-function (policy-description)
  (let ((policy-ht (make-hash-table :test #'equal)))
    (dolist (plan-description (funcall policy-description))
;;       (format t "~S ~S ~S~%"
;; 	      (plan-description-from plan-description)
;; 	      (plan-description-to plan-description)
;; 	      (plan-description-paths plan-description))
      (setf (gethash (list (plan-description-from plan-description)
						   (plan-description-to plan-description))
					 policy-ht)
			`(,(plan-description-paths plan-description)
			  ,(plan-description-durations plan-description))))
    #'(lambda (from to) (gethash (list from to) policy-ht))))


;; Ideally, this function should allow us to choose from many feasible plans from
;; the "from" state to the "to" state. For now, we just commit to one.
(defun plan-using-policy-function (from to policy-function)
  ;; If the goal "to" is the same as "from", then we can either choose to do nothing
  ;; or find a cyclic trajectory. For now, we choose to simply do nothing.
  ;; Technically, we cannot just check if "to" is the same as "from", we must check
  ;; if "to" is satisfied by "from".

  ;;(format t "Planning from: ~S to: ~S~%" from to)
  (if (equal from to)
      (progn ;;(format t "Using current state for the plan.~%")
	     t)
      ;; To be more precise, we should look for a path the leads to a state that is
      ;; consistent with "to".
      (let ((paths-and-durations (funcall policy-function from to)))
		(if (or (null paths-and-durations) (null (first paths-and-durations)))
			;; If no paths exists, then no plan exists. Planning failed.
			nil
			;; Otherwise, choose a plan
			(let (;; for now, just choose the first path 
				  (chosen-path (first (first paths-and-durations))))
			  (cond
				;; If the paths is nil, it means no subgoal is necessary.
				((null chosen-path) (list (list nil to)))
				;; If the first element in the path is a state constraint, the path
				;; is simply a transition from the "from" state to the "to" state
				;; with the subgoals represented by the "path".
				((typep (first chosen-path) 'mtk-csp:constraint)
				 (list (list chosen-path to)))
				;; Otherwise, the chose path is the next state in the plan. We must
				;; find a plan that achieves that state.
				(t (append (plan-using-policy-function from chosen-path policy-function)
						   (plan-using-policy-function chosen-path to policy-function)))))))))
  
;; The paths contains a list of paths that can be taken to achieve the goals.
;; There are three types of paths:
;;   * List of state constraints : These are the subgoals that must be achieved.
;;                                 If there are no state constraints, nothing
;;                                 needs to be done. This typically implies that
;;                                 the transition is uncontrollable.
;;   * State : This is the next state that will eventually lead toward the goal.
(defstruct plan-description
  from					; from state
  to					; to state (i.e. goal state)
  paths		    	 	; List of ways to achieve the goal
  durations)             ; List of durations for the ways to achieve the goal

(defmethod policy-description-bus ()
  (list 
   (make-plan-description
	:from '(= mtk-qsp::bus mtk-qsp::off)
	:to '(= mtk-qsp::bus mtk-qsp::init) 
	:paths (list (list (mtk-qsp:make-state-constraint
						:scope '(mtk-qsp::bus-cmd)
						:relation '((= mtk-qsp::bus-cmd mtk-qsp::turn-on)))))
	:durations (list `((:lower-bound (0) :upper-bound (10)))))
	(make-plan-description
	 :from '(= mtk-qsp::bus mtk-qsp::off)
	 :to '(= mtk-qsp::bus mtk-qsp::on)
	 :paths '((= mtk-qsp::bus mtk-qsp::init)))
	(make-plan-description
	 :from '(= mtk-qsp::bus mtk-qsp::init)
	 :to '(= mtk-qsp::bus mtk-qsp::off)
	 :paths (list (list (mtk-qsp:make-state-constraint 
						 :scope '(mtk-qsp::bus-cmd)
						 :relation '((= mtk-qsp::bus-cmd mtk-qsp::turn-off)))))
	 :durations (list `((:lower-bound (0) :upper-bound (10)))))
	(make-plan-description
	 :from '(= mtk-qsp::bus mtk-qsp::init)
	 :to '(= mtk-qsp::bus mtk-qsp::on)
	 :paths '(())
	 :durations (list `((:lower-bound (0) :upper-bound (,*inf*)))))
	(make-plan-description
	 :from '(= mtk-qsp::bus mtk-qsp::on)
	 :to '(= mtk-qsp::bus mtk-qsp::off)
	 :paths (list (list (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus-cmd)
			     :relation '((= mtk-qsp::bus-cmd mtk-qsp::turn-off)))))
	 :durations (list `((:lower-bound (0) :upper-bound (,*inf*)))))))

(defmethod policy-description-driver-compiled ()
  (list (make-plan-description
	 :from '(= mtk-qsp::driver mtk-qsp::off)
	 :to '(= mtk-qsp::driver mtk-qsp::init) 
	 :paths (list (list (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::driver-cmd)
			     :relation '((= mtk-qsp::driver-cmd mtk-qsp::turn-on)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus)
			     :relation '((= mtk-qsp::bus mtk-qsp::on))))))
	(make-plan-description
	 :from '(= mtk-qsp::driver mtk-qsp::off)
	 :to '(= mtk-qsp::driver mtk-qsp::on)
	 :paths '((= mtk-qsp::driver mtk-qsp::init)))
	(make-plan-description
	 :from '(= mtk-qsp::driver mtk-qsp::init)
	 :to '(= mtk-qsp::driver mtk-qsp::off)
	 :paths (list (list (mtk-qsp:make-state-constraint 
			     :scope '(mtk-qsp::driver-cmd)
			     :relation '((= mtk-qsp::driver-cmd mtk-qsp::turn-off)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus)
			     :relation '((= mtk-qsp::bus mtk-qsp::on))))))
	(make-plan-description
	 :from '(= mtk-qsp::driver mtk-qsp::init)
	 :to '(= mtk-qsp::driver mtk-qsp::on)
	 :paths '(()))
	(make-plan-description
	 :from '(= mtk-qsp::driver mtk-qsp::on)
	 :to '(= mtk-qsp::driver mtk-qsp::off)
	 :paths (list (list (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::driver-cmd)
			     :relation '((= mtk-qsp::driver-cmd mtk-qsp::turn-off)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus)
			     :relation '((= mtk-qsp::bus mtk-qsp::on))))))))

(defmethod policy-description-valve-compiled ()
  (list (make-plan-description
	 :from '(= mtk-qsp::valve mtk-qsp::closed)
	 :to '(= mtk-qsp::valve mtk-qsp::opened) 
	 :paths (list (list (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::valve-cmd)
			     :relation '((= mtk-qsp::valve-cmd mtk-qsp::open)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::driver)
			     :relation '((= mtk-qsp::driver mtk-qsp::on)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus)
			     :relation '((= mtk-qsp::bus mtk-qsp::on))))))
	(make-plan-description
	 :from '(= mtk-qsp::valve mtk-qsp::opened)
	 :to '(= mtk-qsp::valve mtk-qsp::closed)
	 :paths (list (list (mtk-qsp:make-state-constraint 
			     :scope '(mtk-qsp::valve-cmd)
			     :relation '((= mtk-qsp::valve-cmd mtk-qsp::close)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::driver)
			     :relation '((= mtk-qsp::driver mtk-qsp::on)))
			    (mtk-qsp:make-state-constraint
			     :scope '(mtk-qsp::bus)
			     :relation '((= mtk-qsp::bus mtk-qsp::on))))))))

(defmethod convert-qcp-to-tpn-string (qcp)
  (with-output-to-string (stream) 
	(let* ((edges (mtk-graph:get-edge-list (mtk-graph:get-edges qcp))))

	  ;; Output the beginning of the TPN
	  (format stream "<TPN>~%")
	  (format stream "  <FORMAT> SPEC2 </FORMAT>~%")
	  (format stream "  <NAME> Main.run </NAME>~%")
	  ;; Not sure how to get the start and end?? Is it even used by the dispatcher?
	  (format stream "  <START> 0 </START>~%")
	  (format stream "  <END> 0 </END>~%")

	  (loop for edge in edges do
		   (let* ((state-constraint (get-state-constraint edge qcp))
				  (temporal-constraint (mtk-csp-tcsp:get-temporal-constraint edge qcp))
				  (temporal-relation  (first (mtk-csp:get-relation temporal-constraint)))
				  (lb (mtk-utils-bounded-set:get-value (mtk-utils-bounded-set:get-lower-bound temporal-relation)))
				  (ub (mtk-utils-bounded-set:get-value (mtk-utils-bounded-set:get-upper-bound temporal-relation)))
				  (vertices (mtk-graph:get-vertices edge))
				  (u (first vertices))
				  (v (second vertices)))

			 (format stream "  <ARC>~%")
			 (format stream "    <START> ~a </START>~%" (rename-vertex u))
			 (format stream "    <END> ~a </END>~%" (rename-vertex v))
			 (format stream "    <LOWERBOUND> ~a </LOWERBOUND>~%"  (tpn-time-val lb))
			 (format stream "    <UPPERBOUND> ~a </UPPERBOUND>~%" (tpn-time-val ub))

			 (when state-constraint
				 (let* ((state-relation (first (mtk-csp:get-relation state-constraint)))) 
				   ;(format t "~a -> ~a [~a, ~a] ~a~%" u v lb ub state-relation)
				   (format stream "    <PRIMITIVE> ~a </PRIMITIVE>~%" state-relation)))

			 (format stream "  </ARC>~%")
		
			 )
		   )

	  ;; Close the TPN
	  (format stream "</TPN>~%")
	  ))
  )

(defmethod rename-vertex (v)
  (let* ((v-str (format nil "~a" v))
		 (v-str (remove-if #'(lambda (x) (or (eql x #\space) (eql x #\() (eql x #\)) (eql x #\-))) v-str )))
	v-str))

(defmethod tpn-time-val (time)
  (cond 
	((= time *inf*) "+INF")
	((= time *-inf*) "-INF")
	(t (format nil "~a" time))))