


(in-package #:tburton)


(defclass stn-episode (directed-edge)
  ((lb :type number
	   :initarg :lb
	   :accessor get-lb
	   :initform 0
	   :documentation "lowerbound")
   (ub :type number
	   :initarg :ub 
	   :accessor get-ub
	   :initform *inf*
	   :documentation "upperbound")
   ))

(defun make-stn-episode (source target &key (lb 0) (ub *inf*) &allow-other-keys)
  (make-instance 'stn-episode :lb lb :ub ub
				 :vertices (list source target)
				 :source-vertices (list source)
				 :target-vertices (list target)))

(defclass stn (directed-graph) ())

(defun make-stn ()
	(make-instance 'stn))

(defmethod add-episode ((episode stn-episode)(stn stn))
  (add-edge episode stn))

(defmethod is-temporally-consistent ((g stn))
  (let* ((event-ht (make-hash-table))
		 (events (get-vertices g))
		 (episodes (get-edge-list (get-edges g)))
		 (num-events 0)
		 (num-episodes (length episodes)))
	;; Create a hash table that maps events to an index number.
	(loop for event in events do
		 (setf (gethash event event-ht) num-events)
		 (incf num-events))
	(let* (;; Each episode has a lower and upperbound constraint
		   (m (* 2 num-episodes))
		   ;; Each event has an occurence time
		   ;; plus two slack variable, one each for the lower and upper bounds.
		   (n (+ num-events (* 2 num-episodes)))
		   (A (make-array `(,m ,n) :element-type 'number :initial-element 0))
		   (b (make-array m :element-type 'number :initial-element 0))
		   (c (make-array n :element-type 'number :initial-element 0)))
	  ;; Create the A and b matrix
	  (loop for episode in episodes
		 with mi = 0 do
		   (let ((starti (gethash (get-source-vertex episode) event-ht))
				 (endi (gethash (get-target-vertex episode) event-ht))
				 (lb (get-lb episode))
				 (ub (get-ub episode)))
			 ;; b - a + s1 = ub, where s1>=0
			 (setf (aref A mi endi) 1)
			 (setf (aref A mi starti) -1)
			 (setf (aref A mi (+ num-events mi)) 1)
			 (setf (aref b mi) ub)
			 (incf mi)
			 ;; b - a - s2 = lb, where s2>=0
			 (setf (aref A mi endi) 1)
			 (setf (aref A mi starti) -1)
			 (setf (aref A mi (+ num-events mi)) -1)
			 (setf (aref b mi) lb)
			 (incf mi)))
	  ;; Create the c matrix
	  (fill c 1 :start 0 :end num-events)
	  (loop for k being the hash-keys in event-ht using (hash-value v) do
	  	   (format t "~a:~a ~%" k v) )
	  (format t "A ~a~%" A)
	  (format t "b ~a~%" b)
	  (format t "c ~a~%" c)
	  (let ((consistent (simplex:simplex A b c)))
		consistent)
	  )))

(defclass tpn (stn) 
  ((choice :type hash-table
		   :accessor get-choice-ht
		   :initform (make-hash-table)
		   :documentation "A hashtable that maps symbol vertex names to a boolean as to whether it is a choice.")))