(in-package #:tBurton)
(defvar *tBurtonCoAddress* "localhost")
(defvar *tBurtonReaderPort* 15007)
(defvar *tBurtonWriterPort* 15008)

(defun start-tBurton-socket-server (&key (CoAddress "localhost") (ReaderPort 15007) (WriterPort 15008))

  ;; Start a socket server for receiving requests from the coordinator
 
  (setf *tBurtonCoAddress* CoAddress)
  (setf *tBurtonReaderPort* ReaderPort)
  (setf *tBurtonWriterPort* WriterPort)
 
  (let ((socketServer (ignore-errors (socket:make-socket :connect :passive :local-port *tBurtonWriterPort*)))
		(hasDomain nil)
		(hasProblem nil)
		#+(or win32 mswindows) (DomainFile "E:\\tBurton.domain")
		#+(or win32 mswindows) (ProblemFile "E:\\tBurton.problem")
		#-(or win32 mswindows) (DomainFile "/tmp/tBurton.domain")
		#-(or win32 mswindows) (ProblemFile "/tmp/tBurton.problem")
		(PLAN)
		(count 0))

	(when (null socketServer)

			;; Cannot establish connection to the coordinator
			(format t "Cannot establish local tBurton server~%")
			(return-from start-tBurton-socket-server nil))

	(format t "Server established~%")

	(loop

		 ;; Set a max service limit
	   until (> count 500)
		 

	   do (when (null (ros-interface:send-socket-message "Ready" *tBurtonCoAddress* *tBurtonReaderPort*))

			;; Cannot establish connection to the coordinator
			(format t "Cannot connect to the Coordinator at ~A:~A~%"  *tBurtonCoAddress* *tBurtonWriterPort*)
			(close socketServer)
			(return-from start-tBurton-socket-server nil))



	   (let ((socketInput (ignore-errors (socket:accept-connection socketServer)))
				(input nil))

			(when (null socketInput)
			  
			  ;; Cannot establish a local port
			  (format t "Cannot establish a connection to local server~%")
			  (close socketServer)
			  (return-from start-tburton-socket-server nil))

			(format t "Connection established~%")

			
			(setf input (read-all-from-socket socketInput))

			(setf count (+ count 1))

			(when (not (null input))

			  ;; First, determine the type of the content by checking the type marker

			  (let ((type (pop input))
					(content nil))

				(cond ((string-equal "Domain" type)

					   ;; Getting an STN
					   ;; Save it to a file for future use
					   (setf content (pop input))
					   ;;(format t "~A~%" content)
					   (writeFile DomainFile content)
					   (setf hasDomain T)
					   (ros-interface:send-socket-message "Ready" *tBurtonCoAddress* *tBurtonReaderPort*))

					  ((string-equal "Problem" type)

					   ;; Getting an STN
					   ;; Save it to a file for future use
					   (setf content (pop input))
					   ;;(format t "~A~%" content)
					   (writeFile ProblemFile content)
					   (setf hasProblem T)
					   (ros-interface:send-socket-message "Ready" *tBurtonCoAddress* *tBurtonReaderPort*))

					  ;; ((string-equal "StatusRequest" type)

					  ;;  ;; return the current status of tBurton

					  ;;  (when (null Executing)
					  ;; 	 "Ready for executing plans"
					  ;; 	 (ros-interface:send-socket-message "Ready" *tBurtonCoAddress* *tBurtonReaderPort*)))

					  ((string-equal "Stop" type)

					   ;; Stop tBurton and go offline
					   (ros-interface:send-socket-message "Offline" *tBurtonCoAddress* *tBurtonReaderPort*)
					   (close socketInput)
					   (close socketServer)
					   (return-from start-tBurton-socket-server t)
					   ))

				(when (and hasDomain hasProblem)

				  ;; tBurton can start planning now
				  ;; call tBurton with the files received
				  ;; and signaling busy

				  (ros-interface:send-socket-message "Busy" *tBurtonCoAddress* *tBurtonReaderPort*)

;;;;;;;;;;;;;;;;;;;;;;; Modify for tBurton
				  (setf PLAN (tburton-file DomainFile ProblemFile 'boeing-problem :use-zero-episodes t))

				  (ros-interface:send-socket-message "Ready" *tBurtonCoAddress* *tBurtonReaderPort*)


				  ;; Report results to the coordinator
				  (report-tBurton-result PLAN)
;;;;;;;;;;;;;;;;;;;;;;; End modification
				  ;; Stop tBurton
				  (setf hasDomain nil)
				  (setf hasProblem nil)
				  )


				))



			;;(format s "~s" (+ (read s) (read s)))
			(close socketInput)))

	(ros-interface:send-socket-message "Offline" *tBurtonCoAddress* *tBurtonReaderPort*)
	(close socketServer))

  )
 
(defun report-tBurton-result (tBurton-plan)

  ;; Report the results of planning to the coordinator
  ;; through socket messages
  (if (not (null tBurton-plan))
	  (let ((plan-string (with-output-to-string (plan-stream)
						   (write-tpn plan-stream tBurton-plan #|:environment `(,tpn::*always-active*) :force-labeling? t|#))  ))
		;; Now examine what happened.
		(format t "tBurton Planning Result: ~a~%" plan-string)
		(ros-interface:send-socket-message (format nil "ResultTPN;~a" plan-string) *tBurtonCoAddress* *tBurtonReaderPort*))
	  (ros-interface:send-socket-message (format nil "NoSolution") *tBurtonCoAddress* *tBurtonReaderPort*))
  )


(defun read-all-from-socket (socketInput)

  ;; given a socket, return a list of strings
  ;; the first one being the type of the message, if available
  ;; the second item is the content

  (let ((line (read-line socketInput nil nil))
		(position nil))

	(setf position (search ";" line))

	(if (null position)

		(progn
		  ;; This string does not have a mark
		  ;; It is not part of the planning request

		  ;; Maybe it is status request?
		  
		  ;;(format t "Receiving request ~A~%" line)
		  (return-from read-all-from-socket (list line)))

		(progn
		  ;; This string has a mark
		  ;; save the mark as the first item in the output

		  (let ((type (subseq line 0 position))
				(content (subseq line (+ position 1) (length line))))

			;;(format t "Type: ~A~%" type)
			;;(format t "First line: ~A~%" content)

			(setf line (read-line socketInput))

			(loop 
			   until (null line)

			   do (setf content (format nil "~a~%~a" content line))
			   ;;(format t "Reading: ~A~%" line)
				 (setf line (read-line socketInput nil nil)))
			
			(return-from read-all-from-socket (list type content)))))
	)
  )


(defun writeFile (name content)

  ;; write the content to the file specified by name

  ;; there is a bug with the overwrite function
  ;; file may not be completely overwritten
  ;; try deleting the file first, then write a new one

  (ignore-errors (delete-file name))

  (with-open-file (stream name
						  :direction :output
						  :if-exists :overwrite
						  :if-does-not-exist :create)
    (format stream content))

  )
