;;;; Copyright (c) 2012 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:tburton)

(defun test-tburton1 ()
  (test-tburton-file "tburton/tests/boeing-problem/boeing-domain.pddl"
					 "tburton/tests/boeing-problem/boeing-problem.pddl"
					 'boeing-problem))

(defun test-tburton2 ()
  (test-tburton-file "tburton/tests/pike-problem/pike.domain"
					 "tburton/tests/pike-problem/pike.problem"
					 'boeing-problem))

(defun test-tburton3 ()
  (test-tburton-file "tburton/tests/boeing-problem2/tBurton.domain"
					 "tburton/tests/boeing-problem2/tBurton.problem"
					 'boeing-problem
					 :dot-pdf-pathname "tburton/tests/boeing-problem2/tBurton.pdf"))

(defun test-tburton4 ()
  (test-tburton-file "tburton/tests/boeing-problem3/arch-domain.pddl"
					 "tburton/tests/boeing-problem3/arch-problem.pddl"
					 'archproblem
					 :dot-pdf-pathname "tburton/tests/boeing-problem3/arch-domain.pdf"))

(defun test-tburton5 ()
  (test-tburton-file "tburton/tests/boeing-problem3/arch-domain-temporal.pddl"
					 "tburton/tests/boeing-problem3/arch-problem-temporal.pddl"
					 'archproblem
					 :dot-pdf-pathname "tburton/tests/boeing-problem3/arch-domain-temporal.pdf"))

(defun test-tburton-vex-castle ()
  (test-tburton-file "tburton/tests/vex/vex-assembly-domain.pddl"
					 "tburton/tests/vex/vex-assembly-problem.pddl"
					 'vex-assembly-problem-castle
					 :dot-pdf-pathname "tburton/tests/vex/vex-assembly-problem.pdf"))

(defun test-tburton-vex-castle-2 ()
  (test-tburton-file "tburton/tests/vex/vex-assembly-domain-no-typing.pddl"
					 "tburton/tests/vex/vex-assembly-problem-no-typing.pddl"
					 'vex-assembly-problem-castle
					 :dot-pdf-pathname "tburton/tests/vex/vex-assembly-problem-no-typing.pdf"))

(defun test-tburton-rescue-1 ()
  (test-tburton-file "tburton/tests/search-and-rescue/rescue-domain-timed.pddl"
					 "tburton/tests/search-and-rescue/rescue-problem-timed.pddl"
					 'rescue-problem
                     :dot-pdf-pathname "tburton/tests/search-and-rescue/rescue-problem-timed.pdf"
					 :tpn-pathname "tburton/tests/search-and-rescue/rescue-problem-timed.tpn"))

(defun test-tburton-occluded-blocks-1 ()
  (test-tburton-file "tburton/tests/occluded-blocks/occluded-blocks-domain.pddl"
					 "tburton/tests/occluded-blocks/occluded-blocks-problem.pddl"
					 'occluded-blocks-problem
                     :dot-pdf-pathname "tburton/tests/occluded-blocks/occluded-blocks.pdf"
					 :tpn-pathname "tburton/tests/occluded-blocks/occluded-blocks.tpn"))


;; Helper function to call TBurton on some files (Steve added this!)
(defun test-tburton-file (pddl-domain-pathname pddl-problem-pathname probname &key (dot-pdf-pathname nil) (tpn-pathname nil))
  "This is a convenience function that calls tBurton on the given PDDL 
   problem and then prints the resulting TPN. Optional parameters will 
   also result in a graph of the TPN being output to a PDF, and the TPN 
   be written to a file."
  (let ((tpn-solution (tburton-file pddl-domain-pathname pddl-problem-pathname probname)))
	(cond (tpn-solution
		   (format t "~a" (with-output-to-string (tpn-solution-stream)
							(tpn::write-tpn tpn-solution-stream tpn-solution)))
		   (when dot-pdf-pathname
			 (tpn2dot tpn-solution (cl-user::mars-pathname dot-pdf-pathname)))
		   (when tpn-pathname
			 (with-open-file (stream tpn-pathname :direction :output :if-exists :supersede)
			   (tpn::write-tpn stream tpn-solution)))	   
		   t)
		  (:otherwise
		   (format t "No solution found!~%")
		   nil)
		  )))

