;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2013


(in-package :temporal-constraints)


(defclass temporal-duration (copyable-mixin)
  ((controllable?
	:initarg :controllable?
	:accessor temporal-duration-controllable?
	:initform T
	:documentation "T if this duration is controllable, nil
	otherwise."))
  (:documentation "The base class for all temporal
  durations. Controllable by default."))

(defclass bounded-temporal-duration (temporal-duration)
  ((upper-bound
	:initarg :upper-bound
	:accessor btd-upper-bound)
   (lower-bound
	:initarg :lower-bound
	:accessor btd-lower-bound))
  (:documentation "Describes temporal durations where the duration can
  be anywhere between UPPER-BOUND and LOWER-BOUND."))

(defclass looping-temporal-duration (temporal-duration)
  ((n-upper-bound
	:initarg :n-upper-bound
	:accessor ltd-n-upper-bound
	:documentation "Upper bound on the number of loops allowed.")

   (n-lower-bound
	:initarg :n-lower-bound
	:accessor ltd-n-lower-bound
	:documentation "Lower bound on the number of loops allowed.")

   (per-loop-duration
	:initarg :per-loop-duration
	:accessor ltd-per-loop-duration
	:documentation "TEMPORAL-DURATION describing the duration between
	the beginning and end events of a single loop."))
  
  (:documentation "Temporal durations that consist of a number of
  loops that have to be scheduled and each loop instance has its own
  TEMPORAL-DURATION."))

(defclass probabilistic-temporal-duration (temporal-duration)
  ((controllable?
	:initform nil))
  (:documentation "Superclass for probabilistic temporal
  durations. Uncontrollable by default."))
