;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:visibility-graph)

(mtk-utils:deftest-suite :test-visibility-graph)

(mtk-utils:deftest (test-vg-1 :test-visibility-graph) ()
;;  Creates an OBSTACLE-LIST and WAYPOINT-LIST,
;;  representing traversing a pair of square-obstacles that
;;  are origin symmetric.
  (let ((pt1 (make-2d-point -8  0))
		(pt2 (make-2d-point  8  0))
		(pt3 (make-2d-point  2  3))
		(pt4 (make-2d-point  6  3))
		(pt5 (make-2d-point  6 -1))
		(pt6 (make-2d-point  2 -1))
		(pt7 (make-2d-point -6  1))
		(pt8 (make-2d-point -2  1))
		(pt9 (make-2d-point -2 -3))
		(pt10 (make-2d-point -6 -3)))
	(let* ((polygon1 (make1-polygon (list pt3 pt4 pt5 pt6)))
		   (polygon2 (make1-polygon (list pt7 pt8 pt9 pt10)))
		   (obstacle-list `(,polygon1 ,polygon2))
		   (waypoint-list `(,pt1 ,pt2))
		   (vg (make-visibility-graph obstacle-list waypoint-list)))
	  ;; Test GET-COST
	  (mtk-utils:test (+ (sqrt 5) 4 (sqrt 5) (sqrt 5) 4 (sqrt 5))
					  (get-cost vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  ;; Test GET-COST-FROM-PATH
	  (mtk-utils:test (+ (sqrt 5) 4 4 4 4 4 (sqrt 5) (sqrt 5) 4 4 4 4 4 (sqrt 5))
	  				  (get-cost-from-path vg (list pt1
												   pt7 pt8 pt9 pt10 pt7
												   pt6 pt3 pt4 pt5 pt6
												   pt2))
	  				  :test #'equalp
	  				  :fail-info "")
	  ;; Test GET-PATH
	  (mtk-utils:test (list pt1 pt7 pt8 pt6 pt5 pt2)
					  (get-path vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  )))


(mtk-utils:deftest (test-vg-2 :test-visibility-graph) ()
;;  Creates an OBSTACLE-LIST and WAYPOINT-LIST,
;;  representing traversing a pair of square-obstacles that
;;  are origin symmetric, but the second waypoint is in
;;  an obstacle.
  (let ((pt1 (make-2d-point -8  0))
		(pt2 (make-2d-point  8  0))
		;; polygon 1
		(pt3 (make-2d-point  2  3))
		(pt4 (make-2d-point  6  3))
		(pt5 (make-2d-point  6 -1))
		(pt6 (make-2d-point  2 -1))
		;; polygon 2
		(pt7 (make-2d-point -6  1))
		(pt8 (make-2d-point -2  1))
		(pt9 (make-2d-point -2 -3))
		(pt10 (make-2d-point -6 -3))
		;; point in polygon 1
		(pt11 (make-2d-point 4  1)))
	(let* ((polygon1 (make1-polygon (list pt3 pt4 pt5 pt6)))
		   (polygon2 (make1-polygon (list pt7 pt8 pt9 pt10)))
		   (obstacle-list `(,polygon1 ,polygon2))
		   (waypoint-list `(,pt1 ,pt2 ,pt11))
		   (vg (make-visibility-graph obstacle-list waypoint-list)))
	  ;; Test GET-COST with point contained in polygon
	  (mtk-utils:test nil
					  (get-cost vg pt1 pt11)
					  :test #'equalp
					  :fail-info "")
	  ;; Test GET-COST with point outside of polygon
	  (mtk-utils:test  (+ (sqrt 5) 4 (sqrt 5) (sqrt 5) 4 (sqrt 5))
					  (get-cost vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  ;; Test GET-COST-FROM-PATH
	  (mtk-utils:test nil
	  				  (get-cost-from-path vg (list pt1
												   pt7 pt8 pt9 pt10 pt7
												   pt6 pt3 pt4 pt5 pt6
												   pt11))
	  				  :test #'equalp
	  				  :fail-info "")
	  ;; Test GET-PATH
	  (mtk-utils:test (list pt1 pt7 pt8 pt6 pt5 pt2)
					  (get-path vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  )))



(mtk-utils:deftest (test-vg-3 :test-visibility-graph) ()
;;  Creates an OBSTACLE-LIST and WAYPOINT-LIST,
;;  representing traversing a pair of square-obstacles that
;;  are origin symmetric, but eclipse each other at the origin.
  (let ((pt1 (make-2d-point -8  0))
		(pt2 (make-2d-point  8  0))
		;; polygon 1
		(pt3 (make-2d-point -1  3))
		(pt4 (make-2d-point  3  3))
		(pt5 (make-2d-point  3 -1))
		(pt6 (make-2d-point -1 -1))
		;; polygon 2
		(pt7 (make-2d-point -3  1))
		(pt8 (make-2d-point  1  1))
		(pt9 (make-2d-point  1 -3))
		(pt10 (make-2d-point -3 -3))
		;; point in polygon 1
		(pt11 (make-2d-point 0  0)))
	(let* ((polygon1 (make1-polygon (list pt3 pt4 pt5 pt6)))
		   (polygon2 (make1-polygon (list pt7 pt8 pt9 pt10)))
		   (obstacle-list `(,polygon1 ,polygon2))
		   (waypoint-list `(,pt1 ,pt2 ,pt11))
		   (vg (make-visibility-graph obstacle-list waypoint-list)))
	  ;; Test GET-COST with point contained in polygon
	  (mtk-utils:test nil
					  (get-cost vg pt1 pt11)
					  :test #'equalp
					  :fail-info "")
	  ;; Test GET-COST with point outside of polygon
	  (mtk-utils:test (+ (sqrt 58) 4 (sqrt 34))
					  (get-cost vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  ;; Test GET-COST-FROM-PATH
	  (mtk-utils:test nil
	  				  (get-cost-from-path vg (list pt1
												   pt7 pt8 pt9 pt10 pt7
												   pt6 pt3 pt4 pt5 pt6
												   pt11))
	  				  :test #'equalp
	  				  :fail-info "")
	  ;; Test GET-PATH
	  (mtk-utils:test (list pt1 pt3 pt4 pt2)
					  (get-path vg pt1 pt2)
					  :test #'equalp
					  :fail-info "")
	  )))